
import glutil.Camera;
import glutil.TrackBall;
import glutil.UVSphere;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.*;
import javax.media.opengl.*;


/**
 * A Jogl/OpenGL application that an 11-by-11-by-11 cube of
 * spheres in different colors.
 */
public class ColorCubeOfSpheres extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Color Cube of Spheres");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new ColorCubeOfSpheres());
      window.pack();
      window.setVisible(true);
   }
   
   private Camera camera;    // The camera that defines the view.
   private int displayList;  // A display list for drawing one sphere.
   private UVSphere sphere;  // For rendering a sphere.
   
   private JCheckBox useDisplayListCheckbox;  // If checked, use the display list;
                                              // If not checked, draw each sphere directly.

   private JLabel message;  // Displays the elapsed time in milliseconds for the most
                            // recent rendering of the display.
   
   /**
    * Make a regular JPanel to hold the GLJPanel,  Add a small panel at the
    * bottom to hold a check box that determines whether a display list is used
    * for rendering and a label that displays the rendering time of the most
    * recent display of the GLJPanel.  The user can rotate the view with the
    * mouse.
    */
   public ColorCubeOfSpheres() {
      useDisplayListCheckbox = new JCheckBox("Use Display List", true);
      message = new JLabel("Render time:");
      JPanel bottom = new JPanel();
      bottom.setLayout(new BorderLayout(20,20));
      bottom.add(useDisplayListCheckbox, BorderLayout.WEST);
      bottom.add(message, BorderLayout.CENTER);
      bottom.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
      GLJPanel drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      add(bottom, BorderLayout.SOUTH);
      camera = new Camera();
      camera.setLimits(-10,10,-10,10,-10,10);
      new TrackBall(drawable,camera);  // Enable rotation of the view by mouse.
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      sphere = new UVSphere(0.4);

      /* Create a display list that contains all the OpenGL
       * commands that are generated when the sphere is drawn.
       */

      displayList = gl.glGenLists(1);
      gl.glNewList(displayList, GL.GL_COMPILE);
      sphere.render(gl);
      gl.glEndList();
   }

   /**
    * Draw a color cube of spheres, 11 spheres on each side, using
    * different colors.  Each of the red, green, and blue components
    * of the color varies along one dimension of the cube.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      camera.apply(gl);
      boolean useList = useDisplayListCheckbox.isSelected();
      long start = System.currentTimeMillis();  // Starting time for rendering.
      for (int i = 0; i <= 10; i++) {
         float r = i/10.0f;
         for (int j = 0; j <= 10; j++) {
            float g = j/10.0f;
            for (int k = 0; k <= 10; k++) {
               float b = k/10.0f;
               gl.glColor3f(r,g,b);
               gl.glPushMatrix();
               gl.glTranslatef(i-5,j-5,k-5);
               if (useList)
                  gl.glCallList(displayList);
               else
                  sphere.render(gl);
               gl.glPopMatrix();
            }
         }
      }
      gl.glFlush();  // Make sure all commands are sent to graphics card.
      gl.glFinish(); // Wait for all commands to complete, before checking time.
      long time = System.currentTimeMillis() - start;
      message.setText("Render Time: " + time);
   }
   
   

   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
