
import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.media.opengl.*;

import scenegraph2D.*;

/**
 * A 2D OpenGL animation that demonstrates the use of a scene graph to implement
 * hierarchical modeling.  This program is a port of JoglHierarchicalModeling2D,
 * but it uses data structures rather than subroutines to represent objects in
 * the scene.
 */
public class JoglHMWithSceneGraph2D extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Basic JOGL Animation 2D");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new JoglHMWithSceneGraph2D());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The OpenGL display panel.
   private int frameNumber;    // Current frame number, increases by 1 in each frame.
   
   private ComplexObject2D theWorld;  // A data structure representing everything in the scene.
   
   /* The next three variables are required to implement the animation, since
    * the transform that is applied to these objects changes from one frame of
    * the animation to the next. */
   
   private SceneNode2D cart;     // The object that represents the cart.
   private SceneNode2D vanes;    // The rotating vanes of the windmill.
   private SceneNode2D wheel;    // The basic untransformed wheel.
   
   /**
    * Constructor creates the GLJPanel that will be used for drawing and adds
    *  it to the main panel.  It also starts a timer to draw the animation.  And
    *  it sets the preferred size to be 700-by-500.
    */
   public JoglHMWithSceneGraph2D() {
      buildTheWorld();  // Creates the data structure that represents the scene.
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(700,500));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      Timer timer = new Timer(30, new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            frameNumber++;
            drawable.repaint();
         }
      });
      timer.setInitialDelay(1000);
      timer.start();
   }
   
   /**
    * This method is called when the GLJPanel is created.  It initializes
    * the GL context.  Here, it sets the clear color to be sky blue and
    * it sets the xy-limits for drawing so that x ranges from 0 to 7 and
    * y ranges from -1 to 4.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0.5f, 0.5f, 1, 1);
      // The next three lines set up the coordinates system.
      gl.glMatrixMode(GL.GL_PROJECTION);
      gl.glOrtho(0, 7, -1, 4, -1, 1);
      gl.glMatrixMode(GL.GL_MODELVIEW);
   }
   
   /**
    * This method is called when the GLJPanel needs to be redrawn.
    * It draws the current frame of the animation.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT); // Fills the scene with blue.
      gl.glLoadIdentity();
      
      /* Set some transforms that depend on the current frame number */
      
      wheel.setRotation(frameNumber*20);
      vanes.setRotation(frameNumber * (180.0/46));
      cart.setTranslation(-3 + 13*(frameNumber % 300) / 300.0, 0);
      
      /* Now, draw the world by traversing the scene graph. */
      
      theWorld.draw(gl);
   }
   
   /**
    * This method creates the scene graph data structure that represents all the
    * objects in the scene.
    */
   private void buildTheWorld() {
      
      ComplexObject2D ground = new ComplexObject2D();
      ground.setColor(new Color(0.0f, 0.6f, 0.2f));
      ground.add(new PolygonNode(new double[] {-3,-1,1.5,1.65,5,-1} )); // first hill
      ground.add(new PolygonNode(new double[] {-3,-1,3,2.1,7,-1} )); // second hill
      ground.add(new PolygonNode(new double[] {0,-1,6,1.2,20,-1} )); // third hill
      
      PolygonNode road = new PolygonNode( new double[] {0,-0.5,7,-0.5,7,0.5,0,0.5} );
      road.setColor( new Color(0.4f, 0.4f, 0.5f) );

      LineNode stripe = new LineNode(0,0,7,0);  // the stripe on the road
      stripe.setColor(Color.WHITE);
      stripe.setWidth(6);
      
      ComplexObject2D sun = new ComplexObject2D();
      sun.setColor(Color.YELLOW);
      sun.add( new DiskNode(0.5) );
      for (int i = 0; i < 13; i++) {
         double angle = (2*Math.PI/13) * i;
         sun.add(new LineNode(0,0,Math.cos(angle),Math.sin(angle)));
      }
      sun.setTranslation(5,3.3);
      
      ComplexObject2D wheel = new ComplexObject2D();
      wheel.setColor(Color.BLACK);
      wheel.add( new DiskNode(1) );
      DiskNode hubcap = new DiskNode(0.8);
      hubcap.setColor(new Color(0.75f, 0.75f, 0.75f));
      wheel.add(hubcap);
      wheel.add( new DiskNode(0.2) );
      for (int i = 0; i < 15; i++) {
         double angle = (2*Math.PI/15) * i;
         wheel.add(new LineNode(0,0,Math.cos(angle),Math.sin(angle)));
      }
      
      ComplexObject2D cart = new ComplexObject2D();
      cart.setColor(Color.RED);
      cart.add( new TransformedObject(wheel, 0.8, 0, -1.5, -0.1) );
      cart.add( new TransformedObject(wheel, 0.8, 0, 1.5, -0.1) );
      cart.add( new PolygonNode( new double[] { -2.5, 0, 2.5, 0, 2.5, 2, -2.5, 2} ) );
      cart.setScaling(0.3);
      
      ComplexObject2D windmill = new ComplexObject2D();
      PolygonNode post = new PolygonNode( new double[] { -0.05, 0, 0.05, 0, 0.05, 3, -0.05, 3} );
      post.setColor( new Color(0.8f, 0.8f, 0.9f) );
      windmill.add(post);
      PolygonNode vane = new PolygonNode( new double[] { 0,0,0.5,0.1,1.5,0,0.5,-0.1} );
      vane.setColor( new Color(0.4f, 0.4f, 0.8f) );
      ComplexObject2D vanes = new ComplexObject2D();
      vanes.add(vane);
      vanes.add( new TransformedObject( vane, 1, 120, 0, 0) );
      vanes.add( new TransformedObject( vane, 1, 240, 0, 0) );
      vanes.setTranslation(0,3);
      windmill.add(vanes);

      theWorld = new ComplexObject2D();
      theWorld.add(ground);
      theWorld.add(road);
      theWorld.add(stripe);
      theWorld.add(sun);
      theWorld.add( new TransformedObject(windmill, 0.6, 0, 0.72, 1));
      theWorld.add( new TransformedObject(windmill, 0.4, 0, 2.2, 1.6));
      theWorld.add( new TransformedObject(windmill, 0.7, 0, 3.7, 0.8));
      theWorld.add(cart);
      
      this.cart = cart;    // record certain components in global variables, so they can be animated
      this.vanes = vanes;
      this.wheel = wheel;
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
