import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import javax.media.opengl.*;
import com.sun.opengl.util.BufferUtil;
import glutil.Camera;
import glutil.TrackBall;

/**
 * A Jogl/OpenGL application that draws the graph of a certain function
 * z = f(x,y), using a 201-by-201 grid of points to approximate
 * the graph.  The graph is illuminated by two colored lights,
 * as well as by one white light. The program optionally draws axes
 * and a grid lines on the surface, under control of the user.
 * At the user's option, the lights are fixed either with respect to
 * the world or with respect to the viewer.  The user can also turn
 * the three lights on and off and can turn ambient light on and off.
 * Controls for the user options are under the display area.
 * (Note that this program also serves as an example of using 
 * glDrawElements.)
 */
public class LightDemo extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Light Demo with Graph");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new LightDemo());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private Camera camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.

   private Camera untransformedCamera; // Camera with the same transform as the
                                       // initial one in the previous camera, before
                                       // rotation by the user.
   
   private JCheckBox axesCheckbox;     // Should axes be drawn?
   private JCheckBox linesCheckbox;    // Should a grid of lines be drawn on the surface?
   
   private JComboBox lightTransformSelect;  // Decide whether lights are fixed with respect
                                            // to the viewer or with respect to the world.
   
   private JCheckBox ambientLightCheckbox;  // checkboxed for turning lights on and off.
   private JCheckBox redLightCheckbox;
   private JCheckBox blueLightCheckbox;
   private JCheckBox whiteLightCheckbox;
   
   private FloatBuffer vBuf = BufferUtil.newFloatBuffer(201*201*3);  // Buffers for use with vertex arrays.
   private FloatBuffer nBuf = BufferUtil.newFloatBuffer(201*201*3);
   private IntBuffer surfaceIndexBuffer = BufferUtil.newIntBuffer(200*201*2);
   private IntBuffer xLineIndexBuffer = BufferUtil.newIntBuffer(21*201);
   private IntBuffer yLineIndexBuffer = BufferUtil.newIntBuffer(21*201);

   boolean version_1_5;  // Is this OpenGL 1.5 or higher?  If so, vertex buffer objects are used.
   
   int vertexVBO, normalVBO;            // ID's for vertex buffer objects, if used.
   int xLineVBO, yLineVBO, surfaceVBO;
   
   /**
    * Create the GLJPanel where the surface is drawn, with a set of controls underneath.
    */
   public LightDemo() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      axesCheckbox = new JCheckBox("Axes ", true);
      linesCheckbox = new JCheckBox("Lines");
      redLightCheckbox = new JCheckBox("Red Light", true);
      blueLightCheckbox = new JCheckBox("Blue Light", true);
      whiteLightCheckbox = new JCheckBox("White Light", true);
      ambientLightCheckbox = new JCheckBox("Ambient Light", true);
      lightTransformSelect = new JComboBox();
      lightTransformSelect.addItem("Viewer");
      lightTransformSelect.addItem("World");
      JPanel bottom = new JPanel();
      bottom.setLayout(new GridLayout(3,1,3,3));
      bottom.setBackground(Color.GRAY);
      bottom.setBorder(BorderFactory.createLineBorder(Color.GRAY, 2));
      add(bottom, BorderLayout.SOUTH);
      ActionListener al = new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            drawable.repaint();
         }
      };
      axesCheckbox.addActionListener(al);
      linesCheckbox.addActionListener(al);
      redLightCheckbox.addActionListener(al);
      blueLightCheckbox.addActionListener(al);
      whiteLightCheckbox.addActionListener(al);
      ambientLightCheckbox.addActionListener(al);
      lightTransformSelect.addActionListener(al);
      JPanel panel;
      panel = new JPanel();
      panel.add(new JLabel("Attach Lights To:"));
      panel.add(lightTransformSelect);
      bottom.add(panel);
      panel = new JPanel();
      panel.add(axesCheckbox);
      panel.add(Box.createHorizontalStrut(10));
      panel.add(linesCheckbox);
      panel.add(Box.createHorizontalStrut(10));
      panel.add(ambientLightCheckbox);
      bottom.add(panel);
      panel = new JPanel();
      panel.add(whiteLightCheckbox);
      panel.add(Box.createHorizontalStrut(10));
      panel.add(redLightCheckbox);
      panel.add(Box.createHorizontalStrut(10));
      panel.add(blueLightCheckbox);
      bottom.add(panel);
      untransformedCamera = new Camera();
      untransformedCamera.setView(40, 12, 12, 0, 0, 0, 0, 0, 1);
      untransformedCamera.setScale(7);
      camera = new Camera();
      camera.setView(40, 12, 12, 0, 0, 0, 0, 0, 1);
      camera.setScale(7);
      new TrackBall(drawable,camera);  
   }
   

   /**
    * OpenGL initialization.  Set up lights and material, create surface
    * data, vertex arrays, and (if used) vertex buffer options.  Turn on
    * essential options for lit 3D rendering.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0,1);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_LIGHTING);
      
      /* Configure light and material properties. */

      float[] gray = { 0.6f, 0.6f, 0.6f, 1 };
      float[] red = { 0.6f, 0, 0, 1 };
      float[] blue = { 0, 0, 0.6f, 1 };
      
      gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, gray, 0);
      gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, gray, 0);
      gl.glLightfv(GL.GL_LIGHT1, GL.GL_DIFFUSE, red, 0);
      gl.glLightfv(GL.GL_LIGHT1, GL.GL_SPECULAR, red, 0);
      gl.glLightfv(GL.GL_LIGHT2, GL.GL_DIFFUSE, blue, 0);
      gl.glLightfv(GL.GL_LIGHT2, GL.GL_SPECULAR, blue, 0);
      
      gl.glLightModeli(GL.GL_LIGHT_MODEL_TWO_SIDE, GL.GL_TRUE);
      gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE, gray, 0 );
      gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, new float[] {0.3f,0.3f,0.3f}, 0 );
      gl.glMateriali(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, 100);
      
      /* Create the surface data, storing it in Buffers and, if OpenGL version is 1.5
       * or higher, vertex buffer objects. */

      version_1_5 = gl.isExtensionAvailable("GL_VERSION_1_5");
      System.out.println("Version 1.5 available is: " + version_1_5);
      makeElementBuffers();
      computeSurfaceData();
      if (version_1_5) {
         int[] ids = new int[5];
         gl.glGenBuffers(5, ids, 0);
         vertexVBO = ids[0];
         normalVBO = ids[1];
         xLineVBO = ids[2];
         yLineVBO = ids[3];
         surfaceVBO = ids[4];
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vertexVBO);
         gl.glVertexPointer(3,GL.GL_FLOAT,0,0);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, normalVBO);
         gl.glNormalPointer(GL.GL_FLOAT,0,0);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, surfaceVBO);
         gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, 4*2*200*201, surfaceIndexBuffer, GL.GL_STATIC_DRAW);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, xLineVBO);
         gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, 4*21*201, xLineIndexBuffer, GL.GL_STATIC_DRAW);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, yLineVBO);
         gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, 4*21*201, yLineIndexBuffer, GL.GL_STATIC_DRAW);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vertexVBO);
         gl.glBufferData(GL.GL_ARRAY_BUFFER, 4*3*201*201, vBuf, GL.GL_DYNAMIC_DRAW);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, normalVBO);
         gl.glBufferData(GL.GL_ARRAY_BUFFER, 4*3*201*201, nBuf, GL.GL_DYNAMIC_DRAW);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
      }
      else {
         gl.glVertexPointer(3,GL.GL_FLOAT,0,vBuf);
         gl.glNormalPointer(GL.GL_FLOAT,0,nBuf);
      }
   }

   /**
    * Draw the surface, with options selected by the user.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      
      /* Position the lights and set up the projection and viewing transforms. */

      if (lightTransformSelect.getSelectedIndex() == 1) {
            // Lights should be attached to the world.  Transform them
            // along with the geometry.  To do this, set up their
            // positions after applying the transform for the current
            // camera position.
         camera.apply(gl);
         gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, new float[] {1,0,0,0}, 0);
         gl.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, new float[] {0,0,15,1}, 0);
         gl.glLightfv(GL.GL_LIGHT2, GL.GL_POSITION, new float[] {0,0,-15,1}, 0);
      }
      else {
            // Lights should be attached to the viewer.  Set their positions
            // subject to the original camera view, which was in effect when
            // then program started.  (So, before the user rotates the scene,
            // the scene looks the same whether the lights are attached to
            // the viewer or to the world.)
         untransformedCamera.apply(gl);
         gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, new float[] {1,0,0,0}, 0);
         gl.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, new float[] {0,0,15,1}, 0);
         gl.glLightfv(GL.GL_LIGHT2, GL.GL_POSITION, new float[] {0,0,-15,1}, 0);
         camera.apply(gl);  // Now, apply the transform for the geometry.
      }
      
      /* Enable or disable the red, blue, white lights. */
      
      if (whiteLightCheckbox.isSelected())
         gl.glEnable(GL.GL_LIGHT0);
      else
         gl.glDisable(GL.GL_LIGHT0);
      if (redLightCheckbox.isSelected())
         gl.glEnable(GL.GL_LIGHT1);
      else
         gl.glDisable(GL.GL_LIGHT1);
      if (blueLightCheckbox.isSelected())
         gl.glEnable(GL.GL_LIGHT2);
      else
         gl.glDisable(GL.GL_LIGHT2);
      
      /* Set ambient light to dark gray or black, depending on whether ambient light is selected. */
      
      if (ambientLightCheckbox.isSelected())
         gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, new float[] { 0.2f,0.2f,0.2f,1 }, 0);
      else
         gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, new float[] { 0,0,0,1 }, 0);
      
      /* Draw the axes, with lighting turned off, if the user wants them. */
      
      if (axesCheckbox.isSelected()) {
         gl.glDisable(GL.GL_LIGHTING);
         gl.glLineWidth(2);
         gl.glBegin(GL.GL_LINES);
         gl.glColor3f(1, 0, 0);
         gl.glVertex3f(0,0,0);
         gl.glVertex3f(6,0,0);
         gl.glColor3f(0, 0.7f, 0);
         gl.glVertex3f(0,0,0);
         gl.glVertex3f(0,6,0);
         gl.glColor3f(0.25f, 0.25f, 1);
         gl.glVertex3f(0,0,0);
         gl.glVertex3f(0,0,6);
         gl.glEnd();
         gl.glLineWidth(1);
         gl.glEnable(GL.GL_LIGHTING);
      }
      
      /* Draw the surface and, if the selected, an unlit grid of lines on the surface. */
      
      gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
      gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
      if ( ! linesCheckbox.isSelected() ) {
         drawSurface(gl);
      }
      else {
         gl.glPolygonOffset(1,1);
         gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
         drawSurface(gl);
         gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);
         gl.glDisable(GL.GL_LIGHTING);
         gl.glColor3f(0.75f,0.75f,0.75f);
         gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
         drawLinesOnSurface(gl);
         gl.glEnable(GL.GL_LIGHTING);
      }
      gl.glDisableClientState(GL.GL_VERTEX_ARRAY);
      gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
      
   } // end display()
      
   
   /**
    * Store the required indices for use with glDrawElements in the IntBuffers.
    */
   private void makeElementBuffers() {
      for (int i = 0; i < 201; i+=10) {
         for (int j = 0; j < 201; j++)
            xLineIndexBuffer.put(201*i + j);
      }
      for (int i = 0; i < 201; i+=10) {
         for (int j = 0; j < 201; j++)
            yLineIndexBuffer.put(201*j + i);
      }
      for (int i = 0; i < 200; i++) {
         for (int j = 0; j < 201; j++) {
            surfaceIndexBuffer.put(201*(i+1) + j);
            surfaceIndexBuffer.put(201*i + j);
         }
      }
      xLineIndexBuffer.rewind();
      yLineIndexBuffer.rewind();
      surfaceIndexBuffer.rewind();
   }
   
   /**
    * Compute a grid of points on the surface and normals to the surface
    * at those points.  The function is graphed for x and y in the range
    * from -5 to 5.  (This range could be changed, but other parts of the
    * program would have to be changed to adjust the view to the new range.)
    */
   private void computeSurfaceData() {
      double xmin = -5;
      double xmax = 5;
      double ymin = -5;
      double ymax = 5;
      double xRes = 200;
      double yRes = 200;
      float[] surfaceData = new float[201*3];
      float[] normalData = new float[201*3];
      double dx = (xmax - xmin)/xRes;
      double dy = (ymax - ymin)/yRes;
      for (int i = 0; i <= xRes; i++) {
         int v = 0;
         int n = 0;
         double y = ymin + dy*i;
         for (int j = 0; j <= yRes; j++) {
            double x = xmin + dx*j;
            double z1 = funcValue(x,y);
            float[] normal1 = computeUnitNormal(x, y);
            surfaceData[v++] = (float)x;
            surfaceData[v++] = (float)y;
            surfaceData[v++] = (float)z1;
            normalData[n++] = normal1[0];
            normalData[n++] = normal1[1];
            normalData[n++] = normal1[2];
         }
         vBuf.put(surfaceData,0,201*3);
         nBuf.put(normalData,0,201*3);
      }
      vBuf.rewind();
      nBuf.rewind();
   }
   
   
   /**
    * Draw the surface as a series of triangle strips, using glDrawElements and
    * the data that has been stored in the buffers.
    */
   private void drawSurface(GL gl) {
      if (version_1_5) {
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, surfaceVBO);
         for (int i = 0; i < 200; i++)
            gl.glDrawElements(GL.GL_TRIANGLE_STRIP, 402, GL.GL_UNSIGNED_INT, 402*i*4);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
      }
      else {
         for (int i = 0; i < 200; i++) {
            surfaceIndexBuffer.position(402*i);
            gl.glDrawElements(GL.GL_TRIANGLE_STRIP, 402, GL.GL_UNSIGNED_INT, surfaceIndexBuffer);
         }
      }
   }
   
   /**
    * Draw lines on the surface, but only every 10-th line, using glDrawElements and
    * the data that has been stored in the buffers.
    */
   private void drawLinesOnSurface(GL gl) {
      if (version_1_5) {
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, xLineVBO);
         for (int i = 0; i <= 20; i++)
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, 201*i*4);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, yLineVBO);
         for (int i = 0; i <= 20; i++)
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, 201*i*4);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
      }
      else {
         for (int i = 0; i <= 20; i++) {
            xLineIndexBuffer.position(201*i);
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, xLineIndexBuffer);
            yLineIndexBuffer.position(201*i);
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, yLineIndexBuffer);
         }
      }
   }
   
   /**
    * Compute the value of the function at the point (x,y).
    * (Other functions could be substituted here.)
    */
   private double funcValue(double x, double y) {
      return 3*Math.sin(x*x + y*y) / (1 + x*x + y*y);
   }
   
   /**
    * Compute a unit normal to the graph of the function at the point (x,y).
    * This is only an approximation, using nearby points instead of exact derivatives.
    */
   private float[] computeUnitNormal(double x, double y) {
      double epsilon = 0.00001;
      double z = funcValue(x,y);
      double z1 = funcValue(x+epsilon,y);
      double z2 = funcValue(x,y+epsilon);
      // normal is (epsilon,0,z1-z) X (0,epsilon,z2-z)
      double a = -epsilon*(z1-z);
      double b = -epsilon*(z2-z);
      double c = epsilon*epsilon;
      double length = Math.sqrt(a*a + b*b + c*c);
      if (Double.isNaN(length) || Double.isInfinite(length))
         return new float[] { 0,0,1 }; // Doesn't happen in this example!
      else
         return new float[] { (float)(a/length), (float)(b/length), (float)(c/length) };
   }
   
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
