
import java.awt.*;
import javax.swing.*;
import javax.media.opengl.*;

import com.sun.opengl.util.GLUT;

/**
 * A Jogl/OpenGL application that renders four spheres, with different settings,
 */
public class LitAndUnlitSpheres extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Four Spheres");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new LitAndUnlitSpheres());
      window.pack();
      window.setVisible(true);
   }
   
   private GLUT glut;  // For accessing some drawing routines from the
                       // OpenGL Utilties Toolkit.  glut is used here
                       // to draw the wireframe and solid spheres.
   
   /**
    * Make a regular JPanel to hold the GLJPanel (although in this
    * program, that's all it holds.
    */
   public LitAndUnlitSpheres() {
      GLJPanel drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      glut = new GLUT();
   }

   /**
    * Draw four spheres, with different settings. The spheres are
    * actually polyhedra with a rather small number of faces, not 
    * enough to give something that looks exactly like a sphere, 
    * even with lighting and smooth shading.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      gl.glLoadIdentity();
      
      /**
       * Draw the first two spheres without lighting.  When lighting
       * is off, the color of the things that are drawn are taken
       * from the current color.
       */
      
      gl.glDisable(GL.GL_LIGHTING);
      gl.glColor3f(1,1,1);  // Color for unlit spheres.
      
      /**
       * Draw a wireframe sphere with center at (-1.5, 0).
       */
      
      gl.glPushMatrix();
      gl.glTranslated(-1.5,0,0);
      gl.glRotated(15,1,1,1);  // Tilt the sphere for a better view.
      gl.glRotated(90,-1,0,0); // Rotate axis of sphere from z-axis onto y-axis.
      glut.glutWireSphere(0.45, 16, 8);  // Sphere with radius 0.45, 16 "slices" and 8 "stacks".
      gl.glPopMatrix();
      
      /**
       * Draw a solid sphere without lighting.  this shows up as a flat patch
       * of color, since there is none of the shading that would be produced
       * by the lighting.
       */

      gl.glPushMatrix();
      gl.glTranslated(-0.5,0,0);
      gl.glRotated(15,1,1,1);
      gl.glRotated(90,-1,0,0);
      glut.glutSolidSphere(0.45, 16, 8);
      gl.glPopMatrix();
      
      /**
       * Draw the next two spheres with lighting.  When lighting
       * is enabled, the current color is ignored and the current
       * "material" is used instead.  The default material is a
       * light gray.  The previous call to glColor3f has no effect
       * on the drawing.
       */
      
      gl.glEnable(GL.GL_LIGHTING);
      
      /**
       * Draw a llit sphere using flat shading.
       */
      
      gl.glShadeModel(GL.GL_FLAT);
      gl.glPushMatrix();
      gl.glTranslated(0.5,0,0);
      gl.glRotated(15,1,1,1);
      gl.glRotated(90,-1,0,0);
      glut.glutSolidSphere(0.45, 16, 8);
      gl.glPopMatrix();
      
      /**
       * Draw a lit sphere using smooth shading.
       */
      
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glPushMatrix();
      gl.glTranslated(1.5,0,0);
      gl.glRotated(15,1,1,1);
      gl.glRotated(90,-1,0,0);
      glut.glutSolidSphere(0.45, 16, 8);
      gl.glPopMatrix();
      
   }
   
   /**
    * This method is called when the size of the display area is set.
    * It is used here to set the view volume, with limits that will
    * make sure that all four spheres are in view.
    */
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
      GL gl = drawable.getGL();
      double s = 2.8;  // limits of cube that we want to view go from -s to s.
      gl.glMatrixMode(GL.GL_PROJECTION);
      gl.glLoadIdentity();  // Start with the identity transform.
      if (width > height) {  // Expand x limits to match viewport aspect ratio.
         double ratio = (double)width/height;
         gl.glOrtho(-s*ratio, s*ratio, -s, s, -s, s);
      }
      else {  // Expand y limits to match viewport aspect ratio.
         double ratio = (double)height/width;
         gl.glOrtho(-s, s, -s*ratio, s*ratio, -s, s);
      }
      gl.glMatrixMode(GL.GL_MODELVIEW);
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
