
import glutil.*;
import simplescenegraph3d.*;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.media.opengl.*;

/**
 * This program constructs a simple animated scene (using the classes
 * from package simplescenegraph3d).  Included in the scene graph are
 * two "AvatarNodes" which represent viewers as nodes in a scene graph.
 * A pop-up menu below the scene allows the user to see the scene from
 * the point of view of either of these viewers, as well as from the
 * default "global" view.  The global view is represented by a
 * camera object and allows the user to rotate the scene by dragging
 * with the mouse.
 */
public class MovingCameraDemo extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Moving Camera Demo");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new MovingCameraDemo());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   
   private ComplexObject3D world;  // The root node in the scene graph.
   
   private ComplexObject3D wheel;  // Various nodes in the scene graph, used for animation.
   private ComplexObject3D cart;
   private ComplexObject3D horse1;
   private ComplexObject3D horse2;
   private ComplexObject3D horseWithRider;
   private ComplexObject3D merryGoRound;

   private AvatarNode viewerInCar;  // A node in the scene graph representing a viewer riding the cart.
   private AvatarNode viewerOnMerryGoRound;  // A node representing a viewer riding the merry-go-round.
   
   private Camera globalViewer;  // The Camera that represents the global view from "outside" the scene.

   private UVSphere sphere = new UVSphere(48, 32);     // Some shapes to put in the world.
   private Cube cube = new Cube();
   private UVCone cone = new UVCone(0.5,2);
   private UVCylinder cylinder = new UVCylinder(0.5,2);
   
   private JComboBox viewSelect;   // A pop-up menu for selecting point of view.
   
   private int frameNumber;    // Drives the animation.  A timer adds 1
                                // to this value every 30 milliseconds
                                // and repaints the panel.

   private JCheckBox animate;  // Controls whether the animation is running.

   private Timer animator;  // The timer that drives the animation,
                             // which is started and stopped in response
                             // to the animate checKbox.



   /**
    * Creates the OpenGL drawable and adds it to the panel.  Also sets up the
    * user interface and builds the scene graph.
    */
   public MovingCameraDemo() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,500));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      JPanel bottom = new JPanel();
      bottom.setLayout(new FlowLayout(FlowLayout.CENTER));
      add(bottom, BorderLayout.SOUTH);
      animate = new JCheckBox("Animate");
      animate.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            if (animate.isSelected())
               animator.start();
            else
               animator.stop();
         }
      });
      bottom.add(animate);
      animator = new Timer(30, new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            frameNumber++;
            drawable.repaint();
         }
      });      
      viewSelect = new JComboBox();
      viewSelect.addItem("Global View");
      viewSelect.addItem("View from Front of Cart");
      viewSelect.addItem("View from Merry-Go-Round");
      viewSelect.addActionListener( new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            drawable.repaint();
         }
      });
      bottom.add(Box.createHorizontalStrut(10));
      bottom.add(new JLabel("Select view: "));
      bottom.add(viewSelect);
      globalViewer = new Camera();
      globalViewer.setScale(20);
      globalViewer.setView(0, 10, 40, 0, 0, 0, 0, 1, 0);
      new TrackBall(drawable,globalViewer);  
      buildWorld();
   }
   
   
   /**
    * Build the scene graph that represents the world.  This method is
    * called just once, from the constructor.
    */
   private void buildWorld() {
      viewerInCar = new AvatarNode();
      viewerOnMerryGoRound = new AvatarNode();
      world = new ComplexObject3D();
      
      // Add a big, flat cylinder to represent the ground and an annulus to represent a road.
      world.add(new RenderableNode(new UVCylinder(15,1,96,1,10))
                       .setColor(new Color(0,120,0))
                       .setRotation(90,-1,0,0)
                       .setTranslation(0, -1, 0));
      world.add(new RenderableNode(new UVDisk(12.5, 11, 96, 2))
                       .setColor(Color.LIGHT_GRAY)
                       .setRotation(90,-1,0,0)
                       .setTranslation(0, 0.01, 0));
      
      // Add a tree, consisting of a cone and a cylinder.
      
      ComplexObject3D tree = new ComplexObject3D();
      tree.add(new RenderableNode(cone).setTranslation(0,0,1).setColor(new Color(0,180,25)));
      tree.add(new RenderableNode(cylinder).setScaling(0.3,0.3,0.5).setColor(new Color(80,20,0)));
      world.add(new ComplexObject3D(tree).setScaling(3).setRotation(90,-1,0,0).setTranslation(-2,0,-5));
      
      // Add some random shapes around the periphery.
      
      for (double angle = 0; angle < 2*Math.PI; angle += Math.PI / 10) {
            SceneNode3D r;
            switch ((int)(4*Math.random())) {
            case 0:
               r = new RenderableNode(cube).setScaling(0.5,1,0.5);
               break;
            case 1:
               r = new RenderableNode(sphere).setScaling(0.5);
               break;
            case 2:
               r = new RenderableNode(cone).setRotation(90,-1,0,0).setTranslation(0,-1,0);
               break;
            default:
               r = new RenderableNode(cylinder).setRotation(90,-1,0,0).setTranslation(0,-1,0);
            }
            Color c = Color.getHSBColor((float)Math.random(), (float)(0.5+0.5*Math.random()), 0.8f);
            double x = 14*Math.cos(angle);
            double z = 14*Math.sin(angle);
            world.add(new ComplexObject3D(r).setColor(c).setTranslation(x,1,z));
      }
      
      // Create a cart, consisting of a cube and four wheels.  One of the
      // viewer nodes is also added to the card.
      
      wheel = new ComplexObject3D();
      wheel.add(new ComplexObject3D( new RenderableNode( new UVTorus(1,0.75,48,32))).setColor(Color.BLUE));
      SceneNode3D spoke = new RenderableNode(new UVCylinder())
                                .setScaling(0.2,0.2,1.7)
                                .setRotation(90,-1,0,0)
                                .setTranslation(0,-0.85,0)
                                .setColor(Color.YELLOW);
      wheel.add( spoke );
      wheel.add( new ComplexObject3D(spoke).setRotation(60,0,0,1));
      wheel.add( new ComplexObject3D(spoke).setRotation(-60,0,0,1));
      cart = new ComplexObject3D();
      cart.add(new ComplexObject3D(wheel).setTranslation(2,1,1));
      cart.add(new ComplexObject3D(wheel).setTranslation(-2,1,1));
      cart.add(new ComplexObject3D(wheel).setTranslation(2,1,-1));
      cart.add(new ComplexObject3D(wheel).setTranslation(-2,1,-1));
      cart.add(new RenderableNode(new Cube())
                     .setScaling(2.5,0.3,0.8)
                     .setTranslation(0,1,0)
                     .setColor(Color.RED));
      cart.add(viewerInCar.setRotation(-90, 0, 1, 0).setTranslation(.8, 2, 0));
      cart = new ComplexObject3D(cart.setTranslation(0,0,11.5));
      world.add(cart);
      
      // Create a merry-go-round.  The "horses" on the merry-go-round are just
      // cubes that go up and down.  There are two groups of horses; one group
      // goes up while the other goes down.  One of the viewer nodes is added
      // as a "rider" to one of the horses.
      
      SceneNode3D horse = new RenderableNode(new Cube())
                   .setScaling(1,0.5,0.5)
                   .setTranslation(0,2.5,5)
                   .setColor(Color.GREEN);
      horse1 = new ComplexObject3D(horse);
      horse2 = new ComplexObject3D(horse);
      horseWithRider = new ComplexObject3D(horse);
      horseWithRider.add(viewerOnMerryGoRound);
      viewerOnMerryGoRound.setRotation(-80,0,1,0).setTranslation(1,3,5);
      merryGoRound = new ComplexObject3D();
      merryGoRound.add(new RenderableNode(new UVCylinder()) // top of merry-go-round
                       .setScaling(12,12,0.25)
                       .setRotation(90,-1,0,0)
                       .setTranslation(0,0.5,0)
                       .setColor(Color.LIGHT_GRAY));
      merryGoRound.add(new RenderableNode(new UVCylinder(0.5,1,32,1))  // bottom
                        .setScaling(12,12,0.25)
                        .setRotation(90,-1,0,0)
                        .setTranslation(0,4.5,0)
                        .setColor(Color.LIGHT_GRAY));
      SceneNode3D pole = new RenderableNode(new UVCylinder(0.5,1,32,1))
                      .setColor(Color.BLUE)
                      .setScaling(0.2,0.2,4.2)
                      .setRotation(90,-1,0,0)
                      .setTranslation(0,0.8,5);
      merryGoRound.add(new ComplexObject3D(pole));
      merryGoRound.add(horseWithRider);
      for (int i = 1; i < 6; i++) {
         merryGoRound.add(new ComplexObject3D(pole).setRotation(60*i,0,1,0));
         merryGoRound.add(new ComplexObject3D( i % 2 == 0 ? horse1 : horse2 ).setRotation(60*i,0,1,0));
      }
      world.add(new ComplexObject3D(merryGoRound).setScaling(0.8).setTranslation(2,0,4));
      
   } // end buildTheWorld()
   

   /**
    * initialize the GL context
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0.3f,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      gl.glEnable(GL.GL_NORMALIZE);
   }

   
   /**
    * Draw a the world, from the point of view of the viewer currently selected
    * by the user.  Also, apply the transformations appropriate for the current
    * frame number, to implement the animation.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

      // Apply transformations for the current frame.
      // NOTE: This MUST be done before setting up the view, since some
      // of the transformations affect the AvatarNodes that represent two
      // of the possible views.
      
      wheel.setRotation(-(frameNumber*5), 0, 0, 1); // make the car wheels rotate
      cart.setRotation(frameNumber, 0,1,0);  // make the car go around the road
      merryGoRound.setRotation((int)(2.7*frameNumber), 0, 1, 0);  // make the merry-go-round turn
      int horseFrame = frameNumber % 60;  // the remaining stuff makes the horses go up and down
      if (horseFrame >= 30)
         horseFrame = 60-horseFrame;
      double horseOffset = (horseFrame - 15) / 30.0;
      horse1.setTranslation(0, horseOffset, 0);
      horse2.setTranslation(0, -horseOffset, 0);
      horseWithRider.setTranslation(0,horseOffset,0);
      
      // Set up the projection and view for the currently selected viewer.

      switch (viewSelect.getSelectedIndex()) {
      case 0:
         globalViewer.apply(gl);
         break;
      case 1:
         viewerInCar.apply(gl);
         break;
      case 2:
         viewerOnMerryGoRound.apply(gl);
      }
      
      world.draw(gl);  // render the world.
   }
      
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
