
import glutil.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;


/**
 * A Jogl/OpenGL application that draws several wireframe
 * "paddle wheels."  An animation of the wheels rotating
 * can be turned on and off by a checkbox below the display.
 * Each paddle wheel is just a bunch of rectangles.
 * The user can use the mouse to rotate the entire image.
 */
public class PaddleWheels extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Paddle Wheels");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new PaddleWheels());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private Camera camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.

   private int frameNumber;    // Drives the animation.  A timer adds 1
                               // to this value every 30 milliseconds
                               // and repaints the panel.
   
   private JCheckBox animate;  // Controlls whether the animation is running.
   
   private Timer animator;  // The timer that drives the animation,
                            // which is started and stopped in response
                            // to the animate checlbox.
   
   /**
    * Make a regular JPanel to hold the GLJPanel (although in this
    * program, that's all it holds.
    */
   public PaddleWheels() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      animate = new JCheckBox("Animate the rotation of the wheels");
      animate.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            if (animate.isSelected())
               animator.start();
            else
               animator.stop();
         }
      });
      JPanel bottom = new JPanel();
      bottom.setLayout(new FlowLayout(FlowLayout.CENTER));
      bottom.add(animate);
      add(bottom, BorderLayout.SOUTH);
      animator = new Timer(30, new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            frameNumber++;
            drawable.repaint();
         }
      });      
      camera = new Camera();
      camera.setScale(6);
      new TrackBall(drawable,camera);  
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      gl.glLightModeli(GL.GL_LIGHT_MODEL_TWO_SIDE, GL.GL_TRUE);
   }

   /**
    * Draw a set of three "paddle wheels", which rotate when
    * animation is turned on.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      camera.apply(gl);
            
      /**
       * A red paddle wheel with 20 flaps, centered at -4, -1
       */
      
      gl.glColor3f(1,0,0);
      gl.glPushMatrix();
      gl.glTranslated(-4,-1,0);
      drawPaddleWheel(gl,20);
      gl.glPopMatrix();
      
      /**
       * A green paddle wheel with 32 flaps, centered at -4, -1
       */
      
      gl.glColor3f(0,0.7f,0);
      gl.glPushMatrix();
      gl.glTranslated(0,1.5,0);
      drawPaddleWheel(gl,32);
      gl.glPopMatrix();
      
      /**
       * A blue paddle wheel with 12 flaps, centered at -4, -1
       */
      
      gl.glColor3f(0.4f,0.4f,1);
      gl.glPushMatrix();
      gl.glTranslated(4,-1,0);
      drawPaddleWheel(gl,12);
      gl.glPopMatrix();
      
   }
   
   /**
    * Draws a "paddle wheel" with a specified number of sides.
    * The wheel actually just consists of a bunch of identical
    * polygons which have been fanned out into a circle around
    * the x-axis.  The wheel is centered at (0,0,0), is 2 units
    * wide and 4 units tall.  The wheel has a built in rotation
    * about the x-axis, which depends on the frame number, 
    * giving a wheel that rotates completely in 360 frames.
    */
   private void drawPaddleWheel(GL gl, int paddles) {
      gl.glPushMatrix();
      gl.glRotated(frameNumber, 1, 0, 0);
      for (int i = 0; i < paddles; i++) {
         gl.glRotated(360.0/paddles, 1, 0, 0);
         gl.glBegin(GL.GL_POLYGON);
            // Note:  We should specify a normal for the polygon here.
            // I get away without doing it because the normal to the
            // polygon is (1,0,0), which happens to be the default.
         gl.glVertex2d(-0.7,1);
         gl.glVertex2d(0.7,1);
         gl.glVertex2d(1,2);
         gl.glVertex2d(-1,2);
         gl.glEnd();
      }
      gl.glPopMatrix();
   }
   
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
