
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Point2D;
import java.util.ArrayList;

import javax.swing.*;
import javax.media.opengl.*;

/**
 * This program lets the user experiment with various primitive types
 * and various options.  Drawing is 2D, with no lighting.  The program
 * draws a primitive using several points.  The user can drag the points
 * around, can delete all the points, and can add a point by double-clicking.
 * A set of UI components let the user select the primitive type to be
 * drawn as well as other options such as line stippling.
 */
public class PrimitiveTypes extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Primitive Explorer");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new PrimitiveTypes());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private ArrayList<Point2D> points = new ArrayList<Point2D>();
   
   private static int[] sizes = { 1, 2, 3, 4, 5, 7, 10, 20 };
   
   private static int[] types = {
      GL.GL_POINTS, GL.GL_LINES, GL.GL_LINE_STRIP, GL.GL_LINE_LOOP, GL.GL_POLYGON,
      GL.GL_TRIANGLES, GL.GL_TRIANGLE_STRIP, GL.GL_TRIANGLE_FAN,
      GL.GL_QUADS, GL.GL_QUAD_STRIP
   };
   
   private JButton resetButton, clearButton;
   
   private JComboBox pointSizeSelect, lineWidthSelect;
   
   private JComboBox stipplePatternSelect, stippleMultiplierSelect;
   
   private JComboBox primitiveSelect;
   
   private JCheckBox antialiasBox, stippleBox, smoothBox, colorBox;
      
   /**
    * Make a regular JPanel to hold the GLJPanel (although in this
    * program, that's all it holds.
    */
   public PrimitiveTypes() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(500,500));
      drawable.addGLEventListener(this);
      drawable.addMouseListener( new Mouser());
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      resetButton = new JButton("RESET");
      clearButton = new JButton("CLEAR");
      pointSizeSelect = new JComboBox();
      addToSelect(pointSizeSelect, new String[] { "1", "2", "3", "4", "5", "7", "10", "20" });
      lineWidthSelect = new JComboBox();
      addToSelect(lineWidthSelect, new String[] { "1", "2", "3", "4", "5", "7", "10", "20" });
      stippleMultiplierSelect = new JComboBox();
      addToSelect(stippleMultiplierSelect, new String[] { "1", "2", "3", "4", "5" });
      stipplePatternSelect = new JComboBox();
      addToSelect(stipplePatternSelect, new String[] { "FF00", "F0CC", "5555", "EAEA", "0101" });
      primitiveSelect = new JComboBox();
      addToSelect(primitiveSelect, new String[] {
            "GL_POINTS", "GL_LINES", "GL_LINE_STRIP", "GL_LINE_LOOP", "GL_POLYGON",
            "GL_TRIANGLES", "GL_TRIANGLE_STRIP", "GL_TRIANGLE_FAN",
            "GL_QUADS", "GL_QUAD_STRIP"
         } );
      antialiasBox = new JCheckBox("Line and point antialiasing");
      stippleBox = new JCheckBox("Line stippling");
      smoothBox = new JCheckBox("GL_SMOOTH");
      colorBox = new JCheckBox("Colors", true);
      ActionListener listener = new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            drawable.repaint();
         }
      };
      pointSizeSelect.addActionListener(listener);
      lineWidthSelect.addActionListener(listener);
      stipplePatternSelect.addActionListener(listener);
      stippleMultiplierSelect.addActionListener(listener);
      primitiveSelect.addActionListener(listener);
      antialiasBox.addActionListener(listener);
      stippleBox.addActionListener(listener);
      colorBox.addActionListener(listener);
      smoothBox.addActionListener(listener);
      resetButton.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            reset();  // return to default state.
            drawable.repaint();
         }
      });
      clearButton.addActionListener( new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            points.clear();  // Delete all points.
            drawable.repaint();
         }
      });
      reset();
      JPanel bottom = new JPanel();
      bottom.setBackground(Color.LIGHT_GRAY);
      bottom.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY,3));
      bottom.setLayout(new GridLayout(0,1));
      add(bottom, BorderLayout.SOUTH);
      JPanel row = new JPanel();
      row.setBackground(Color.WHITE);
      bottom.add(row);
      row.add(new JLabel("Primitive Type:"));
      row.add(primitiveSelect);
      row.add(Box.createHorizontalStrut(25));
      row.add(resetButton);
      row.add(Box.createHorizontalStrut(25));
      row.add(clearButton);
      row = new JPanel();
      row.setBackground(Color.WHITE);
      bottom.add(row);
      row.add(new JLabel("Point Size:"));
      row.add(pointSizeSelect);
      row.add(Box.createHorizontalStrut(25));
      row.add(new JLabel("Line Width:"));
      row.add(lineWidthSelect);
      row = new JPanel();
      row.setBackground(Color.WHITE);
      bottom.add(row);
      row.add(colorBox);
      row.add(Box.createHorizontalStrut(25));
      row.add(smoothBox);
      row.add(Box.createHorizontalStrut(25));
      row.add(antialiasBox);
      row = new JPanel();
      row.setBackground(Color.WHITE);
      bottom.add(row);
      row.add(stippleBox);
      row.add(Box.createHorizontalStrut(25));
      row.add(new JLabel("Pattern:"));
      row.add(stipplePatternSelect);
      row.add(Box.createHorizontalStrut(25));
      row.add(new JLabel("Multiplier:"));
      row.add(stippleMultiplierSelect);
   }
   
   private void addToSelect(JComboBox select, String[] items) {
      for (String item : items)
         select.addItem(item);
   }
   
   /**
    * A class for letting the user drag the points and
    * create new points by double-clicking.
    */
   private class Mouser implements MouseListener, MouseMotionListener  {
      int startX,startY;
      Point2D startPoint;
      int draggedPoint = -1;
      public void mousePressed(MouseEvent e) {
         if (draggedPoint < 0 && e.getClickCount() == 2) {
            double x = -1.5 + ((double)e.getX()/drawable.getWidth())*3;
            double y = 1.5 - ((double)e.getY()/drawable.getHeight())*3;
            points.add(new Point2D.Double(x,y));
            drawable.repaint();
            return;
         }
         draggedPoint = -1;
         int x = e.getX();
         int y = e.getY();
         for (int i = 0; i < points.size(); i++) {
            int px = (int)(((points.get(i).getX()+1.5)/3)*drawable.getWidth());
            int py = (int)(((1.5-points.get(i).getY())/3)*drawable.getHeight());
            if (Math.abs(px-x) < 8 && Math.abs(py-y) < 8) {
               draggedPoint = i;
               startX = x;
               startY = y;
               startPoint = new Point2D.Double(points.get(i).getX(), points.get(i).getY());
               drawable.addMouseMotionListener(this);
               return;
            }
         }
      }
      public void mouseReleased(MouseEvent e) {
         if (draggedPoint < 0)
            return;
         drawable.removeMouseMotionListener(this);
         draggedPoint = -1;
         startPoint = null;
      }
      public void mouseDragged(MouseEvent e) {
         if (draggedPoint < 0)
            return;
         int x = e.getX();
         int y = e.getY();
         if (x < 0)
            x = 0;
         else if (x > drawable.getWidth())
            x = drawable.getWidth();
         if (y < 0)
            y = 0;
         else if (y > drawable.getHeight())
            y = drawable.getHeight();
         double dx = ((double)(x-startX)/drawable.getWidth())*3;
         double dy = ((double)(startY-y)/drawable.getHeight())*3;
         points.get(draggedPoint).setLocation(startPoint.getX()+dx, startPoint.getY()+dy);
         drawable.repaint();
      }
      public void mouseClicked(MouseEvent e) {}
      public void mouseEntered(MouseEvent e) {}
      public void mouseExited(MouseEvent e) {}
      public void mouseMoved(MouseEvent e) {}
   }
   
   private void reset() {
      points.clear();
      for (int i = 0; i < 8; i++) {
         double angle = (2*Math.PI/8) * i;
         Point2D pt = new Point2D.Double(Math.cos(angle), Math.sin(angle));
         points.add(pt);
      }
      antialiasBox.setSelected(false);
      stippleBox.setSelected(false);
      smoothBox.setSelected(false);
      colorBox.setSelected(true);
      stipplePatternSelect.setSelectedItem("FF00");
      stippleMultiplierSelect.setSelectedIndex(0);
      lineWidthSelect.setSelectedIndex(0);
      pointSizeSelect.setSelectedIndex(0);
      primitiveSelect.setSelectedIndex(0);
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0,1);
      gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
      gl.glMatrixMode(GL.GL_PROJECTION);
      gl.glLoadIdentity();
      gl.glOrtho(-1.5,1.5,-1.5,1.5,-1,1);
      gl.glMatrixMode(GL.GL_MODELVIEW);
      gl.glLoadIdentity();
   }

   /**
    * Draw a set of three "paddle wheels", which rotate when
    * animation is turned on.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT);
      
      // First, install all user-selected options.
      short pattern = 0;
      if (stippleBox.isSelected()) {
         try {
            int x = Integer.parseInt(((String)stipplePatternSelect.getSelectedItem()).trim(), 16);
            if (x < 0 || x > 0xFFFF)
               throw new Exception();
            pattern = (short)x;
         }
         catch (Exception e) {  // can't happen!
            pattern = (short)0xFF00;
         }
      }
      if (smoothBox.isSelected())
         gl.glShadeModel(GL.GL_SMOOTH);
      else
         gl.glShadeModel(GL.GL_FLAT);
      if (stippleBox.isSelected())
         gl.glEnable(GL.GL_LINE_STIPPLE);
      else
         gl.glDisable(GL.GL_LINE_STIPPLE);
      if (antialiasBox.isSelected()) {
         gl.glEnable(GL.GL_POINT_SMOOTH);
         gl.glEnable(GL.GL_LINE_SMOOTH);
         gl.glEnable(GL.GL_BLEND);
      }
      else {
         gl.glDisable(GL.GL_POINT_SMOOTH);
         gl.glDisable(GL.GL_LINE_SMOOTH);
         gl.glDisable(GL.GL_BLEND);
      }
      gl.glPointSize( sizes[ pointSizeSelect.getSelectedIndex() ] );
      gl.glLineWidth( sizes[ lineWidthSelect.getSelectedIndex() ] );
      gl.glLineStipple( stippleMultiplierSelect.getSelectedIndex()+1, pattern );
      
      // Draw the primitive.
      
      gl.glColor3d(0.7,0.7,0.7);
      int primitiveNum = primitiveSelect.getSelectedIndex();
      gl.glBegin( types[primitiveNum] ); // Draw selected primitive.
      for (int i = 0; i < points.size(); i++) {
         if (colorBox.isSelected()) {
            Color c = Color.getHSBColor( (float)i/points.size(), 1, 1 );
            gl.glColor3fv(c.getColorComponents(null),0);
         }
         Point2D p = points.get(i);
         gl.glVertex2d( p.getX(), p.getY() );
      }
      gl.glEnd();
      
      // Draw extra stuff.
      
      gl.glDisable(GL.GL_LINE_SMOOTH);  // set up for drawing default lines
      gl.glDisable(GL.GL_BLEND);
      gl.glDisable(GL.GL_LINE_STIPPLE);
      gl.glLineWidth(1);
      if (primitiveNum >= 4) { // draw polygon edges in black
         gl.glColor3f(0,0,0);
         gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
         gl.glBegin( types[primitiveNum] );
         for (int i = 0; i < points.size(); i++) {
            Point2D p = points.get(i);
            gl.glVertex2d( p.getX(), p.getY() );
         }
         gl.glEnd();
         gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL);
      }
      gl.glColor3f(1,1,1);
      double radius = (7.0/getWidth())*3;
      for (Point2D p : points) {  // white hexagon around each point
         gl.glBegin(GL.GL_LINE_LOOP);
         double x = p.getX();
         double y = p.getY();
         for (int j = 0; j < 6; j++) {
            double angle = (2*Math.PI/6)*j;
            gl.glVertex2d(x + radius*Math.cos(angle), y + radius*Math.sin(angle));
         }
         gl.glEnd();
      }
   }
   
   
   
   
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
