import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;


/**
 * A panel that shows a GeneralPath made up of five quadratic Bezier curves.
 * The curves are shown along with their control points, with dotted lines
 * from the control point to the endpoints of the curve.  The user can drag
 * the endpoints and the control points.
 * This class also contains a main() routine that opens a window that
 * displays a panel of the type defined by this class.
 */
public class QuadraticBezierEdit extends JPanel {
   
   public static void main(String[] args) {
      JFrame frame = new JFrame("QuadraticBezierEdit");
      frame.setContentPane( new QuadraticBezierEdit() );
      frame.pack();
      frame.setLocation(200,150);
      frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      frame.setVisible(true);
   }
   
   private AffineTransform transform;  // The transform that was applied to the
                                       // graphics context the last time this 
                                       // panel was painted. (This is set in the
                                       // method applyLimits().)
   
   private float pixelSize;  // The width and height of one pixel, as measured by
                             // the coordinate system that was used the last time
                             // the panel was painted.  (This is set in the method
                             // applyLimits().)
   
   private double[] limitsRequested = new double[] { -5, 5, 5, -5 };
       // A coordinate system will be applied to this panel that is just
       // large enough to show the rectangle with corners at (limits[0],limits[1])
       // and (limits[2],limits[3]), allowing for the unit length in both
       // the horizontal and vertical directions to be preserved.
   
   private Point2D.Float[] points = {   // The points that define the Bezier curves.
         new Point2D.Float(-4,0),     // First endpoint.
         new Point2D.Float(-3,4),     // Control point for first curve.
         new Point2D.Float(-1,3),     // Second endpoint.
         new Point2D.Float(1,3.5F),   // Control point for second curve.
         new Point2D.Float(0,2),      // Etc.
         new Point2D.Float(-1,-2),
         new Point2D.Float(1,-1),
         new Point2D.Float(0.5F,-2),
         new Point2D.Float(3,-3),
         new Point2D.Float(2.5F,-1),
         new Point2D.Float(4,0),
   };

   
   /**
    * The constructor defines a QuadraticBezierEdit panel with background color white,
    * preferred size (500,500), and a gray border.   It also sets up mouse listeners
    * to allow the user to drag the points.
    */
   public QuadraticBezierEdit() {
      setPreferredSize( new Dimension(500,500) );
      setBackground(Color.WHITE);
      setBorder(BorderFactory.createLineBorder(Color.GRAY,2));
      MouseHandler mh = new MouseHandler();
      addMouseListener( mh );
      addMouseMotionListener( mh );
   }
   
   
   /**
    * Draws the Quadratic Bezier curves in this panel.
    */
   protected void paintComponent(Graphics g) {
      super.paintComponent(g);
      Graphics2D g2 = (Graphics2D)g.create();
      g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      applyLimits(g2, getWidth(), getHeight(), limitsRequested, true);
      GeneralPath path = new GeneralPath();
      path.moveTo(points[0].x, points[0].y);
      for (int i = 1; i < points.length; i += 2)
         path.quadTo(points[i].x, points[i].y, points[i+1].x, points[i+1].y);
      g2.setStroke( new BasicStroke(pixelSize,BasicStroke.CAP_BUTT,BasicStroke.JOIN_ROUND,0,new float[] { 2*pixelSize,2*pixelSize }, 0) );
      g2.setColor( Color.BLUE );
      for (int i = 1; i < points.length; i += 2) {
         g2.draw( new Line2D.Float(points[i].x,points[i].y,points[i-1].x,points[i-1].y) );
         g2.draw( new Line2D.Float(points[i].x,points[i].y,points[i+1].x,points[i+1].y) );
      }
      for (int i = 0; i < points.length;  i++) {
         if ( (i & 1) == 0)
            g2.setColor(Color.BLACK);
         else
            g2.setColor(Color.BLUE);
         g2.fill( new Ellipse2D.Float(points[i].x - 4*pixelSize, points[i].y - 4*pixelSize, 8*pixelSize, 8*pixelSize) );
      }
      g2.setColor(Color.DARK_GRAY);
      g2.setStroke( new BasicStroke(2*pixelSize) );
      g2.draw(path);
   }
   
   
   /**
    * Applies a coordinate transform to a Graphics2D graphics context.  The upper left corner of 
    * the viewport where the graphics context draws is assumed to be (0,0).  (This method sets
    * the global variables pixelSize and transform.)
    * @param g2 The drawing context whose transform will be set.
    * @param width The width of the viewport where g2 draws.
    * @param height The height of the viewport where g2 draws.
    * @param limitsRequested  Specifies a rectangle that will be visible in the viewport.  Under the transform,
    * the rectangle with corners (limitsRequested[0],limitsRequested[1]) and (limitsRequested[2],limitsRequested[3])
    * will just fit in the viewport.
    * @param preserveAspect if preserveAspect is false, then the limitsRequested rectangle will exactly fill
    * the viewport; if it is true, then the limits will be expanded in one direction, horizontally or
    * vertically, to make the aspect ratio of the displayed rectangle match the aspect ratio of the
    * viewport.  Note that when preserveAspect is false, the units of measure in the horizontal and
    * vertical directions will be different.
    */
   private void applyLimits(Graphics2D g2, int width, int height, double[] limitsRequested, boolean preserveAspect) {
      double[] limits = limitsRequested;
      if (preserveAspect) {
         double displayAspect = Math.abs((double)height / width);
         double requestedAspect = Math.abs(( limits[3] - limits[2] ) / ( limits[1] - limits[0] ));
         if (displayAspect > requestedAspect) {
            double excess = (limits[3] - limits[2]) * (displayAspect/requestedAspect - 1);
            limits = new double[] { limits[0], limits[1], limits[2] - excess/2, limits[3] + excess/2 };
         }
         else if (displayAspect < requestedAspect) {
            double excess = (limits[1] - limits[0]) * (requestedAspect/displayAspect - 1);
            limits = new double[] { limits[0] - excess/2, limits[1] + excess/2, limits[2], limits[3] };
         }
      }
      double pixelWidth = Math.abs(( limits[1] - limits[0] ) / width);
      double pixelHeight = Math.abs(( limits[3] - limits[2] ) / height);
      pixelSize = (float)Math.min(pixelWidth,pixelHeight);
      g2.scale( width / (limits[1]-limits[0]), height / (limits[3]-limits[2]) );
      g2.translate( -limits[0], -limits[2] );
      transform = g2.getTransform();
   }
   
   
   /**
    * Tries to find a point in the points array that is close to the point with
    * pixel coordinates (x,y).  The point (x,y) is transformed into user coordinates
    * by applying the inverse of the transform that was applied when the panel
    * was painted.  If one of the points in the array is within 5 pixels of (x,y),
    * vertically or horizontally, then the index of the last such point in the array
    * is returned.  If none of the points are that close to (x,y), then the return
    * value is -1.
    */
   private int findPoint( int x, int y ) {
      Point2D.Float p = new Point2D.Float(x,y);
      try {
         transform.inverseTransform(p, p);
      }
      catch (NoninvertibleTransformException e) {
      }
      for (int i = points.length-1; i >= 0; i--) {
         if (Math.abs( points[i].x - p.x ) < 5*pixelSize && Math.abs( points[i].y - p.y ) < 5*pixelSize)
            return i;
      }
      return -1;
   }
   
   
   /**
    * A class that implements dragging of control points and endpoint of the Bezier curves.
    * Also changes the cursor when the mouse is near a point that can be dragged.
    */
   private class MouseHandler implements MouseListener, MouseMotionListener {
      private int draggedPoint = -1;  // Index of point currently being dragged; -1 if none.
      public void mousePressed(MouseEvent e) {
         if (draggedPoint != -1)
            return;
         draggedPoint = findPoint(e.getX(),e.getY());
      }
      public void mouseReleased(MouseEvent e) {
         if (draggedPoint == -1)
            return;
         draggedPoint = -1;
      }
      public void mouseDragged(MouseEvent e) {
         if (draggedPoint == -1)
            return;
         int x = Math.min(getWidth(), Math.max(0,e.getX()));
         int y = Math.min(getHeight(), Math.max(0,e.getY()));
         Point2D.Float p = new Point2D.Float(x,y);
         try {
            transform.inverseTransform(p,p);
         }
         catch (NoninvertibleTransformException e1) {
         }
         points[draggedPoint].setLocation(p);
         repaint();
      }
      public void mouseMoved(MouseEvent e) {
         if (findPoint(e.getX(), e.getY()) >= 0)
            setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
         else
            setCursor(Cursor.getDefaultCursor());
      }
      public void mouseClicked(MouseEvent e) {
      }
      public void mouseEntered(MouseEvent e) {
      }
      public void mouseExited(MouseEvent e) {
      }
   }
   
}
