
import glutil.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.net.URL;

import javax.swing.*;
import javax.imageio.ImageIO;
import javax.media.opengl.*;

import com.sun.opengl.util.ImageUtil;
import com.sun.opengl.util.texture.Texture;
import com.sun.opengl.util.texture.TextureIO;

/**
 * This program demonstrates using automatic texture coordinate generation,
 * using either object-linear or eye-linear coord generation.
 */
public class TextureCoordinateGeneration extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Auto-generated Texture Coordinates");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new TextureCoordinateGeneration());
      window.pack();
      window.setVisible(true);
   }
   
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private Camera camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.
   
   private JTextField[] coefficientInput;  // Input coefficients for linear equation.
   private float[] coefficients;           // The coefficients.

   private JComboBox selectEyeOrObject; // Choose betwee object-linear and eye-linear coord generation.
   private JButton defaultsButton;  // Sets coord generation to default values.
   private JButton applyButton;   // Gets the coefficients from the input boxes.
   
   
   /* The following variables are used for selecting/rendering one object.
      Note that the objects that are used come with texture coordinates. */   
   private UVCylinder cylinder = new UVCylinder(1, 2.5, 96, 32, 10);
   private UVCone cone = new UVCone(1, 2.5, 96, 32, 10);
   private UVSphere sphere = new UVSphere(1.25, 96, 48);
   private Cube cube = new Cube(2);
   private UVTorus torus = new UVTorus(1.25, 0.25, 96, 48);
   private Renderable[] objects = { sphere, cylinder, cube, cone, torus};
   private String[] objectNames = { "Sphere", "Cylinder", "Cube", "Cone", "Torus" };
   private JComboBox objectSelect;  // Where the user selects which object to render.
   
   /* There are three choices of texture for the user to choose from. */
   private Texture[] textures = new Texture[3];
   private String[] textureNames = { "Brick Texture", "Metal Texture", "Earth Texture" };
   private String[] textureFiles = { "brick001.jpg", "metal003.gif", "Earth-1024x512.jpg" };
   private JComboBox textureSelect; // Where the user selects which texture to use.
   
   
   /**
    * Set up the JPanel containing a GLJPanel with some controls
    * at the bottom.  Set up a Camera and TrackBall for 3D viewing.
    */
   public TextureCoordinateGeneration() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      JPanel bottom = new JPanel();
      bottom.setLayout(new GridLayout(0,1,3,3));
      bottom.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
      add(bottom, BorderLayout.SOUTH);
      JPanel row1 = new JPanel();
      row1.setLayout(new FlowLayout(FlowLayout.CENTER));
      bottom.add(row1);
      JPanel row2 = new JPanel();
      row2.setLayout(new FlowLayout(FlowLayout.CENTER));
      bottom.add(row2);
      JPanel row3 = new JPanel();
      row3.setLayout(new FlowLayout(FlowLayout.CENTER));
      bottom.add(row3);
      JPanel row4 = new JPanel();
      row4.setLayout(new FlowLayout(FlowLayout.CENTER));
      bottom.add(row4);
      objectSelect = new JComboBox();
      for (String s : objectNames)
         objectSelect.addItem(s);
      objectSelect.setSelectedIndex(4);
      row1.add(Box.createHorizontalStrut(10));
      row1.add(objectSelect);
      objectSelect.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            camera.setView(0,0,30,0,0,0,0,1,0); // reset view to default
            repaint();
         }
      });
      textureSelect = new JComboBox();
      for (String s : textureNames)
         textureSelect.addItem(s);
      row1.add(Box.createHorizontalStrut(10));
      row1.add(textureSelect);
      textureSelect.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            repaint();
         }
      });
      selectEyeOrObject = new JComboBox();
      selectEyeOrObject.addItem("Object Coords");
      selectEyeOrObject.addItem("Eye Coords");
      selectEyeOrObject.addItem("Standard Coords");
      selectEyeOrObject.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            getCoefficients();
         }
      });
      row1.add(Box.createHorizontalStrut(10));
      row1.add(selectEyeOrObject);
      coefficientInput = new JTextField[8];
      coefficients = new float[] { 1, 0, 0, 0, 0, 1, 0, 0 };
      for (int i = 0; i < 8; i++) {
         coefficientInput[i] = new JTextField("" + coefficients[i], 5);
      }
      String[] labels = { "* x + ", "* y + ", "* z +", "* w" };
      row2.add(new JLabel("s = "));
      row3.add(new JLabel("t = "));
      for (int i = 0; i < 4; i++) {
         row2.add(coefficientInput[i]);
         row2.add(new JLabel(labels[i]));
         row3.add(coefficientInput[i+4]);
         row3.add(new JLabel(labels[i]));
      }
      applyButton = new JButton("  Apply  ");
      applyButton.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            getCoefficients();
         }
      });
      row4.add(applyButton);
      defaultsButton = new JButton("Defaults");
      defaultsButton.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            coefficients = new float[] { 1, 0, 0, 0, 0, 1, 0, 0 };
            for (int i = 0; i < 8; i++)
               coefficientInput[i].setText("" + coefficients[i]);
            selectEyeOrObject.setSelectedIndex(0);
         }
      });
      row4.add(defaultsButton);
      camera = new Camera();
      camera.setScale(2);
      new TrackBall(drawable,camera);  
   }
   
   /**
    * Get the coefficients out of the coefficient input boxes,
    * or report an error if the data does not represent float values.
    */
   private void getCoefficients() {
      float[] nums = new float[8];
      for (int i = 0; i < 8; i++) {
         String input = coefficientInput[i].getText().trim();
         try {
            nums[i] = Float.parseFloat(input);
         }
         catch (NumberFormatException e) {
            JOptionPane.showMessageDialog(this, "Illegal value in input " + i
                        + ".\nThe input must be a legal number.");
            coefficientInput[i].selectAll();
            coefficientInput[i].requestFocus();
         }
      }
      coefficients = nums;
      repaint();
   }

   /**
    * Set up lighting, initialize textures, etc.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      float[] dimwhite = { 0.4f, 0.4f, 0.4f };
      gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, dimwhite, 0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      if (gl.isExtensionAvailable("GL_VERSION_1_2"))
         gl.glLightModeli(GL.GL_LIGHT_MODEL_COLOR_CONTROL, GL.GL_SEPARATE_SPECULAR_COLOR);
      gl.glLightModeli(GL.GL_LIGHT_MODEL_LOCAL_VIEWER, GL.GL_TRUE);
      float[] white = { 1, 1, 1, 1 };
      gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE, white, 0);
      gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, white, 0);
      gl.glMateriali(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, 128);
      gl.glClearColor(0,0,0.3f,1);
      try {
         for (int i = 0; i < textureNames.length; i++) {
            URL textureURL;
            textureURL = getClass().getClassLoader().getResource("textures/" + textureFiles[i]);
            BufferedImage img = ImageIO.read(textureURL);
            ImageUtil.flipImageVertically(img);
            textures[i] = TextureIO.newTexture(img, true);
            textures[i].setTexParameteri(GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT);
            textures[i].setTexParameteri(GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT);
         }
      }
      catch (Exception e) {
         System.out.println("Error while loading textures: " + e);
      }
   }

   /**
    * Draw a textured 3D object, with a texture transform.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      camera.apply(gl);
      
      gl.glPushAttrib(GL.GL_TEXTURE_BIT);  // Save texture sttate.
      
      /* Bind and enable the selected texture. */
      
      int texnum = textureSelect.getSelectedIndex();
      textures[texnum].enable();
      textures[texnum].bind();
      
      /* Set up texture coordinate generation, if it's turned on. */
      
      int gennum = selectEyeOrObject.getSelectedIndex();
      if (gennum == 0) { // Enable object-linear texture coordinate generation.
         gl.glTexGenfv(GL.GL_S, GL.GL_OBJECT_PLANE, coefficients, 0);   
         gl.glTexGenfv(GL.GL_T, GL.GL_OBJECT_PLANE, coefficients, 4);
         gl.glTexGeni(GL.GL_S, GL.GL_TEXTURE_GEN_MODE, GL.GL_OBJECT_LINEAR);
         gl.glTexGeni(GL.GL_T, GL.GL_TEXTURE_GEN_MODE, GL.GL_OBJECT_LINEAR);
         gl.glEnable(GL.GL_TEXTURE_GEN_S);
         gl.glEnable(GL.GL_TEXTURE_GEN_T);
      }
      else if (gennum == 1) { // Enable eye-linear texture coordinate generation.
         gl.glPushMatrix();    // Eye plane is tranformed by modelview matrix.
         gl.glLoadIdentity();  // I don't want andy transformation in this example.
         gl.glTexGenfv(GL.GL_S, GL.GL_EYE_PLANE, coefficients, 0);   
         gl.glTexGenfv(GL.GL_T, GL.GL_EYE_PLANE, coefficients, 4);
         gl.glPopMatrix();
         gl.glTexGeni(GL.GL_S, GL.GL_TEXTURE_GEN_MODE, GL.GL_EYE_LINEAR);
         gl.glTexGeni(GL.GL_T, GL.GL_TEXTURE_GEN_MODE, GL.GL_EYE_LINEAR);
         gl.glEnable(GL.GL_TEXTURE_GEN_S);
         gl.glEnable(GL.GL_TEXTURE_GEN_T);
      }
      else {
         /* In the remaining case, gennum = 2, do not use texture coordinate generation.
          * With TexGen turned off, the built-in texture coords of the objects are used,
          * as specified when the objects are rendered.  Note that when tex coord generation
          * is on, tex coords specified with glTexCoord* are ignored.  */
      }
      
      /* Draw the selected object. */

      int selectedObject = objectSelect.getSelectedIndex();
      gl.glRotated(15,3,2,0);
      gl.glRotated(90,-1,0,0);
      if (selectedObject == 1 || selectedObject == 3)
         gl.glTranslated(0,0,-1.25);
      objects[selectedObject].render(gl);
      
      gl.glPopAttrib();  // Restore texture state.
   }
   
   
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
