

import glutil.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.net.URL;

import javax.swing.*;
import javax.imageio.ImageIO;
import javax.media.opengl.*;

import com.sun.opengl.util.ImageUtil;
import com.sun.opengl.util.texture.Texture;
import com.sun.opengl.util.texture.TextureIO;

/**
 * A panel of type TextureDemo displays a textured object,
 * which the user can rotate by dragging the mouse.  The main
 * routine in this class creates a window containing six
 * panels of type Texture Demo, with a variety of different
 * objects and textures.
 */
public class TextureDemo extends GLJPanel implements GLEventListener {

   public static void main(String[] args){
      JFrame window = new JFrame("Textured Objects");
      JPanel content = new JPanel();
      window.setContentPane(content);
      content.setLayout(new GridLayout(2,3,3,3));
      content.setBackground(Color.LIGHT_GRAY);
      content.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY, 3));
      content.add( new TextureDemo(new UVSphere(1, 40, 20), "textures/Earth-1024x512.jpg", 0) );
      content.add( new TextureDemo(new UVSphere(1, 40, 20), "textures/NightEarth-512x256.jpg", 0) );
      content.add( new TextureDemo(new UVCylinder(0.75,2,20,10), "textures/Earth-1024x512.jpg", -1) );
      content.add( new TextureDemo(new Cube(1.6), "textures/brick001.jpg", 0) );
      content.add( new TextureDemo(new Cube(1.6), "textures/NightEarth-512x256.jpg", 0) );
      content.add( new TextureDemo(new UVCone(0.85,2,20,10), "textures/metal003.gif", -0.75) );
      window.pack();
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setVisible(true);
   }
      
   private Camera camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.
   
   private Renderable object;  // The object displayed in the panel.

   private String textureFileName;  // Name of the file that contains the texture image.

   private double zOffset; // Object is moved by this amount in the z-direction,
                           // then rotated by 90 degrees about the x-axis.
   
   /**
    * Create a TextureDemo panel to show a given object, with a given texture.
    * @param object The object of type Renderable that will be displayed.  Should be sized
    * to fit (after zOffset is applied) in a 2-by-2-by-2 cube centered at the origin.
    * @param textureFileName  the file name of the resource file that contains the texture image.
    * @param zOffset before being rendered, the object is moved in the z-direction by this
    * amount and is then rotated by 90 degrees about the x-axis.  The zOffset can be 
    * supplied to center the object about the origin.  The rotation is because my objects
    * are all oriented along the z-axis; the rotation orients them along the y-axis instead.
    */
   public TextureDemo(Renderable object, String textureFileName, double zOffset) {
      setPreferredSize(new Dimension(250,250));
      this.object = object;
      this.textureFileName = textureFileName;
      this.zOffset = zOffset;
      addGLEventListener(this);
      camera = new Camera();
      camera.setScale(1.5);
      new TrackBall(this,camera);  
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  Here, in addition to the usual setup for 3D,
    * it sets a white material color, suitable for texturing.
    * It also loads, binds and enables the texture. (Since the
    * one texture will be used throughout the program, it
    * makes sense to do this once in the init method.)
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      gl.glColor3f(1,1,1);
      try {
         Texture texture;
         URL textureURL;
         textureURL = getClass().getClassLoader().getResource(textureFileName);
         if (textureURL != null) {
            BufferedImage img = ImageIO.read(textureURL);
            ImageUtil.flipImageVertically(img);
            texture = TextureIO.newTexture(img, true);
            texture.bind();
            texture.enable();
         }
      }
      catch (Exception e) {
         e.printStackTrace();
      }
   }

   /**
    * Display method renders the object.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      camera.apply(gl);
      gl.glRotated(90,-1,0,0);
      gl.glTranslated(0,0,zOffset);
      object.render(gl);
   }
   
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
