
import glutil.*;

import java.awt.*;
import java.awt.event.*;
import java.nio.ByteBuffer;

import javax.swing.*;
import javax.media.opengl.*;

import com.sun.opengl.util.BufferUtil;

/**
 * This program demonstrates using one and two-dimensional textures
 * created directly from color data computed by the program and 
 * stored as a sequence of red, green, and blue values.  The 1D 
 * texture that is used is a spectrum.  The 2D texture is a
 * Mandelbrot set.  (The textures can also be animated.)
 */
public class TextureLoading extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Texture1D");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new TextureLoading());
      window.pack();
      window.setVisible(true);
   }
   
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private Camera camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.

   private int frameNumber;    // Drives the animation.  A timer adds 1
                               // to this value every 30 milliseconds
                               // and repaints the panel.
   
   private JCheckBox animate;  // Controls whether the animation is running.
   
   private Timer animator;  // The timer that drives the animation,
                            // which is started and stopped in response
                            // to the animate checkbox.
   
   /* The following variables are used for selecting/rendering one object.
      Note that the objects that are used come with texture coordinates. */   
   private UVCylinder cylinder = new UVCylinder(1, 2.5, 96, 32, 10);
   private UVCone cone = new UVCone(1, 2.5, 96, 32, 10);
   private UVSphere sphere = new UVSphere(1.25, 96, 48);
   private Cube cube = new Cube(2);
   private UVTorus torus = new UVTorus(1.25, 0.25, 96, 48);
   private Renderable[] objects = { sphere, cylinder, cube, cone, torus};
   private String[] objectNames = { "Sphere", "Cylinder", "Cube", "Cone", "Torus" };
   private JComboBox objectSelect;  // Where the user selects which object to render.
   private JComboBox textureSelect;  // Selects between 1D and 2D textures.
   
   private JComboBox transformSelect; // Where the user selects which texture transform to use.
      
   
   /**
    * Set up the JPanel containing a GLJPanel with some controls
    * at the bottom.  Set up a Camera and TrackBall for 3D viewing.
    */
   public TextureLoading() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      animate = new JCheckBox("Animate");
      animate.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            if (animate.isSelected())
               animator.start();
            else
               animator.stop();
         }
      });
      JPanel bottom = new JPanel();
      bottom.setLayout(new FlowLayout(FlowLayout.CENTER));
      bottom.add(animate);
      add(bottom, BorderLayout.SOUTH);
      animator = new Timer(30, new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            frameNumber++;
            drawable.repaint();
         }
      });
      objectSelect = new JComboBox();
      for (String s : objectNames)
         objectSelect.addItem(s);
      objectSelect.setSelectedIndex(2);
      bottom.add(Box.createHorizontalStrut(10));
      bottom.add(objectSelect);
      objectSelect.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            camera.setView(0,0,30,0,0,0,0,1,0); // reset view to default
            repaint();
         }
      });
      textureSelect = new JComboBox();
      textureSelect.addItem("Texture1D");
      textureSelect.addItem("Texture2D");
      bottom.add(Box.createHorizontalStrut(10));
      bottom.add(textureSelect);
      textureSelect.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            frameNumber = 0;
            repaint();
         }
      });
      transformSelect = new JComboBox();
      transformSelect.addItem("Rotate");
      transformSelect.addItem("Scale");
      transformSelect.addItem("Translate");
      bottom.add(Box.createHorizontalStrut(10));
      bottom.add(transformSelect);
      transformSelect.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            frameNumber = 0;
            repaint();
         }
      });
      camera = new Camera();
      camera.setScale(2);
      new TrackBall(drawable,camera);  
   }

   /**
    * Set up lighting, initialize textures, etc.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      float[] dimwhite = { 0.4f, 0.4f, 0.4f };
      gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, dimwhite, 0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      if (gl.isExtensionAvailable("GL_VERSION_1_2"))
         gl.glLightModeli(GL.GL_LIGHT_MODEL_COLOR_CONTROL, GL.GL_SEPARATE_SPECULAR_COLOR);
      gl.glLightModeli(GL.GL_LIGHT_MODEL_LOCAL_VIEWER, GL.GL_TRUE);
      float[] white = { 1, 1, 1, 1 };
      gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE, white, 0);
      gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, white, 0);
      gl.glMateriali(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, 128);
      gl.glClearColor(0,0,0.3f,1);
      
      /* Make a 1D texture from 256 RGB values representing a spectrum.
       * The color data is first stored in a ByteBuffer, using three bytes
       * per color.  The red, blue, and green components are stored separately
       * as bytes for each color.  When the color data in the buffer is passed to 
       * glTexImage1D, it takes bytes 3 at a time and uses them to create the
       * colors in the texture.  Note that the data is treated as unsigned bytes
       * in the range 0 to 255, even though Java doesn't really have such a type.
       */
      
      ByteBuffer textureData1D = BufferUtil.newByteBuffer(3*256);
      for (int i = 0; i < 256; i++) {
         Color c = Color.getHSBColor(1.0f/256 * i, 1, 1); // A color of the spectrum.
         textureData1D.put((byte)c.getRed());  
         textureData1D.put((byte)c.getGreen()); 
         textureData1D.put((byte)c.getBlue()); 
      }
      textureData1D.rewind();
      gl.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_RGB, 256, 0,
               GL.GL_RGB, GL.GL_UNSIGNED_BYTE, textureData1D);
      gl.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT);
      
      /*
       * Make a 2D texture by computing the individual pixel colors.  The
       * texture shows the famous Mandelbrot set, which is defined by a
       * computational procedure that need'nt concern us here.
       */
      ByteBuffer textureData2D = BufferUtil.newByteBuffer(3*256*512);
      for (int row = 0; row < 256; row++) {
         double cy = -1.25 + 2.5/256*row;
         for (int col = 0; col < 512; col++) {
            double cx = -2.5 + 4.0/512*col;
            int count = 0;
            double x = cx;
            double y = cy;
            while (count < 50 && x*x + y*y < 16) {
               double newx = x*x - y*y + cx;
               y = 2*x*y + cy;
               x = newx;
               count++;
            }
            Color c;
            if (count == 50)
               c = Color.BLACK;
            else
               c = Color.getHSBColor(count/50.0f, 1, 0.8f);
            int rgb = c.getRGB();  // The color components packed into a 32-bit int.
            textureData2D.put((byte)(rgb>>16)); // Red component of c.
            textureData2D.put((byte)(rgb>>8));  // Green component of c.
            textureData2D.put((byte)(rgb));       // Blue component of c.
         }
      }
      textureData2D.rewind();
      gl.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGB, 512, 256, 0,
            GL.GL_RGB, GL.GL_UNSIGNED_BYTE, textureData2D);
      gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT);
      gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT);
      
      /* Note: since I am not providing mipmaps for the texture, the following
       * two commands are required.  The default MIN filter requires
       * mipmaps, and if the mipmaps aren't there, you don't get any texturing
       * at all! 
       */
      
      gl.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
      gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
   }

   /**
    * Draw a textured 3D object, with a texture transform.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      camera.apply(gl);
      
      /* Apply the texture transform, using the texture matrix.
       * The transform that is applied depends on which transform
       * type the user has selected.  It also depends on the current
       * frame number.
       */
      
      gl.glMatrixMode(GL.GL_TEXTURE);
      gl.glLoadIdentity();
      switch (transformSelect.getSelectedIndex()) {
      case 0:
         gl.glTranslated(0.5,0.5,0);
         gl.glRotated(frameNumber,0,0,1);
         gl.glTranslated(-0.5,-0.5,0);
         break;
      case 1: 
         int scaleFrame = frameNumber % 200;
         double scale = 1 + ((scaleFrame < 100)? scaleFrame/100.0 : (200-scaleFrame)/100.0);
         gl.glTranslated(0.5,0.5,0);
         gl.glScaled(scale,scale,scale);
         gl.glTranslated(-0.5,-0.5,0);
         break;
      default:
         gl.glTranslated(frameNumber/200.0,frameNumber/313.0,0);
      }
      gl.glMatrixMode(GL.GL_MODELVIEW);
      
      /* Draw the selected object. */
      
      if (textureSelect.getSelectedIndex() == 0)
         gl.glEnable(GL.GL_TEXTURE_1D); // Use 1D spectrum texture.
      else
         gl.glEnable(GL.GL_TEXTURE_2D); // Use 2D Mandelbrot texture.

      int selectedObject = objectSelect.getSelectedIndex();
      gl.glRotated(15,3,2,0);
      gl.glRotated(90,-1,0,0);
      if (selectedObject == 1 || selectedObject == 3)
         gl.glTranslated(0,0,-1.25);
      objects[selectedObject].render(gl);
      
      gl.glDisable(GL.GL_TEXTURE_1D);
      gl.glDisable(GL.GL_TEXTURE_2D);
   }
   
   
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
