
import glutil.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.net.URL;
import java.nio.FloatBuffer;

import javax.swing.*;
import javax.imageio.ImageIO;
import javax.media.opengl.*;

import com.sun.opengl.util.BufferUtil;
import com.sun.opengl.util.ImageUtil;
import com.sun.opengl.util.texture.Texture;
import com.sun.opengl.util.texture.TextureIO;


/**
 * This program draws a cylinder inside a sphere, where the
 * sphere is represented as a random cloud of points.  It uses
 * vertex arrays to draw the objects, not glBegin/glEnd.  If
 * the OpenGL version number is at least 1.5, it uses
 * vertex buffer objects.
 */
public class VertexArrayDemo extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("VertexArrayDemo");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new VertexArrayDemo());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private Camera camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.

   private int frameNumber;    // Drives the animation.  A timer adds 1
                               // to this value every 30 milliseconds
                               // and repaints the panel.
   
   private JCheckBox animate;  // Controls whether the animation is running.
   
   private Timer animator;  // The timer that drives the animation,
                            // which is started and stopped in response
                            // to the animate checkbox.
   
   private boolean version_1_5;  // Is the OpenGL version at least 1.5? 
                                 // The value is set in the init() method.
   
   /**
    * Constructor used in the applet version, where I want fewer points,
    * since the viewport is smaller.
    */
   public VertexArrayDemo(int spherePoints) {
      this();
      spherePointCount = spherePoints;
   }
   
   /**
    * Constructor creates the panel, with an "Animate" button to control
    * whether the animation is running.
    */
   public VertexArrayDemo() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      animate = new JCheckBox("Animate");
      animate.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            if (animate.isSelected())
               animator.start();
            else
               animator.stop();
         }
      });
      JPanel bottom = new JPanel();
      bottom.setLayout(new FlowLayout(FlowLayout.CENTER));
      bottom.add(animate);
      add(bottom, BorderLayout.SOUTH);
      animator = new Timer(30, new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            frameNumber++;
            drawable.repaint();
         }
      });      
      camera = new Camera();
      camera.setScale(1.4);
      new TrackBall(drawable,camera);  
   }
   
   //---------------------- sphere as a point cloud ---------------------
   
   // "Vertex arrays" are actually defined as "Buffers", belonging to the
   // class java.nio.Buffer.  Here, FloatBuffers are used.
   
   private int spherePointCount = 100000;

   private FloatBuffer sphereVertices;  // Random points on a unit sphere.
   private FloatBuffer sphereTexCoords;   // Texture coordinates for each point.
   
   // For using Vertex Buffer Objects, we need ID numbers.  These are generated
   // in the init() method and are used only if the OpenGL version is 1.5 or higher.
   
   private int spherePointBufferID;  // VBO ID for spherePointCloud.
   private int sphereTexCoordID;     // VBO ID for sphereTexCoords.
   
   private Texture sphereTexture;  // A texture object for texturing the sphere.

   /**
    * Generate random points on a unit sphere.  Store them in spherePointCloud,
    * and store corresponding texture coordinates in sphereTexCoords.
    */
   private void createPointSphere() {
      sphereVertices = BufferUtil.newFloatBuffer(spherePointCount*3);
      sphereTexCoords = BufferUtil.newFloatBuffer(spherePointCount*2);
      for (int i = 0; i < spherePointCount; i++) {
         double s = Math.random();
         double t = Math.random();
         sphereTexCoords.put(2*i,(float)s);
         sphereTexCoords.put(2*i+1,(float)t);
         double u = s * Math.PI * 2;
         double z = t * 2 - 1;
         double r = Math.sqrt(1-z*z);
         double x = r * Math.cos(u);
         double y = r * Math.sin(u);
         sphereVertices.put(3*i,(float)x);
         sphereVertices.put(3*i+1,(float)y);
         sphereVertices.put(3*i+2,(float)z);
      }
   }
   
   /**
    * Draw pointCt points from the sphere data buffers  Number
    * of points must be less than or equal to spherePointCount.
    * NOTE that for a point on the unit sphere, the coordinates of
    * the point are the same as the coordinates of the unit normal
    * vector to that point.  So, we can use the same data for
    * the points as for the normals.
    */
   private void drawPointSphere(GL gl, int pointCt) {
      if (version_1_5) {
         // Use glBindBuffer to say what buffer to work on, then use
         // glVertexPointer to set the position where the data starts
         // in the buffer.  In this case, the data starts at position 0.
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, spherePointBufferID);
         gl.glVertexPointer(3,GL.GL_FLOAT,0,0);
         gl.glNormalPointer(GL.GL_FLOAT,0,0);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereTexCoordID);
         gl.glTexCoordPointer(2, GL.GL_FLOAT, 0, 0);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER,0); // Leave no VBO bound.
      }
      else {
         // Use glVertexPointer to set the buffer from which the
         // data will be read.  The internal position pointer of
         // the buffer indicates where the data starts in the buffer.
         // For the sphere, that internal position pointer is never
         // moved from 0, so the data starts at the start of the buffer.
         gl.glVertexPointer(3,GL.GL_FLOAT,0,sphereVertices);
         gl.glNormalPointer(GL.GL_FLOAT,0,sphereVertices);
         gl.glTexCoordPointer(2, GL.GL_FLOAT, 0, sphereTexCoords);
      }
      gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
      gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
      gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
      sphereTexture.enable();
      gl.glDrawArrays(GL.GL_POINTS, 0, pointCt);  // Generate the actual primitives!!
      gl.glDisableClientState(GL.GL_VERTEX_ARRAY);
      gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
      gl.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY);
      sphereTexture.disable();
   }
   
   //------------- cylinder as a triangle strip and two triangle fans --------------
   
   // The data for the side, top, and bottom of the cylinder are all stored in
   // the same buffer, cylinderPoints.  The side data starts at position 0,
   // followed by the vertices for the top and then the vertices for the bottom.
   // There are (cylinderVertexCount + 1) vertices around the top and around 
   // the bottom of the cylinder.  For drawing the top and bottom, an extra
   // vertex is added at the center to serve as the starting point for the
   // triangle fan.  There are (cylinderVertexCount+1)*2 vertices for the side
   // and (cylinderVertexCount+2) vertices for the top and for the bottom.
   // The variables cylinderTopStart and cylinderBottomStart give the
   // positions for the start of the data from the top and bottom, in 
   // terms of number of vertices.  Normals are stored only for the side,
   // since for the top and for the bottom, the same normal is used for each vertex.
   
   private int cylinderVertexCount = 32;
   
   private int cylinderTopStart = (cylinderVertexCount+1)*2;
   private int cylinderBottomStart = cylinderTopStart + (cylinderVertexCount+2);
   
   private FloatBuffer cylinderPoints;      // Vertices for side, top, bottom.
   private FloatBuffer cylinderSideNormals; // Normals for side only.
   
   private int cylinderPointBufferID;  // VBO buffer ID for cylinderPoints.
   private int cylinderNormalBufferID; // VBO buffer ID for cylinderSideNormals.
   
   /**
    * Create the data for the cylinder, using the specified radius and height.
    */
   private void createCylinder(double radius, double height) {

      cylinderPoints = BufferUtil.newFloatBuffer( ((cylinderVertexCount+1)*4 + 2) * 3 );
      cylinderSideNormals = BufferUtil.newFloatBuffer( ((cylinderVertexCount+1)*2) * 3 );
      
      // Data for the side of the cylinder.
      
      for (int i = 0; i <= cylinderVertexCount; i++) {
         double angle = (2*Math.PI/cylinderVertexCount) * i;
         double x = Math.cos(angle);
         double y = Math.sin(angle);
         
         cylinderSideNormals.put((float)x);      // Generate a point at top of strip.
         cylinderSideNormals.put((float)y);
         cylinderSideNormals.put(0);
         cylinderPoints.put((float)(radius*x));
         cylinderPoints.put((float)(radius*y));
         cylinderPoints.put((float)height);
         
         cylinderSideNormals.put((float)x);     // Generate a point at bottom of strip.
         cylinderSideNormals.put((float)y);
         cylinderSideNormals.put(0);
         cylinderPoints.put((float)(radius*x));
         cylinderPoints.put((float)(radius*y));
         cylinderPoints.put(0);
      }
      
      // Data for the top of the cylinder.
      
      cylinderPoints.put(0);    // Center vertex for top of cylinder.
      cylinderPoints.put(0);
      cylinderPoints.put((float)height);
      for (int i = 0; i <= cylinderVertexCount; i++) {  // Vertices around the top.
         double angle = (2*Math.PI/cylinderVertexCount) * i;
         double x = Math.cos(angle);
         double y = Math.sin(angle);
         cylinderPoints.put((float)(radius*x));
         cylinderPoints.put((float)(radius*y));
         cylinderPoints.put((float)height);
      }
      
      // Data for the bottom of the cylinder.  (Note that the vertices
      // are generated in the opposite order from the top, since the top
      // and bottom face in opposite directions.)
      
      cylinderPoints.put(0);  // Center vertex for bottom of cylinder.
      cylinderPoints.put(0);
      cylinderPoints.put(0);
      for (int i = cylinderVertexCount; i >= 0; i--) {  // Vertices around the bottom.
         double angle = (2*Math.PI/cylinderVertexCount) * i;
         double x = Math.cos(angle);
         double y = Math.sin(angle);
         cylinderPoints.put((float)(x*radius));
         cylinderPoints.put((float)(y*radius));
         cylinderPoints.put(0);
      }
      
      // Rewind the FloatBuffers.  (This is only necessary because I used the
      // relative "put" methods to drop data into the buffers.  For the
      // sphere, using the indexed put method, the position pointer in the
      // buffer never moved from zero.)
      
      cylinderPoints.rewind();  // Restore position pointer to start.
      cylinderSideNormals.rewind();  // Restore position pointer to start.
   }
   
   /**
    * Draw the cylinder.  Side, top, and bottom are drawn separately.
    */
   private void drawCylinder(GL gl) {
      
      // Set up the vertex and normal pointers
      
      if (version_1_5) {
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, cylinderPointBufferID);
         gl.glVertexPointer(3,GL.GL_FLOAT,0,0);  // Data starts at position 0.
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, cylinderNormalBufferID);
         gl.glNormalPointer(GL.GL_FLOAT,0,0);  // Data starts at position 0.
      }
      else {
         gl.glVertexPointer(3,GL.GL_FLOAT,0,cylinderPoints);
         gl.glNormalPointer(GL.GL_FLOAT,0,cylinderSideNormals);
      }

      // Draw the side, using data from the vertex buffer and from the normal buffer.
      
      gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
      gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
      gl.glDrawArrays(GL.GL_QUAD_STRIP, 0, (cylinderVertexCount+1)*2);
      gl.glDisableClientState(GL.GL_NORMAL_ARRAY);  // Turn off normal array.
                                                    // Leave vertex array enabled.

      // Draw the top and bottom, using data from vertex buffer only.
      
      gl.glNormal3f(0,0,1); // Normal for all vertices for the top.
      gl.glDrawArrays(GL.GL_TRIANGLE_FAN, cylinderTopStart, cylinderVertexCount+2);
      gl.glNormal3f(0,0,-1); // Normal for all vertices for the top.
      gl.glDrawArrays(GL.GL_TRIANGLE_FAN, cylinderBottomStart, cylinderVertexCount+2);
            
      gl.glDisableClientState(GL.GL_VERTEX_ARRAY);  // Turn off vertex array.
      
   }

   //------------------------------------------------------------------------------
   

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    * It also initializes the data for the sphere and cylinder.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      version_1_5 = gl.isExtensionAvailable("GL_VERSION_1_5");
      System.out.println("OpenGL Version 1.5 available: " + version_1_5);

      createPointSphere();
      createCylinder(0.7,1.3);
      
      if (version_1_5) {
         
         // For OpenGL Version 1.5 or higher, vertex buffer objects are
         // used to store the vertex, normal, and texture coordinates.

         int[] bufferID = new int[4];    // Get 4 buffer IDs for the data.
         gl.glGenBuffers(4,bufferID,0);

         spherePointBufferID = bufferID[0];
         sphereTexCoordID = bufferID[1];
         cylinderPointBufferID = bufferID[2];
         cylinderNormalBufferID = bufferID[3];
         
         // Use glBindBuffer to say which buffer object you want to work on,
         // then load the data into the buffer.  Buffer size is specified
         // in bytes.  Since a float value takes up 4 bytes, the number of
         // float values in the buffer is multiplied by 4 to get the n
         // of bytes.  For example, the spherePointCloud buffer contains
         // 3 float data for each of spherePointCount points; the number of
         // float values is spherePointCount*3, and so the number of bytes
         // is spherePointCount*3*4.
         
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, spherePointBufferID);
         gl.glBufferData(GL.GL_ARRAY_BUFFER, spherePointCount*3*4, 
               sphereVertices, GL.GL_STATIC_DRAW);
         
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereTexCoordID);
         gl.glBufferData(GL.GL_ARRAY_BUFFER, spherePointCount*2*4, 
               sphereTexCoords, GL.GL_STATIC_DRAW);
         
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, cylinderPointBufferID);
         gl.glBufferData(GL.GL_ARRAY_BUFFER, ((cylinderVertexCount+1)*4+2)*3*4, 
               cylinderPoints, GL.GL_STATIC_DRAW);
         
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, cylinderNormalBufferID);
         gl.glBufferData(GL.GL_ARRAY_BUFFER, (cylinderVertexCount+1)*2*3*4, 
               cylinderSideNormals, GL.GL_STATIC_DRAW);
      
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);  // Leave no buffer ID bound.
      }
      
      try {
         URL textureURL;
         textureURL = getClass().getClassLoader().getResource("textures/Earth-1024x512.jpg");
         if (textureURL != null) {
            BufferedImage img = ImageIO.read(textureURL);
            ImageUtil.flipImageVertically(img);
            sphereTexture = TextureIO.newTexture(img, true);
            sphereTexture.bind();
         }
      }
      catch (Exception e) {
         System.out.println("Could not load texture image for sphere!");
      }
      
   }

   /**
    * Draws the sphere and the cylinder.  A rotation is applied
    * to the cylinder, depending on the frame number, to make it
    * tumble around inside the sphere.  If animation is in progress,
    * then the number of points shown on the sphere depends on the
    * frameNumber.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      camera.apply(gl);
      
      gl.glPushMatrix();
      gl.glRotatef(90,-1,0,0);      
      int pointCt;
      if (animator.isRunning()) {
         int pointsPerFrame = spherePointCount/500;
         int x = frameNumber % 1000;
         if (x < 500)
            pointCt = pointsPerFrame*(x);
         else
            pointCt = pointsPerFrame*(1000-x);
      }
      else 
         pointCt = spherePointCount;
      gl.glColor3f(1,1,1);
      if (pointCt > 0)
         drawPointSphere(gl, pointCt);
      gl.glPopMatrix();
      
      gl.glPushMatrix();
      gl.glRotatef(frameNumber/3,1,0,0);
      gl.glRotatef(frameNumber/2,0,0,1);
      gl.glColor3d(0.3,0,0);
      gl.glRotatef(90,-1,0,0);
      gl.glTranslated(0,0,-0.65);
      drawCylinder(gl);
      gl.glPopMatrix();
   }
      
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
