
import glutil.*;
import simplescenegraph3d.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;

/**
 * Creates a "world" consisting of a 41-by-41 plane with various shapes
 * scattered randomly on it.  The user has the experience of walking
 * around, by using the up/down arrow keys to move forward and backward
 * and the left/right arrow keys to turn.
 */
public class WalkThroughDemo extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Walk-through Demo");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new WalkThroughDemo());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   
   private SceneNode3D theWorld;  // Contains the plane and all the shapes.

   private UVSphere sphere = new UVSphere(48, 32);     // Some shapes to put in the world.
   private Cube cube = new Cube();
   private UVCone cone = new UVCone(0.5,2);
   private UVCylinder cylinder = new UVCylinder(0.5,2);
   
   private SimpleAvatar viewer;  // Represents the viewpoint of the viewer.

   /**
    * Create the world and the viewer's "avatar".  Set up keyboard, focus, and
    * mouse handling.
    */
   public WalkThroughDemo() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,350));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      final JLabel message = new JLabel("CLICK TO GET FOCUS", JLabel.CENTER);
       message.setOpaque(true);
       message.setBackground(Color.LIGHT_GRAY);
       message.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
       add(message,BorderLayout.SOUTH);
      viewer = new SimpleAvatar();
      buildWorld();
      setBorder(BorderFactory.createLineBorder(Color.GRAY,4));
      addMouseListener(new MouseAdapter() {
         public void mousePressed(MouseEvent evt) {
            requestFocus();
         }
      });
      addFocusListener(new FocusListener() {
         public void focusLost(FocusEvent e) {
            setBorder(BorderFactory.createLineBorder(Color.GRAY,4));
            message.setText("CLICK TO GET FOCUS");
         }
         public void focusGained(FocusEvent e) {
            setBorder(BorderFactory.createLineBorder(Color.CYAN,4));
            message.setText("Arrow Keys = Move; R = reset world; HOME = reset view");
         }
      });
      addKeyListener(new KeyAdapter() {
         public void keyPressed(KeyEvent e) {
            int code = e.getKeyCode();
            if (code == KeyEvent.VK_UP)
               viewer.move(0.05);
            else if (code == KeyEvent.VK_DOWN)
               viewer.move(-0.05);
            else if (code == KeyEvent.VK_LEFT)
               viewer.turn(1);
            else if (code == KeyEvent.VK_RIGHT)
               viewer.turn(-1);
            else if (code == KeyEvent.VK_HOME)
               viewer.home();
            else if (code == KeyEvent.VK_R) {
               buildWorld();
               viewer.home();
            }
            drawable.repaint();
         }
      });
   }
   
   /**
    * Create a SceneNode3D to represent the contents of the world.  The
    * plane is a scaled cube that has been translated in the negative y
    * direction so that its top is at y = -1.  The other objects have
    * their centers on the xz-plane, which is the same plane in which the
    * viewer moves, so the centers of the objects are at eye height.
    * The viewer starts out at (0,0), so we avoid placing objects within
    * 2 units of that position.  Otherwise, each position on a 41-by-41
    * grid has a 2% chance of being occupied. All the objects are scaled to
    * just fit within a square of side 1.  Cubes, cones, and cylinders are
    * scaled so that they sit on the plane.  Spheres have radius 0.5 and
    * float above the plane.
    */
   private void buildWorld() {
      ComplexObject3D world = new ComplexObject3D();
      world.add( new RenderableNode(cube).setScaling(21, 1, 21)
            .setColor(Color.WHITE)
            .setTranslation(0,-2,0));
      for (int i = -20; i <= 20; i++)
         for (int j = -20; j <= 20; j++) {
            if (Math.random() < 0.02 && (Math.abs(i) > 2 || Math.abs(j) > 2)) {
               SceneNode3D r;
               switch ((int)(4*Math.random())) {
               case 0:
                  r = new RenderableNode(cube).setScaling(0.5,1,0.5);
                  break;
               case 1:
                  r = new RenderableNode(sphere).setScaling(0.5);
                  break;
               case 2:
                  r = new RenderableNode(cone).setRotation(90,-1,0,0).setTranslation(0,-1,0);
                  break;
               default:
                  r = new RenderableNode(cylinder).setRotation(90,-1,0,0).setTranslation(0,-1,0);
               }
               Color c = Color.getHSBColor((float)Math.random(), (float)(0.5+0.5*Math.random()), 0.8f);
               world.add(new ComplexObject3D(r).setColor(c).setTranslation(i,0,j));
            }
         }
      theWorld = world;
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0.3f,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      gl.glEnable(GL.GL_NORMALIZE);
   }

   /**
    * Draw a the world, from the point of view of the viewer's avatar.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

      viewer.apply(gl);  // Set up projection and viewing transforms.
      
      theWorld.draw(gl);
   }
      
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
