package glsl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import javax.media.opengl.GL;


/**
 * A GLSLProgram object represents an OpenGL Shading Language program
 * that include a vertex shader, a fragment shader, or both.  This class
 * incorporates most of the OpenGL 2.0 methods for dealing with GLSL
 * programs.  It requires OpenGL 2.0 or higher to work.
 *    To use a GLSLProgram, you must add at least one string of source
 * code for either the fragment shader or the vertex shader (or both),
 * using the methods addFragmentShaderCode() and addVertexShaderCode().
 * The program can then be installed in a GL context, using the
 * install() method.  You can optionally compile the shaders and link
 * them before installing; if this has not already been done, it will 
 * be done automatically by the install() method.  The compile and
 * link methods make it possible to obtain the infologs from the 
 * compilation and linking, which can be useful during debugging.
 *    Note that it is assumed that the same GL context (for the same
 * GLDrawable) is used in all method calls.  This class is not meant
 * for sharing programs among several OpenGL contexts.
 */
public class GLSLProgram {
   
   private ArrayList<String> vertexProgram;
   private ArrayList<String> fragmentProgram;
   
   private int vertexShaderID = -1;
   private int fragmentShaderID = -1;
   private int programID = -1;
   
   private boolean vertexShaderCompiled;
   private boolean fragmentShaderCompiled;
   private boolean programLinked;
   
   private Map<String, Integer> attributeBindings;
   
   /**
    * Returns the OpenGL program ID number that is currently being used to
    * represent the GLSL program.  This method should be called only after
    * the program has been linked, either by calling the linkProgram() method
    * or automatically when the program is installed with the install() method.
    * Note that re-linking a program will change the program ID!
    * @throws IllegalStateException if called before the program is linked
    */
   public int getProgramID() { // return value can change after re-linking!
      if (!programLinked)
         throw new IllegalStateException("Program ID is only available after program linking.");
      return programID;
   }
   
   /**
    * Can be called after the program has been successfully linked to find the location
    * of a uniform variable in the program.  The location can be used in the
    * OpenGL method glUniform* to assign a value to the uniform variable.
    * @param gl The OpenGL context in which the program has been linked.
    * @param uniformName The name of the uniform variable, as used in the shader program.
    * @return The location of the uniform variable, or -1 if the location can't be found.
    *    Note that only "active" uniform variables have a location.  To have an active
    *    uniform, it is not enought to declare it in the GLSL program; it must actually
    *    be used in the program.
    * @throws IllegalStateException if called before the program is linked
    */
   public int getUniformLocation(GL gl, String uniformName) {  // return value can change after re-linking!
      if (!programLinked)
         throw new IllegalStateException("Uniform locations are only available after program linking.");
      return gl.glGetUniformLocation(programID, uniformName);
   }
   
   /**
    * Sets the attribute index number that is to be associated with a given attribute
    * variable in the GLSL program.  This method can be called at any time and does
    * not require an OpenGL context as a parameter.  The assignment is not actually
    * made when the method is called.  Attribute name/value pairs are stored and are
    * applied when the GLSL program has be successfully linked.  Calling this method
    * after the program has been linked will force the program to be re-linked the
    * next time it is installed in an OpenGL context.
    * @param attributeVariableName The name of the attribute variable, as declared
    *    in the GLSL vertex shader program.
    * @param attributeIndex  The attribute number that should be associated with the
    *    variable.  The index should be greater than zero, since zero is reserved for
    *    the vertex location.  The index that is assigned here can be used in the
    *    OpenGL method gl.glVertexAttrib* to set the value of the attribute variable.
    *    (It can also be used with glVertexAttribPointer when using vertex arrays.)
    */
   public void bindAttributeLocation(String attributeVariableName, int attributeIndex) {
      if (attributeBindings == null)
         attributeBindings = new HashMap<String, Integer>();
      attributeBindings.put(attributeVariableName, attributeIndex);
      programLinked = false;
   }
   
   /**
    * Adds a source code string to the vertex shader program.  Although multiple
    * source code strings are allowed, it is common to use just one.  You do NOT
    * need a separate string for each line in the program.  You can use one big
    * string with source code lines separated by newline characters.  Note that
    * calling this method after the vertex shader has been compiled will force it
    * to be recompiled, and the GLSL program to be re-linked, the next time
    * the GLSL program is installed in an OpenGL context.
    */
   public void addVertexShaderCode(String sourceString) {
      if (sourceString == null)
         throw new IllegalArgumentException("Source string cannot be null.");
      if (vertexProgram == null)
         vertexProgram = new ArrayList<String>();
      vertexProgram.add(sourceString);
      vertexShaderCompiled = false;
      programLinked = false;
   }
   
   /**
    * Adds a source code string to the fragment shader program.  Although multiple
    * source code strings are allowed, it is common to use just one.  You do NOT
    * need a separate string for each line in the program.  You can use one big
    * string with source code lines separated by newline characters.  Note that
    * calling this method after the vertex shader has been compiled will force it
    * to be recompiled, and the GLSL program to be re-linked, the next time
    * the GLSL program is installed in an OpenGL context.
    */
   public void addFragmentShaderCode(String sourceString) {
      if (sourceString == null)
         throw new IllegalArgumentException("Source string cannot be null.");
      if (fragmentProgram == null)
         fragmentProgram = new ArrayList<String>();
      fragmentProgram.add(sourceString);
      fragmentShaderCompiled = false;
      programLinked = false;
   }

   /**
    * Compiles the vertex shader program.  The program consists of strings that
    * have been added using the addVertexShaderCode() method.  If an error occurs
    * during compilation, the return value of the method is false.  Calling
    * this method is optional; if it has not been called before the program is
    * linked, then it will be called automatically.
    * @param gl  The OpenGL context that is use to do the compilation.
    * @param infoLog if non-null, the infolog from the compilation is placed
    *   in this StringBuffer.  The infolog contains information about the compilation,
    *   including any error messages that are generated when the compilation fails.
    *   The infolog is essential for debugging during development.
    * @return true if the compilation succeeds, false if not.
    */
   public boolean compileVertexShader(GL gl, StringBuffer infoLog) {
      if (vertexProgram == null)
         return true;
      String[] sourceStrings = new String[vertexProgram.size()];
      int[] sourceStringLengths = new int[vertexProgram.size()];
      for (int i = 0; i < sourceStrings.length; i++) {
         sourceStrings[i] = vertexProgram.get(i);
         sourceStringLengths[i] = sourceStrings[i].length();
      }
      if (vertexShaderID > 0)
         gl.glDeleteShader(vertexShaderID);
      vertexShaderID = gl.glCreateShader(GL.GL_VERTEX_SHADER);
      gl.glShaderSource(vertexShaderID, sourceStrings.length, sourceStrings, sourceStringLengths, 0);
      gl.glCompileShader(vertexShaderID);
      int[] result = new int[1];
      gl.glGetShaderiv(vertexShaderID,GL.GL_COMPILE_STATUS,result,0);
      vertexShaderCompiled = result[0] != 0;
      if (infoLog != null) {
         gl.glGetShaderiv(vertexShaderID,GL.GL_INFO_LOG_LENGTH,result,0);
         if (result[0] > 0) {
            byte[] log = new byte[result[0]];
            gl.glGetShaderInfoLog(vertexShaderID, log.length, result, 0, log, 0);
            for (int i = 0; i < result[0]; i++)
               infoLog.append((char)log[i]);
         }
      }
      return vertexShaderCompiled;
   }
   
   /**
    * Compiles the fragment shader program.  The program consists of strings that
    * have been added using the addFragmentShaderCode() method.  If an error occurs
    * during compilation, the return value of the method is false. Calling
    * this method is optional; if it has not been called before the program is
    * linked, then it will be called automatically.
    * @param gl  The OpenGL context that is use to do the compilation.
    * @param infoLog if non-null, the infolog from the compilation is placed
    *   in this StringBuffer.  The infolog contains information about the compilation,
    *   including any error messages that are generated when the compilation fails.
    *   The infolog is essential for debugging during development.
    * @return true if the link succeeds, false if not.
    */
   public boolean compileFragmentShader(GL gl, StringBuffer infoLog) {
      if (fragmentProgram == null)
         return true;
      String[] sourceStrings = new String[fragmentProgram.size()];
      int[] sourceStringLengths = new int[fragmentProgram.size()];
      for (int i = 0; i < sourceStrings.length; i++) {
         sourceStrings[i] = fragmentProgram.get(i);
         sourceStringLengths[i] = sourceStrings[i].length();
      }
      if (fragmentShaderID > 0)
         gl.glDeleteShader(fragmentShaderID);
      fragmentShaderID = gl.glCreateShader(GL.GL_FRAGMENT_SHADER);
      gl.glShaderSource(fragmentShaderID, sourceStrings.length, sourceStrings, sourceStringLengths, 0);
      gl.glCompileShader(fragmentShaderID);
      int[] result = new int[1];
      gl.glGetShaderiv(fragmentShaderID,GL.GL_COMPILE_STATUS,result,0);
      fragmentShaderCompiled = result[0] != 0;
      if (infoLog != null) {
         gl.glGetShaderiv(fragmentShaderID,GL.GL_INFO_LOG_LENGTH,result,0);
         if (result[0] > 0) {
            byte[] log = new byte[result[0]];
            gl.glGetShaderInfoLog(fragmentShaderID, log.length, result, 0, log, 0);
            for (int i = 0; i < result[0]; i++)
               infoLog.append((char)log[i]);
         }
      }
       return fragmentShaderCompiled;
   }
   
   /**
    * Links the GLSL program.  If an error occurs during compilation, the return 
    * value of the method is false. Calling this method is optional; if it has not 
    * been called before the program is installed, then it will be called automatically.
    * Calling this method will cause the vertex and fragment shaders to be compiled,
    * if they have not already been compiled.  Note that re-linking a program will
    * assigne a new programID to the program!
    * @param gl  The OpenGL context that is use to do the linking.
    * @param infoLog if non-null, the infolog from the linking is placed
    *   in this StringBuffer.   (However, if a compilation is attempted and
    *   fails, then the infolog from the failed compilation is returned instead.) 
    *   The infolog contains information about the linking (or compilation),
    *   including any error messages that are generated.
    *   The infolog is essential for debugging during development.
    * @return true if the compilation succeeds, false if not.    */
   public boolean linkProgram(GL gl, StringBuffer infoLog) {
      programLinked = false;
      if (fragmentProgram == null && vertexProgram == null) {
         if (infoLog != null)
            infoLog.append("(No vertex or fragment shader code has been specified.)\n");
         return false;
      }
      if (vertexProgram != null && !vertexShaderCompiled) {
         if (!compileVertexShader(gl, infoLog)) {
            if (infoLog != null)
               infoLog.append("\n\n(Error occured during vertex shader compilation.)\n");
            return false;
         }
         else if (infoLog != null)
            infoLog.setLength(0); // throw away info log from successful compilation.
      }
      if (fragmentProgram != null && !fragmentShaderCompiled) {
         if (!compileFragmentShader(gl, infoLog)) {
            if (infoLog != null)
               infoLog.append("\n\n(Error occured during fragment shader compilation.)\n");
            return false;
         }
         else if (infoLog != null)
            infoLog.setLength(0); // throw away info log from successful compilation.
      }
      if (programID > 0)
         gl.glDeleteProgram(programID);
      programID = gl.glCreateProgram();
      if (fragmentShaderID > 0)
         gl.glAttachShader(programID, fragmentShaderID);
      if (vertexShaderID > 0)
         gl.glAttachShader(programID, vertexShaderID);
      gl.glLinkProgram(programID);
      int[] result = new int[1];
      gl.glGetProgramiv(programID,GL.GL_LINK_STATUS,result,0);
      programLinked = result[0] != 0;
      if (infoLog != null) {
         gl.glGetProgramiv(programID,GL.GL_INFO_LOG_LENGTH,result,0);
         if (result[0] > 0) {
            byte[] log = new byte[result[0]];
            gl.glGetProgramInfoLog(programID, log.length, result, 0, log, 0);
            for (int i = 0; i < result[0]; i++)
               infoLog.append((char)log[i]);
         }
      }
      if (programLinked && attributeBindings != null) {
         for (Map.Entry<String,Integer> pair : attributeBindings.entrySet())
            gl.glBindAttribLocation(programID, pair.getValue(), pair.getKey());
      }
      return programLinked;
   }
   
   /**
    * Installs the GLSL program in an OpenGL context.  If the program needs to
    * be linked, that is done automatically.  The return value tells whether
    * the GLSL program has been successfully installed.  If installation
    * is successful, then the program will be used for rendering in the
    * OpenGL context, until a different program is installed or until
    * gl.glUseProgram(0) is called to return to standard rendering.
    */
   public boolean install(GL gl) {
      if (! programLinked) {
         if (! linkProgram(gl,null) )
            return false;
      }
      while (gl.glGetError() != GL.GL_NO_ERROR) {
         // Clear any previous error state!
      }
      gl.glUseProgram(programID);
      return gl.glGetError() == GL.GL_NO_ERROR;
   }
   
   /**
    * Deletes the GLSLprogram object and its shaders from the OpenGL context.
    * If the program is later installed in an OpenGL context it will be compiled 
    * and linked again.
    */
   public void dispose(GL gl) {
      if (programID >= 0)
         gl.glDeleteProgram(programID);
      if (vertexShaderID >= 0)
         gl.glDeleteShader(vertexShaderID);
      if (fragmentShaderID >= 0)
         gl.glDeleteShader(fragmentShaderID);
      programID = fragmentShaderID = vertexShaderID = -1;
      programLinked = fragmentShaderCompiled = vertexShaderCompiled = false;
   }
   
}
