package glsl;

import glutil.*;

import java.awt.*;
import java.awt.event.*;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;

import javax.swing.*;
import javax.media.opengl.*;

import com.sun.opengl.util.BufferUtil;


/**
 * This is a version of the IcosphereIFS sample program that uses
 * a GLSL program with a vertex shader and a fragment shader.
 * The program is represented by an object of type GLSLProgram.
 * One goal of this example was to demostrate three kinds of
 * variables that can be used in a GLSL shader:  attribute 
 * variables, uniform variables, and varying variables.  The
 * specific program that is used modifies vertex colors based
 * on the recursion level at which the vertex was generated
 * during the computation of the data for the icosphere.  (It 
 * turned out not to be as interesting as I had hoped.)
 * 
 * This program draws an "icosphere", that is an an approximation
 * for a sphere that is obtained by starting with an icosahedron
 * and possibly subdividing the triangles of the icosahedron one or more
 * times.  The user can select the number of subdivisions, up to 5.
 * The program uses vertex buffer arrays and glDrawElements.
 */
public class IcosphereIFS_GLSL extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Icosphere");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new IcosphereIFS_GLSL());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private Camera camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.
   
   private boolean version_1_5;  // Is the OpenGL version at least 1.5? 
    // The value is set in the init() method.

   private boolean version_2_0;  // Is the OpenGL version at least 2.0? 
    // The value is set in the init() method.

   private JComboBox levelSelect;  // For selecting the number of subdivisions.
   private JCheckBox showEdges;    // Should edges of triangles be drawn?
   
   private static float[][] icoshedronVertices;  // Vertices of an icosahedron.
   private static int[][] icoshedronFaces;  // Faces of the icosahedron, with a face
                                            // represented as 3 indices into the vertex array.
   
   private FloatBuffer icosphereVertexBuffer; // Buffer to hold icosphere vertex data.
   private IntBuffer icosphereIndexBuffer;    // Buffer to hold icosphere face data, with
                                              //    a face represented as 3 indices into
                                              //    the vertex buffer.
   private FloatBuffer recLevelAttribBuffer;  // Buffer to hold attribute values for each
                                              //    vertex.  The attribute associated with
                                              //    the vertex is the recursion level at
                                              //    which the vertex is generated.
   private int vertexCount, indexCount;  // Number of float values stored in the buffers.
                                         // (Except 1/3 this amount in the attribute buffer.)
   
   private int icosphereVertexID;  // For OpenGL 1.5+, a VBO ID for the vertex data.
   private int icopsphereIndexID;  // For OpenGL 1.5+, a VBO ID for the face data.
   private int recLevelAttribID;  // For OpenGL 1.5+, a VBO ID for the vertex attribute data.
   
   private boolean dataValid = false;  // Set to false when buffer data has to be computed.
                                       //   This is checked in the display() method.
   
   private GLSLProgram glslProgram;   // The GLSL program.
   private boolean programOK;         // Set when the program is created, if creation succeeds.
   
   private JCheckBox glslCheckbox;    // Used to turn the GLSL program on and off.
   
   static{  // Initialize the data for the icosahedron.
      float t = (float)((Math.sqrt(5) - 1)/2);
      icoshedronVertices = new float[][] {
            new float[] { -1,-t,0 },
            new float[] { 0,1,t },
            new float[] { 0,1,-t },
            new float[] { 1,t,0 },
            new float[] { 1,-t,0 },
            new float[] { 0,-1,-t },
            new float[] { 0,-1,t },
            new float[] { t,0,1 },
            new float[] { -t,0,1 },
            new float[] { t,0,-1 },
            new float[] { -t,0,-1 },
            new float[] { -1,t,0 },
      };
      for (float[] v : icoshedronVertices) {
            // Normalize the vertices to have unit length. 
         float length = (float)Math.sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
         v[0] /= length;
         v[1] /= length;
         v[2] /= length;
      }      
      icoshedronFaces = new int[][] {
            { 3, 7, 1 },
            { 4, 7, 3 },
            { 6, 7, 4 },
            { 8, 7, 6 },
            { 7, 8, 1 },
            { 9, 4, 3 },
            { 2, 9, 3 },
            { 2, 3, 1 },
            { 11, 2, 1 },
            { 10, 2, 11 },
            { 10, 9, 2 },
            { 9, 5, 4 },
            { 6, 4, 5 },
            { 0, 6, 5 },
            { 0, 11, 8 },
            { 11, 1, 8 },
            { 10, 0, 5 },
            { 10, 5, 9 },
            { 0, 8, 6 },
            { 0, 10, 11 },
      };
   }
   
   
   /**
    * Constructor.  Creates the GLPanel and adds it to the main panel
    * with some controlls at the bottom.  Installs a "trackball" that
    * lets the user rotate the view by dragging the mouse.
    */
   public IcosphereIFS_GLSL() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      JPanel bottom = new JPanel();
      bottom.setLayout(new FlowLayout(FlowLayout.CENTER));
      add(bottom, BorderLayout.SOUTH);
      levelSelect = new JComboBox();
      levelSelect.addItem("No subdivision");
      levelSelect.addItem("1 subdivision level");
      levelSelect.addItem("2 subdivision levels");
      levelSelect.addItem("3 subdivision levels");
      levelSelect.addItem("4 subdivision levels");
      levelSelect.addItem("5 subdivision levels");
      levelSelect.setSelectedIndex(3);
      bottom.add(levelSelect);
      levelSelect.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            dataValid = false;
            drawable.repaint();
         }
      });
      showEdges = new JCheckBox("Show edges of triangles.");
      showEdges.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            drawable.repaint();
         }
      });
      bottom.add(showEdges);
      glslCheckbox = new JCheckBox("Use GLSL", true);
      glslCheckbox.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            drawable.repaint();
         }
      });
      bottom.add(glslCheckbox);
      camera = new Camera();
      camera.setScale(1.4);
      new TrackBall(drawable,camera);  
   }

   
   /**
    * This method creates the GLSL program.  It is called only once.  If the program
    * is successfully created, the variable programOK is set to true, and the program
    * is ready to be used for rendering.  Note that this method separately compiles
    * the vertex and fragment shaders before linking the program.  The compilation
    * is done mostly for debugging; if it were not done explicitly, it would be
    * done automatically when the linkProgram() method in glslProgram is called.
    * The program here is represented by an object of type GLSLProgram, a class
    * that encapsulates most of the operations on GLSL programs.  See that class
    * for more details.
    */
   private void makeGLSLProgram(GL gl) {
      programOK = false;
      StringBuffer prog = new StringBuffer();
      glslProgram = new GLSLProgram();
      glslProgram.bindAttributeLocation("recursion_level", 1);
      prog.append("attribute float recursion_level;\n");
      prog.append("uniform float max_level;\n");
      prog.append("varying float recursion_multiplier;\n");
      prog.append("void main() {\n");
      prog.append("  gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n");
      prog.append("  vec3 normal = normalize(gl_NormalMatrix*gl_Normal);\n");
      prog.append("  float dmult = max(normal.z,0.0);\n"); // N dot L, where L is (0,0,1).
      prog.append("  gl_FrontColor = gl_LightModel.ambient*gl_FrontMaterial.ambient;\n");
      prog.append("  gl_FrontColor += gl_LightSource[0].diffuse*gl_FrontMaterial.diffuse*dmult;\n");
      prog.append("  recursion_multiplier = 2.0 * recursion_level / max_level;\n");
      prog.append("}\n");
      glslProgram.addVertexShaderCode(prog.toString());
      StringBuffer infoLog = new StringBuffer();
      boolean ok = glslProgram.compileVertexShader(gl, infoLog);
      System.out.println("\nInfo log from vertex shader compilation:\n\n" + infoLog + "\n\n");
      if (!ok)
         return;
      prog.setLength(0);
      prog.append("varying float recursion_multiplier;\n");
      prog.append("void main(void) {\n");
      prog.append("   vec4 c = gl_Color;\n");
      prog.append("   c.r = c.r*recursion_multiplier;\n");
      prog.append("   gl_FragColor = c;\n");
      prog.append("}\n");
      glslProgram.addFragmentShaderCode(prog.toString());
      ok = glslProgram.compileFragmentShader(gl, infoLog);
      System.out.println("\nInfo log from fragment shader compilation:\n\n" + infoLog + "\n\n");
      if (!ok)
         return;
      infoLog.setLength(0);
      ok = glslProgram.linkProgram(gl, infoLog);
      System.out.println("Info log from GLSL linker:\n\n" + infoLog + "\n\n");
      if (!ok)
         return;
      programOK = true;
   }
   
   /**
    *  Create an icosphere, based on an icosahedron.  The number of times
    *  that the faces of the icoasahedron are to be subdivided is given
    *  by the level parameters.  The icosphere is considered as an
    *  approximation for the unit sphere.  The data for the icosphere
    *  is stored in the instance variables icosphereVertexBuffer,
    *  icosphereFaceBuffer, vertexCount, and indexCount.
    */
   private void makeIcosphere(int level) {
      ArrayList<Float> verts = new ArrayList<Float>();         // Vertex data.
      ArrayList<Integer> faceIndx = new ArrayList<Integer>();  // Face data.
      ArrayList<Float> recLevelAttrib = new ArrayList<Float>();  // Vertex attributes.
      for (float[] v : icoshedronVertices) { // Add icosahedron vertex to vertex data.
         verts.add(v[0]);
         verts.add(v[1]);
         verts.add(v[2]);
         recLevelAttrib.add((float)level);
      }
      for (int[] f: icoshedronFaces) {
            // Subdivide each face of the icosahedron the given number of times.
         subdivide(f[0],f[1],f[2],verts,recLevelAttrib,faceIndx,level);
      }
      vertexCount = verts.size();
      indexCount = faceIndx.size();
      icosphereVertexBuffer = BufferUtil.newFloatBuffer(verts.size());
      for (float x : verts)   // Copy vertex data into the vertex buffer.
         icosphereVertexBuffer.put(x);
      icosphereVertexBuffer.rewind();
      icosphereIndexBuffer = BufferUtil.newIntBuffer(faceIndx.size());
      for (int i : faceIndx)  // Copy the face data into the face index buffer.
         icosphereIndexBuffer.put(i);
      icosphereIndexBuffer.rewind();
      recLevelAttribBuffer = BufferUtil.newFloatBuffer(recLevelAttrib.size());
      for (float f : recLevelAttrib)
         recLevelAttribBuffer.put(f);
      recLevelAttribBuffer.rewind();
   }
   

   /**
    * Subdivides a triangular face on the unit sphere and stores the
    * data for all the (sub-)faces that are generated into the list of
    * vertex coordinates and the list of vertex indices for faces. 
    * (Note that a given vertex will actually be generated twice, and that
    * no attempt is made to eliminate this redundancy.)
    * @param v1  Index in vertex list of the first vertex of the face.
    * @param v2  Index in vertex list of the second vertex of the face.
    * @param v3  Index in vertex list of the third vertex of the face.
    * @param vertices  The vertex list.
    * @param recLevelAttrib Holds for each vertex the recursion level
    *    at the time the vertex is generated.  This array list will hold
    *    one float value for each vertex and is therefore 1/3 of the size
    *    of the vertex array list.
    * @param faces  The list of vertex indices for each face that is generated.
    * @param level  The number of times the face is to be subdivided.
    */
   private void subdivide(int v1, int v2, int v3, ArrayList<Float> vertices,
             ArrayList<Float> recLevelAttrib, ArrayList<Integer> faces, int level) {
      if (level == 0) {
             // For level 0, add the vertex indices for this face to the vertex data.
         faces.add(v1);
         faces.add(v2);
         faces.add(v3);
      }
      else {  // Subdivide the face into 4 triangles, and then subdivide
             // each of those triangles (level-1) times. The new vertices
             // that are generated are placed in the vertex list.  There
             // is a new vertex half-way between each pair of vertices
             // of the original face.
         float a1 = (vertices.get(3*v1) + vertices.get(3*v2));
         float a2 = (vertices.get(3*v1+1) + vertices.get(3*v2+1));
         float a3 = (vertices.get(3*v1+2) + vertices.get(3*v2+2));
         float length = (float)Math.sqrt(a1*a1+a2*a2+a3*a3);
         a1 /= length;
         a2 /= length;
         a3 /= length;
         int indexA = vertices.size()/3;
         vertices.add(a1);
         vertices.add(a2);
         vertices.add(a3);
         recLevelAttrib.add((float)(level-1));

         float b1 = (vertices.get(3*v3) + vertices.get(3*v2));
         float b2 = (vertices.get(3*v3+1) + vertices.get(3*v2+1));
         float b3 = (vertices.get(3*v3+2) + vertices.get(3*v2+2));
         length = (float)Math.sqrt(b1*b1+b2*b2+b3*b3);
         b1 /= length;
         b2 /= length;
         b3 /= length;
         int indexB = vertices.size()/3;
         vertices.add(b1);
         vertices.add(b2);
         vertices.add(b3);
         recLevelAttrib.add((float)(level-1));

         float c1 = (vertices.get(3*v1) + vertices.get(3*v3));
         float c2 = (vertices.get(3*v1+1) + vertices.get(3*v3+1));
         float c3 = (vertices.get(3*v1+2) + vertices.get(3*v3+2));
         length = (float)Math.sqrt(c1*c1+c2*c2+c3*c3);
         c1 /= length;
         c2 /= length;
         c3 /= length;
         int indexC = vertices.size()/3;
         vertices.add(c1);
         vertices.add(c2);
         vertices.add(c3);
         recLevelAttrib.add((float)(level-1));
         
         subdivide(v1,indexA,indexC,vertices,recLevelAttrib,faces,level-1);
         subdivide(indexA,v2,indexB,vertices,recLevelAttrib,faces,level-1);
         subdivide(indexC,indexB,v3,vertices,recLevelAttrib,faces,level-1);
         subdivide(indexA,indexB,indexC,vertices,recLevelAttrib,faces,level-1);
      }
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    * It also initializes the data for the sphere and cylinder.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0.25f,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      version_1_5 = gl.isExtensionAvailable("GL_VERSION_1_5");
      System.out.println("OpenGL Version 1.5 available: " + version_1_5);
      version_2_0 = gl.isExtensionAvailable("GL_VERSION_2_0");
      System.out.println("OpenGL Version 2.0 available: " + version_2_0);
      if (version_1_5) {
         // For OpenGL Version 1.5 or higher, vertex buffer objects are
         // used to store the vertex, normal, and texture coordinates.
         int[] bufferID = new int[3];    // Get 3 buffer IDs for the data.
         gl.glGenBuffers(3,bufferID,0);
         icosphereVertexID = bufferID[0];
         icopsphereIndexID = bufferID[1];
         recLevelAttribID = bufferID[2];
      }
   }

   /**
    * Draws the icosphere (computing the data for it first, if necessary),
    * and turning on the GLSL program if appropriate.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      camera.apply(gl);
      
      int max_recursion_level = levelSelect.getSelectedIndex();
      
      if (version_2_0) {
         if(glslCheckbox.isSelected()) {  // Install GLSL program.
            if (glslProgram == null)
               makeGLSLProgram(gl); // Creates the program.
            if (programOK) {
               if (!glslProgram.install(gl))
                  System.out.println("Program didn't instal!");
               else {
                  int id = glslProgram.getUniformLocation(gl, "max_level");
                  if (id == -1)
                     System.out.println("Can't find max_level uniform variable!");
                  else
                     gl.glUniform1f(id, max_recursion_level);
               }
            }
         }
         else {
            gl.glUseProgram(0);  // Turn off GLSL program.
         }
      }
      
      if (! dataValid) {  // Create the data for the icosphere, and set up vertex arrays.
         makeIcosphere(max_recursion_level); // Compute the icosphere data.
         if (version_1_5) {
                // Load the vertex and face data into VBOs.
            gl.glBindBuffer(GL.GL_ARRAY_BUFFER, icosphereVertexID);
            gl.glBufferData(GL.GL_ARRAY_BUFFER, vertexCount*4, 
                  icosphereVertexBuffer, GL.GL_STATIC_DRAW);
            gl.glVertexPointer(3,GL.GL_FLOAT,0,0);
            gl.glNormalPointer(GL.GL_FLOAT,0,0);
            if (version_2_0) { 
                  // In version 2.0, we also create a vertex array to hold
                  // the attribute values that are used in the GLSL program.
                  // The attribute associated with a vertex is the recursion
                  // level at which that vertex was generated.
               gl.glBindBuffer(GL.GL_ARRAY_BUFFER, recLevelAttribID);
               gl.glBufferData(GL.GL_ARRAY_BUFFER, (vertexCount/3)*4, 
                     recLevelAttribBuffer, GL.GL_STATIC_DRAW);
               gl.glVertexAttribPointer(1,1,GL.GL_FLOAT,false,0,0);
            }
            gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
            gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, icopsphereIndexID);
            gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, indexCount*4, 
                  icosphereIndexBuffer, GL.GL_STATIC_DRAW);
            gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
         }
         else {
            gl.glVertexPointer(3,GL.GL_FLOAT,0,icosphereVertexBuffer);
            gl.glNormalPointer(GL.GL_FLOAT,0,icosphereVertexBuffer);
            if (version_2_0)
               gl.glVertexAttribPointer(1, 1, GL.GL_FLOAT, false, 0, recLevelAttribBuffer);
         }
         dataValid = true;
      }
      
      gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
      gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
      if (version_2_0)
         gl.glEnableVertexAttribArray(1);
      gl.glColor3f(0.7f,0.7f,0.7f);
      if (version_1_5) {
             // Draw the triangles using the isosphereIndexBuffer VBO for the
             // element data (as well as the isosphereVertexBuffer).
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, icopsphereIndexID);
         gl.glDrawElements(GL.GL_TRIANGLES, indexCount, GL.GL_UNSIGNED_INT, 0);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
      }
      else {
         gl.glDrawElements(GL.GL_TRIANGLES, indexCount, GL.GL_UNSIGNED_INT, icosphereIndexBuffer);
      }
      
      if (version_2_0)
         gl.glUseProgram(0); // Don't ever use shader programs for line drawing.
      
      if (showEdges.isSelected()) {
            // Redraw the triangles with polygon mode set to GL_LINES to show
            // the outlines of the triangles.  Lighting is disabled and polygon
            // offset is used to avoid depth buffer problems caused by the
            // triangles and their outlines being at the same depth.
         gl.glPolygonOffset(-1,-1);
         gl.glEnable(GL.GL_POLYGON_OFFSET_LINE);
         gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
         gl.glDisable(GL.GL_LIGHTING);
         gl.glColor3f(0,0,0);
         if (version_1_5) {
            gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, icopsphereIndexID);
            gl.glDrawElements(GL.GL_TRIANGLES, indexCount, GL.GL_UNSIGNED_INT, 0);
            gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
         }
         else
            gl.glDrawElements(GL.GL_TRIANGLES, indexCount, GL.GL_UNSIGNED_INT, icosphereIndexBuffer);
         gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL);
         gl.glDisable(GL.GL_POLYGON_OFFSET_LINE);
         gl.glEnable(GL.GL_LIGHTING);
      }
      gl.glDisableClientState(GL.GL_VERTEX_ARRAY);
      gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
      if (version_2_0)
         gl.glDisableVertexAttribArray(1);

   }
      
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
