package glsl;

import glutil.*;
import scenegraph3d.*;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.media.opengl.*;

/**
 * This is a version of the MovingLightDemo sample program that adds
 * a fragment shader.  The shader converts color to grayscale.  When
 * the GLSL program is enabled, objects appear in shades of gray.
 * This illustrates the use of a GLSL program, represented by an
 * object of the GLSLProgram class.  The GLSL program in this case is
 * just a very simple fragment shader.  This requires OpenGL version 2.0 or
 * higher.  If it is not available, the checkbox that turns on the
 * GLSL program will be disabled.
 * 
 *     This program constructs a simple animated scene (using the classes
 * from packagescenegraph3d).  Included in the scene graph are several
 * "LightNodes" which represent OpenGL lights as nodes in a scene graph.
 * Two of the lights are in a "sun" and a "moon" that circle the world.
 * Two are in headlights attached to a moving cart. There is also a viewpoint
 * light, but with the brightness turned down from the default.  The
 * user can turn the viewpoint light on and off with a checkbox.
 * The user can animate the scene and rotate it by mouse, in the
 * default view.  The viewer can also choose to see the scene from
 * the point of view of two "AvatarNodes" that represent moving viewpoints
 * within the scene.  This program is a modification and extension of
 * the program MovingCameraDemo, with the addition of lights and
 * materials.  It uses a more advanced version of scene graphs, defined
 * in the scenegraph3d package instead of simplescenegraph3d, with
 * support for lights and material.
 */
public class MovingLightDemoGLSL extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Moving Light Demo with GLSL Fragment Shader");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new MovingLightDemoGLSL());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   
   private ComplexObject3D world;  // The root node in the scene graph.
   
   private ComplexObject3D wheel;  // Various nodes in the scene graph, used for animation.
   private ComplexObject3D cart;
   private ComplexObject3D horse1;
   private ComplexObject3D horse2;
   private ComplexObject3D horseWithRider;
   private ComplexObject3D merryGoRound;
   private ComplexObject3D sun;
   private ComplexObject3D moon;
   
   private SceneNode3D headlightSphere;  // The visible headlight on the car, so emissive color can be changed.
   private LightNode headlightLight;  // The light in the headlight of a car; only on at night.

   private AvatarNode viewerInCar;  // A node in the scene graph representing a viewer riding the cart.
   private AvatarNode viewerOnMerryGoRound;  // A node representing a viewer riding the merry-go-round.
   
   private Camera globalViewer;  // The Camera that represents the global view from "outside" the scene.

   private UVSphere sphere = new UVSphere(48, 32);     // Some shapes to put in the world.
   private Cube cube = new Cube();
   private UVCone cone = new UVCone(0.5,2);
   private UVCylinder cylinder = new UVCylinder(0.5,2);
   
   private JComboBox viewSelect;   // A pop-up menu for selecting point of view.
   
   private JCheckBox useViewLight;   // controls whether the viewpoint light is turned on
   
   private int frameNumber;    // Drives the animation.  A timer adds 1
                                // to this value every 30 milliseconds
                                // and repaints the panel.

   private JCheckBox animate;  // Controls whether the animation is running.

   private Timer animator;  // The timer that drives the animation,
                             // which is started and stopped in response
                             // to the animate checKbox.
   
   private JCheckBox grayScaleCheckbox;  // For turning the fragment shader on/off.
   private GLSLProgram grayScaleProgram; // The GLSL program.
   private boolean programOK;            // Set to true when GLSL program has been compiled successfully.


   /**
    * Creates the OpenGL drawable and adds it to the panel.  Also sets up the
    * user interface and builds the scene graph.
    */
   public MovingLightDemoGLSL() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(700,500));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      JPanel bottom = new JPanel();
      bottom.setLayout(new FlowLayout(FlowLayout.CENTER));
      add(bottom, BorderLayout.SOUTH);
      animate = new JCheckBox("Animate");
      animate.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            if (animate.isSelected())
               animator.start();
            else
               animator.stop();
         }
      });
      bottom.add(animate);
      animator = new Timer(30, new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            frameNumber++;
            drawable.repaint();
         }
      });
      useViewLight = new JCheckBox("Viewpoint Light");
      useViewLight.addActionListener( new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            drawable.repaint();
         }
      } );
      bottom.add(Box.createHorizontalStrut(10));
      bottom.add(useViewLight);
      viewSelect = new JComboBox();
      viewSelect.addItem("Global View");
      viewSelect.addItem("From Front of Cart");
      viewSelect.addItem("From Merry-Go-Round");
      viewSelect.addActionListener( new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            drawable.repaint();
         }
      });
      bottom.add(Box.createHorizontalStrut(10));
      bottom.add(new JLabel("V" +
            "iew: "));
      bottom.add(viewSelect);
      grayScaleCheckbox = new JCheckBox("Grayscale");
      grayScaleCheckbox.addActionListener( new ActionListener() {
         public void actionPerformed(ActionEvent arg0) {
            drawable.repaint();
         }
      });
      bottom.add(Box.createHorizontalStrut(10));
      bottom.add(grayScaleCheckbox);
      globalViewer = new Camera();
      globalViewer.setScale(25);
      globalViewer.setView(0, 20, 80, 0, 0, 0, 0, 1, 0);
      new TrackBall(drawable,globalViewer);  
      buildWorld();
   }
   
   
   /**
    * Build the scene graph that represents the world.  This method is
    * called just once, from the constructor.
    */
   private void buildWorld() {
      viewerInCar = new AvatarNode();
      viewerOnMerryGoRound = new AvatarNode();
      world = new ComplexObject3D();
      
      sun = new ComplexObject3D();  // Create a sun consisting of a light inside a sphere.
      sun.add( new RenderableNode(sphere)
                        .setAmbientAndDiffuseColor(0.3f, 0.3f, 0.1f)
                        .setEmissionColor(0.6f, 0.6f, 0.3f)
                        .setScaling(2.5)
                        .setTranslation(0,20,0) );
      sun.add( new LightNode().setDiffuseColor(0.6f,0.6f,0.4f)
                            .setSpecularColor(0.3f,0.3f,0.2f)
                            .setPosition(0,0,0)
                            .setTranslation(0,20,0));
      world.add(sun);
      
      moon = new ComplexObject3D();  // Moon is a light inside a smaller sphere.
      moon.add( new RenderableNode(sphere)
                        .setAmbientAndDiffuseColor(0.2f, 0.2f, 0.2f)
                        .setEmissionColor(0.5f, 0.5f, 0.5f)
                        .setScaling(1)
                        .setTranslation(0,20,0) );
      moon.add( new LightNode().setDiffuseColor(0.3f,0.3f,0.35f)
                            .setSpecularColor(0.1f,0.1f,0.1f)
                            .setPosition(0,0,0)
                            .setTranslation(0,20,0));
      world.add(moon);
      
      // Create a headlight object, containing a light inside a sphere.
      // Later, two copies of the headlight are added to the cart with
      // different transformations, so we actually get two headlights in the scene.
      
      headlightLight =  new LightNode().setDiffuseColor(0.5f,0.5f,0.4f)
                                  .setPosition(0,0,0)
                                  .setSpotCutoff(20)
                                  .setSpotDirection(1, 0, -0.2f);
      ComplexObject3D headlight = new ComplexObject3D();
      headlightSphere = new RenderableNode(new UVSphere())
                                .setAmbientAndDiffuseColor(0.5f, 0.5f, 0.2f)
                                .setScaling(0.1f, 0.25f, 0.25f);
      headlight.add(headlightSphere);
      headlight.add(headlightLight);

      // Add a big, flat cylinder to represent the ground and an annulus to represent a road.
      // The materials for these objects have no specular color.
      
      world.add(new RenderableNode(new UVCylinder(15,1,96,1,10))
                       .setAmbientAndDiffuseColor(0, 0.4f, 0)
                       .setRotation(90,-1,0,0)
                       .setTranslation(0, -1, 0));
      world.add(new RenderableNode(new UVDisk(12.5, 11, 96, 2))
                       .setAmbientAndDiffuseColor(Color.LIGHT_GRAY)
                       .setRotation(90,-1,0,0)
                       .setTranslation(0, 0.01, 0));
      
      // Add a tree, consisting of a cone and a cylinder.
      
      ComplexObject3D tree = new ComplexObject3D();
      tree.add( new RenderableNode(cone).setTranslation(0,0,1)
                     .setAmbientAndDiffuseColor(0, 0.6f, 0.1f) );
      tree.add( new RenderableNode(cylinder).setScaling(0.3,0.3,0.5)
                     .setAmbientAndDiffuseColor(0.3f,0.15f,0) );
      world.add(new ComplexObject3D(tree).setScaling(3).setRotation(90,-1,0,0).setTranslation(-2,0,-5));
      
      // Add some random shapes around the periphery.  These objects have specular color white.
      
      for (double angle = 0; angle < 2*Math.PI; angle += Math.PI / 10) {
            SceneNode3D r;
            switch ((int)(4*Math.random())) {
            case 0:
               r = new RenderableNode(cube).setScaling(0.5,1,0.5);
               break;
            case 1:
               r = new RenderableNode(sphere).setScaling(0.5);
               break;
            case 2:
               r = new RenderableNode(cone).setRotation(90,-1,0,0).setTranslation(0,-1,0);
               break;
            default:
               r = new RenderableNode(cylinder).setRotation(90,-1,0,0).setTranslation(0,-1,0);
            }
            r.setSpecularColor(0.3f,0.3f,0.3f);
            Color c = Color.getHSBColor((float)Math.random(), (float)(0.5+0.5*Math.random()), 0.6f);
            double x = 14*Math.cos(angle);
            double z = 14*Math.sin(angle);
            world.add(new ComplexObject3D(r).setAmbientAndDiffuseColor(c).setTranslation(x,1,z));
      }
      
      // Create a cart, consisting of a cube and four wheels.  One of the
      // viewer nodes is also added to the card.
      
      wheel = new ComplexObject3D();
      wheel.add(new ComplexObject3D( new RenderableNode( new UVTorus(1,0.75,48,32))).setAmbientAndDiffuseColor(Color.BLUE));
      SceneNode3D spoke = new RenderableNode(new UVCylinder())
                                .setScaling(0.2,0.2,1.7)
                                .setRotation(90,-1,0,0)
                                .setTranslation(0,-0.85,0)
                                .setAmbientAndDiffuseColor(Color.YELLOW);
      wheel.add( spoke );
      wheel.add( new ComplexObject3D(spoke).setRotation(60,0,0,1));
      wheel.add( new ComplexObject3D(spoke).setRotation(-60,0,0,1));
      cart = new ComplexObject3D();
      cart.add(new ComplexObject3D(wheel).setTranslation(2,1,1));
      cart.add(new ComplexObject3D(wheel).setTranslation(-2,1,1));
      cart.add(new ComplexObject3D(wheel).setTranslation(2,1,-1));
      cart.add(new ComplexObject3D(wheel).setTranslation(-2,1,-1));
      cart.add(new RenderableNode(new Cube())
                     .setScaling(2.5,0.3,0.8)
                     .setTranslation(0,1,0)
                     .setAmbientAndDiffuseColor(Color.RED)
                     .setSpecularColor(1,1,1));
      cart.add(new ComplexObject3D(headlight).setTranslation(2.5,1.05,0.4));
      cart.add(new ComplexObject3D(headlight).setTranslation(2.5,1.05,-0.4));
      cart.add(viewerInCar.setRotation(-90, 0, 1, 0).setTranslation(.8, 2, 0));
      cart = new ComplexObject3D(cart.setTranslation(0,0,11.5));
      world.add(cart);
      
      // Create a merry-go-round.  The "horses" on the merry-go-round are just
      // cubes that go up and down.  There are two groups of horses; one group
      // goes up while the other goes down.  One of the viewer nodes is added
      // as a "rider" to one of the horses.
      
      SceneNode3D horse = new RenderableNode(new Cube())
                   .setScaling(1,0.5,0.5)
                   .setTranslation(0,2.5,5)
                   .setAmbientAndDiffuseColor(Color.GREEN);
      horse1 = new ComplexObject3D(horse);
      horse2 = new ComplexObject3D(horse);
      horseWithRider = new ComplexObject3D(horse);
      horseWithRider.add(viewerOnMerryGoRound);
      viewerOnMerryGoRound.setRotation(-80,0,1,0).setTranslation(1,3,5);
      merryGoRound = new ComplexObject3D();
      merryGoRound.add(new RenderableNode(new UVCylinder()) // top of merry-go-round
                       .setScaling(12,12,0.25)
                       .setRotation(90,-1,0,0)
                       .setTranslation(0,0.5,0)
                       .setAmbientAndDiffuseColor(Color.LIGHT_GRAY));
      merryGoRound.add(new RenderableNode(new UVCylinder(0.5,1,32,1))  // bottom
                        .setScaling(12,12,0.25)
                        .setRotation(90,-1,0,0)
                        .setTranslation(0,4.5,0)
                        .setAmbientAndDiffuseColor(Color.LIGHT_GRAY));
      SceneNode3D pole = new RenderableNode(new UVCylinder(0.5,1,32,1))
                      .setAmbientAndDiffuseColor(Color.BLUE)
                      .setScaling(0.2,0.2,4.2)
                      .setRotation(90,-1,0,0)
                      .setTranslation(0,0.8,5);
      merryGoRound.add(new ComplexObject3D(pole));
      merryGoRound.add(horseWithRider);
      for (int i = 1; i < 6; i++) {
         merryGoRound.add(new ComplexObject3D(pole).setRotation(60*i,0,1,0));
         merryGoRound.add(new ComplexObject3D( i % 2 == 0 ? horse1 : horse2 ).setRotation(60*i,0,1,0));
      }
      world.add(new ComplexObject3D(merryGoRound).setScaling(0.8).setTranslation(2,0,4));
      world.setShininess(50);  // This value will be used  for all objects in the world.
      
   } // end buildTheWorld()
   

   /**
    * initialize the GL context
    */
   public void init(GLAutoDrawable drawable) {
      drawable.setGL(new DebugGL(drawable.getGL()));
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0.3f,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, new float[] { 0.3f, 0.3f, 0.3f, 1 }, 0 );
      gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, new float[] { 0.1f, 0.1f, 0.1f, 1 }, 0 );
      gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, new float[] { 0.05f, 0.05f, 0.05f, 1 }, 0);
      gl.glLightModeli(GL.GL_LIGHT_MODEL_LOCAL_VIEWER, GL.GL_TRUE);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_NORMALIZE);
      if ( ! gl.isExtensionAvailable("GL_VERSION_2_0") )
         grayScaleCheckbox.setEnabled(false);
   }

   /**
    * Installs the GLSL program for use in the OpenGL context.  The first time
    * this method is called, the shader is compiled and checked for errors, but
    * note that it is not required to compile the shader before installing the
    * program, since the install() method in GLSLProgram will compile the shader
    * if it has not already been done.  If the program does not compile correctly,
    * then it is not installed.  The info log from the compilation is displayed
    * in any case, whether or not the compilation succeeds. 
    */
   private void installGrayScaleProgram(GL gl) {
      if (grayScaleProgram == null) {
         System.out.println("Creating GLSL program for grayscale.\n");
         grayScaleProgram = new GLSLProgram();
         StringBuffer prog = new StringBuffer();
         prog.append("void main(void) {\n");
         prog.append("   vec4 c = gl_Color;\n");
         prog.append("   float gray =  0.30*c.r + 0.59*c.g + 0.11*c.b;\n");
         prog.append("   gl_FragColor = vec4(gray,gray,gray,c.a);\n");
         prog.append("}\n");
         grayScaleProgram.addFragmentShaderCode(prog.toString());
         StringBuffer infoLog = new StringBuffer();
         programOK = grayScaleProgram.compileFragmentShader(gl, infoLog);
         System.out.println("Fragment shader compilation log:\n\n" + infoLog + "\n\n");
         if (!programOK)
            System.out.println("ERROR: Can't compile fragment shader.");
      }
      if (programOK)
         grayScaleProgram.install(gl);
   }
   
   /**
    * Draw a the world, from the point of view of the viewer currently selected
    * by the user.  Also, apply the transformations appropriate for the current
    * frame number, to implement the animation.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      
      // Check the grayScaleCheckbox.  If it is checked, use the
      // GLSL program that converts color to grayscale.  Otherwise,
      // use default processing, by calling gl.glUseProgram(0) to
      // turn off shader programs.
      
      if (grayScaleCheckbox.isSelected())
         installGrayScaleProgram(gl);
      else
         gl.glUseProgram(0);
      
      // Apply transformations for the current frame.
      // NOTE: This MUST be done before setting up the view, since some
      // of the transformations affect the AvatarNodes that represent two
      // of the possible views.
      
      int sunFrame = frameNumber % 587;
      boolean daytime = sunFrame < 0.23*587 || sunFrame > 0.77*587;
      if (daytime) {
         headlightSphere.setEmissionColor(0,0,0);
         headlightLight.setLit(false);
      }
      else {
         headlightSphere.setEmissionColor(0.5f, 0.5f, 0.2f);
         headlightLight.setLit(true);
      }

      sun.setRotation(sunFrame/587.0*360, 0, 0, 1);  // sun revolves once every 587 frames
      moon.setRotation(sunFrame/587.0*360 - 180, 0, 0, 1);  // moon is 180 degrees behind of sun
      wheel.setRotation(-(frameNumber*5), 0, 0, 1); // make the car wheels rotate
      cart.setRotation(frameNumber, 0,1,0);  // make the car go around the road
      merryGoRound.setRotation((int)(2.7*frameNumber), 0, 1, 0);  // make the merry-go-round turn
      int horseFrame = frameNumber % 60;  // the remaining stuff makes the horses go up and down
      if (horseFrame >= 30)
         horseFrame = 60-horseFrame;
      double horseOffset = (horseFrame - 15) / 30.0;
      horse1.setTranslation(0, horseOffset, 0);
      horse2.setTranslation(0, -horseOffset, 0);
      horseWithRider.setTranslation(0,horseOffset,0);
      
      // Set up the projection and view for the currently selected viewer.

      switch (viewSelect.getSelectedIndex()) {
      case 0:
         globalViewer.apply(gl);
         break;
      case 1:
         viewerInCar.apply(gl);
         break;
      case 2:
         viewerOnMerryGoRound.apply(gl);
      }
      
      gl.glPushAttrib(GL.GL_LIGHTING_BIT); // To limit changes made by turning on lights.// Turn on lights.  Start with light 1 since light 0 is used as a viewpoint light.
      
      world.turnOnLights(gl, 1); // Turn on lights.  Start with light 1 since light 0 is used as a viewpoint light.
      
      if (useViewLight.isSelected()) // turn on viewpoint light if user has selected it.
         gl.glEnable(GL.GL_LIGHT0);
      
      world.draw(gl);  // render the world.
      
      gl.glPopAttrib();
   }
      
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
