package glutil;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

/**
 * Represents a viewer that can be moved around and rotated.
 * It is assumed that the "up" direction is the positive y-axis.
 * The avatar has a position on the xz-plane and a rotation angle
 * giving the direction of the avatar's direction of view as an
 * angle with the negative direction of the z-axis.  The avatar
 * also has a height, which is really just its y-coordinate.
 * 
 * The apply() method is meant to be called at the beginning of 
 * display(), before anything is drawn.  It sets up both the 
 * projection and the viewing transformation.
 */
public class SimpleAvatar {

   private double fieldOfView = 45;  // data for gluPerspective
   private double near = 0.1;
   private double far = 100;
   
   private double x,y,z;  // position
   
   private double angle;  // angle with negative z-axis, in degrees
   
   private GLU glu = new GLU();
   
   /**
    * Sets the parameters that are used to create the projection
    * transformation.  The transformation is created using the 
    * gluPerspective() method from the GLU library, and the parameters
    * to this method are the same as the parameters to that method.
    * @param fieldOfView number of angles that the view subtends in
    *     the vertical direction.  Default is 45.  Must be strictly between
    *     0 and 180.
    * @param near The distance of the near clipping plane from the user.
    *     Default 0.1.  Must be greater than zero.
    * @param far  The distance of the far clipping plane from the user.
    *     Default is 100.  Must be greater than near.
    */
   public void setPerspectiveParamerters(double fieldOfView, double near, double far) {
      if (fieldOfView <= 0 || fieldOfView >= 180)
         throw new IllegalArgumentException("Field of view must be 0 to 180, exclusive");
      if (near <= 0)
         throw new IllegalArgumentException("Near distance must be strictly positive.");
      if (far <= near)
         throw new IllegalArgumentException("Far distance must be greater than near");
      this.fieldOfView = fieldOfView;
      this.near = near;
      this.far = far;
   }
   
   /**
    * Rotates the avatar's view by a given angle from its current view.
    * Positive angles correspond to turning towards the left; negative
    * angles, towards the right.
    * @param angle the angle to turn, given in degrees.
    */
   public void turn(double angle) {
      this.angle += angle;
   }
   
   /**
    * Move the avatar by a given distance along its current direction of
    * view.  Positive distances move the avatar forward; negative distances
    * move the avatar backwards.
    */
   public void move(double distance) {
      double a = angle/180 * Math.PI;
      x += -distance*Math.sin(a);
      z += -distance*Math.cos(a);
   }
   
   /**
    * Set the height of the avatar above the xz-plane.  Default is 0.
    * (This is just the avatar's y-coordinate.)
    */
   public void setHeight( double height ) {
      y = height;
   }
   
   /**
    * Set the avatar's position on the xz-plane.
    */
   public void setPosition(double x, double z) {
      this.x = x;
      this.z = z;
   }
   
   /**
    * Return the avatar to position x=0, z=0, with angle of view = 0.
    * The height is not changed.
    */
   public void home() {
      x = z = angle = 0;
   }
   
   public double getFieldOfView() {
      return fieldOfView;
   }

   public double getNear() {
      return near;
   }

   public double getFar() {
      return far;
   }

   public double getX() {
      return x;
   }

   public double getHeight() {
      return y;
   }

   public double getZ() {
      return z;
   }

   public double getAngle() {
      return angle;
   }

   /**
    * Apply the projection and viewing transformation needed to implement
    * the avatar's view to a given OpenGL drawing context.  This is meant
    * to be called at the beginning of display, before drawing the world.
    */
   public void apply(GL gl) {
      int[] viewPort = new int[4];
      gl.glGetIntegerv(GL.GL_VIEWPORT, viewPort, 0);
      double aspect = (double)viewPort[2] / viewPort[3];
      gl.glMatrixMode(GL.GL_PROJECTION);
      gl.glLoadIdentity();
      glu.gluPerspective(fieldOfView, aspect, near, far);
      gl.glMatrixMode(GL.GL_MODELVIEW);
      gl.glLoadIdentity();
      gl.glRotated(-angle,0,1,0);
      gl.glTranslated(-x,-y,-z);
   }
}
