package glutil;

import javax.media.opengl.GL;

/**
 * Represents a polyhedral approximation for a cone.  The polyhedron is created
 * by a set of lines from the vertex of the cone to its base (the "slices") and by
 * a set of circles perpendicular to the axis (the "stacks").  The cone has
 * its base in the xy-plane, with center at (0,0,0), and its axis lies
 * along the positive direction of the z-axis.  Drawing the  bottom of the
 * cone is optional.  By default, texture coordinates for the cone are
 * generated, but this feature can be turned off.  The texture will wrap once
 * around the cone.  If the bottom is present, it will be textured
 * by a circular (actually polygonal) piece cut out of the image texture.
 */
public class UVCone implements Renderable {
   
   private boolean makeTexCoords = true;
   private final int slices;
   private final int stacks;
   private final int rings;
   private final double radius;
   private final double height;
   private boolean drawBottom;
   
   /**
    * Create a cone with radius 0.5, 20 slices, and 10 stacks.
    * Bottom is drawn with 5 rings.
    */
   public UVCone() {
      this(0.5,1,20,10,5);
   }
   
   /**
    * Create a cone with radius 0.5, 20 slices, and 10 stacks.
    * Bottom is drawn with 5 rings if the parameter
    * is true; if the parameter is false, bottom is not drawn.
    */
   public UVCone(boolean drawCaps) {
      this(0.5,1,20,10,5);
      drawBottom = drawCaps;
   }
   
   /**
    * Create a cone with a given radius and height, 20 slices, and 10 stacks.
    * Bottom is drawn with 5 rings.
    */
   public UVCone(double radius, double height) {
      this(radius, height, 20, 10, 5);
   }
   
   /**
    * Create a cone with given radius, height, number of slices, and number of stacks.
    * The bottom of the cone is drawn using 5 rings.
    */
   public UVCone(double radius, double height, int slices, int stacks) {
      this(radius, height, slices, stacks, 5);
   }
   
   /**
    * Create a cone with a given radius, number of slices, number of stacks, and
    * number of rings.  The number of slices is the number of divisions parallel to the 
    * axis (like the slices of an orange).  The number of stacks is the number
    * of divisions perpendicular the axis.  If the number or rings is less than or
    * equal to zero, then the bottom is not drawn.  If the number of
    * rings is positive, then the bottom is drawn, and is divided
    * radially into the specified number of rings for drawing; the number of axial 
    * divisions is the same as the number of slices.
    */
   public UVCone(double radius, double height, int slices, int stacks, int rings) {
      if (radius <= 0)
         throw new IllegalArgumentException("Radius must be positive.");
      if (height <= 0)
         throw new IllegalArgumentException("Height must be positive.");
      if (slices < 3)
         throw new IllegalArgumentException("Number of slices must be at least 3.");
      if (slices < 2)
         throw new IllegalArgumentException("Number of stacks must be at least 2.");
      this.radius = radius;
      this.height = height;
      this.slices = slices;
      this.stacks = stacks;
      if (rings <= 0) {
         drawBottom = false;
         this.rings = 0;
      }
      else {
         this.rings = rings;
         drawBottom = true;
      }
   }
   
   public boolean getMakeTexCoords() {
      return makeTexCoords;
   }

   /**
    * Set whether or not texture coordinates should be generated
    * when this cone is rendered.
    * @param makeTexCoords
    */
   public void setMakeTexCoords(boolean makeTexCoords) {
      this.makeTexCoords = makeTexCoords;
   }

   public boolean getDrawBottom() {
      return drawBottom;
   }

   /**
    * Determine whether to draw the bottom cap of the cone.
    * If drawBottom is true and the number of rings was set to zero
    * or less, then 5 rings will be used.
    */
   public void setDrawBottom(boolean drawBottom) {
      this.drawBottom = drawBottom;
   }

   public int getRings() {
      return rings;
   }

   public double getHeight() {
      return height;
   }

   public int getSlices() {
      return slices;
   }

   public int getStacks() {
      return stacks;
   }
   
   public double getRadius() {
      return radius;
   }

   /**
    * Render the cone in an OpenGL context.
    */
   public void render(GL gl) {
      for (int j = 0; j < stacks; j++) {
         double z1 = (height/stacks) * j;
         double z2 = (height/stacks) * (j+1);
         gl.glBegin(GL.GL_QUAD_STRIP);
         for (int i = 0; i <= slices; i++) {
            double longitude = (2*Math.PI/slices) * i;
            double sinLong = Math.sin(longitude);
            double cosLong = Math.cos(longitude);
            double x = cosLong;
            double y = sinLong;
            double nz = radius/height;
            double normLength = Math.sqrt(x*x+y*y+nz*nz);
            gl.glNormal3d(x/normLength,y/normLength,nz/normLength);
            if (makeTexCoords)
               gl.glTexCoord2d(1.0/slices * i, 1.0/stacks * (j+1));
            gl.glVertex3d((height-z2)/height*radius*x,(height-z2)/height*radius*y,z2);
            if (makeTexCoords)
               gl.glTexCoord2d(1.0/slices * i, 1.0/stacks * j);
            gl.glVertex3d((height-z1)/height*radius*x,(height-z1)/height*radius*y,z1);
         }
         gl.glEnd();
      }
      if (drawBottom) {
         int rings = this.rings;
         if (rings <= 0)
            rings = 5;
         gl.glNormal3d(0,0,-1);
         for (int j = 0; j < rings; j++) {
            double d1 = (1.0/rings) * j;
            double d2 = (1.0/rings) * (j+1);
            gl.glBegin(GL.GL_QUAD_STRIP);
            for (int i = 0; i <= slices; i++) {
               double angle = (2*Math.PI/slices) * i;
               double sin = Math.sin(angle);
               double cos = Math.cos(angle);
               if (makeTexCoords)
                  gl.glTexCoord2d(0.5*(1+cos*d2),0.5*(1+sin*d2));
               gl.glVertex3d(radius*cos*d2,radius*sin*d2,0);
               if (makeTexCoords)
                  gl.glTexCoord2d(0.5*(1+cos*d1),0.5*(1+sin*d1));
               gl.glVertex3d(radius*cos*d1,radius*sin*d1,0);
            }
            gl.glEnd();
         }
      }
   }

}
