package glutil;

import javax.media.opengl.GL;

/**
 * Represents a polyhedral approximation for a cylinder.  The polyhedron is created
 * by a set of lines parallel to the axis of the cylinder (the "slices") and by
 * a set of circles perpendicular to the axis (the "stacks").  The cylinder has
 * its base in the xy-plane, with center at (0,0,0), and its axis lies
 * along the positive direction of the z-axis.  Drawing the top and bottom of the
 * cylinder is optional.  By default, texture coodinates for the cylinder are
 * generated, but this feature can be turned off.  The texture will wrap once
 * around the cylinder.  If the top or bottom is present, it will be textured
 * by a circular (actually polygonal) piece cut out of the image texture.
 */
public class UVCylinder implements Renderable {
   
   private boolean makeTexCoords = true;
   private final int slices;
   private final int stacks;
   private final int rings;
   private final double radius;
   private final double height;
   private boolean drawTop;
   private boolean drawBottom;
   
   /**
    * Create a cylinder with radius 0.5, height 1, 20 slices, and 10 stacks.
    * Top and bottom caps are drawn with 5 rings.
    */
   public UVCylinder() {
      this(0.5,1,20,10,5);
   }
   
   /**
    * Create a cylinder with radius 0.5, height 1,  20 slices, and 10 stacks.
    * Top and bottom caps are drawn with 5 rings if the parameter
    * is true; if the parameter is false, caps are not drawn.
    */
   public UVCylinder(boolean drawCaps) {
      this(0.5,1,20,10,5);
      drawTop = drawCaps;
      drawBottom = drawCaps;
   }
   
   /**
    * Create a cylinder with a given radius and height, 20 slices, and 10 stacks.
    * Top and bottom caps are drawn with 5 rings.
    */
   public UVCylinder(double radius, double height) {
      this(radius, height, 20, 10, 5);
   }
   
   /**
    * Create a cylinder with given radius, height, number of slices, and number of stacks.
    * The bottom of the cone is drawn using 5 rings.
    */
   public UVCylinder(double radius, double height, int slices, int stacks) {
      this(radius, height, slices, stacks, 5);
   }
   
   /**
    * Create a cylinder with a given radius, number of slices, number of stacks, and
    * number of rings.  The number of slices is the number of divisions parallel to the 
    * axis (like the slices of an orange).  The number of stacks is the number
    * of divisions perpendicular the axis.  If the number or rings is less than or
    * equal to zero, then the top and bottom caps are not drawn.  If the number of
    * rings is positive, then the top and bottom caps are drawn, and they are divided
    * radially into the specified number of rings for drawing; the number of axial 
    * divisions is the same as the number of slices.
    */
   public UVCylinder(double radius, double height, int slices, int stacks, int rings) {
      if (radius <= 0)
         throw new IllegalArgumentException("Radius must be positive.");
      if (height <= 0)
         throw new IllegalArgumentException("Height must be positive.");
      if (slices < 3)
         throw new IllegalArgumentException("Number of slices must be at least 3.");
      if (slices < 2)
         throw new IllegalArgumentException("Number of stacks must be at least 2.");
      this.radius = radius;
      this.height = height;
      this.slices = slices;
      this.stacks = stacks;
      if (rings <= 0) {
         drawTop = drawBottom = false;
         this.rings = 0;
      }
      else {
         this.rings = rings;
         drawTop = drawBottom = true;
      }
   }
   
   public boolean getMakeTexCoords() {
      return makeTexCoords;
   }

   /**
    * Set whether or not texture coordinates should be generated
    * when this cylinder is rendered.
    * @param makeTexCoords
    */
   public void setMakeTexCoords(boolean makeTexCoords) {
      this.makeTexCoords = makeTexCoords;
   }

   public boolean getDrawTop() {
      return drawTop;
   }

   /**
    * Determine whether to draw the top cap of the cylinder.
    * If drawTop is true and the number of rings was set to zero
    * or less, then 5 rings will be used.
    */
   public void setDrawTop(boolean drawTop) {
      this.drawTop = drawTop;
   }

   public boolean getDrawBottom() {
      return drawBottom;
   }

   /**
    * Determine whether to draw the bottom cap of the cylinder.
    * If drawBottom is true and the number of rings was set to zero
    * or less, then 5 rings will be used.
    */
   public void setDrawBottom(boolean drawBottom) {
      this.drawBottom = drawBottom;
   }

   public int getRings() {
      return rings;
   }

   public double getHeight() {
      return height;
   }

   public int getSlices() {
      return slices;
   }

   public int getStacks() {
      return stacks;
   }
   
   public double getRadius() {
      return radius;
   }

   /**
    * Render the cylinder in an OpenGL context.
    */
   public void render(GL gl) {
      for (int j = 0; j < stacks; j++) {
         double z1 = (height/stacks) * j;
         double z2 = (height/stacks) * (j+1);
         gl.glBegin(GL.GL_QUAD_STRIP);
         for (int i = 0; i <= slices; i++) {
            double longitude = (2*Math.PI/slices) * i;
            double sinLong = Math.sin(longitude);
            double cosLong = Math.cos(longitude);
            double x = cosLong;
            double y = sinLong;
            gl.glNormal3d(x,y,0);
            if (makeTexCoords)
               gl.glTexCoord2d(1.0/slices * i, 1.0/stacks * (j+1));
            gl.glVertex3d(radius*x,radius*y,z2);
            if (makeTexCoords)
               gl.glTexCoord2d(1.0/slices * i, 1.0/stacks * j);
            gl.glVertex3d(radius*x,radius*y,z1);
         }
         gl.glEnd();
      }
      if (drawTop) {
         int rings = this.rings;
         if (rings <= 0)
            rings = 5;
         gl.glNormal3d(0,0,1);
         for (int j = 0; j < rings; j++) {
            double d1 = (1.0/rings) * j;
            double d2 = (1.0/rings) * (j+1);
            gl.glBegin(GL.GL_QUAD_STRIP);
            for (int i = 0; i <= slices; i++) {
               double angle = (2*Math.PI/slices) * i;
               double sin = Math.sin(angle);
               double cos = Math.cos(angle);
               if (makeTexCoords)
                  gl.glTexCoord2d(0.5*(1+cos*d1),0.5*(1+sin*d1));
               gl.glVertex3d(radius*cos*d1,radius*sin*d1,height);
               if (makeTexCoords)
                  gl.glTexCoord2d(0.5*(1+cos*d2),0.5*(1+sin*d2));
               gl.glVertex3d(radius*cos*d2,radius*sin*d2,height);
            }
            gl.glEnd();
         }
      }
      if (drawBottom) {
         int rings = this.rings;
         if (rings <= 0)
            rings = 5;
         gl.glNormal3d(0,0,-1);
         for (int j = 0; j < rings; j++) {
            double d1 = (1.0/rings) * j;
            double d2 = (1.0/rings) * (j+1);
            gl.glBegin(GL.GL_QUAD_STRIP);
            for (int i = 0; i <= slices; i++) {
               double angle = (2*Math.PI/slices) * i;
               double sin = Math.sin(angle);
               double cos = Math.cos(angle);
               if (makeTexCoords)
                  gl.glTexCoord2d(0.5*(1+cos*d2),0.5*(1+sin*d2));
               gl.glVertex3d(radius*cos*d2,radius*sin*d2,0);
               if (makeTexCoords)
                  gl.glTexCoord2d(0.5*(1+cos*d1),0.5*(1+sin*d1));
               gl.glVertex3d(radius*cos*d1,radius*sin*d1,0);
            }
            gl.glEnd();
         }
      }
   }

}
