package glutil;

import javax.media.opengl.GL;

/**
 * Represents a disk lying in the xy-plane, centered at (0,0) and facing toward
 * the positive direction of the z-axis.  The disk is approximated by a regular
 * polygon with a given number of side (default = 32).  This class can also represent
 * an annulus (a disk with a smaller disk removed).  Texture coordinated are generated
 * by default, but this can be turned of.  The texture for the disk is simply a cutout
 * from the full texture.
 */
public class UVDisk implements Renderable {
   
   private final double outerRadius;
   private final double innerRadius;
   private final int slices;
   private final int rings;
   private boolean makeTexCoords = true;
   
   /**
    * Create a disk of radius 1 with 32 slices and 5 rings.
    */
   public UVDisk() {
      this(1,0,32,5);
   }
   
   /**
    * Create a disk with a given radius with 32 slices and 5 rings.
    */
   public UVDisk(double radius) {
      this(radius, 0, 32, 5);
   }
   
   /**
    * Create a disk of radius 1 with given numbers of slices and rings.
    */
   public UVDisk(int slices, int rings) {
      this(1,0,slices,rings);
   }
   
   /**
    * Create a disk of a given radius with given numbers of slices and rings.
    */
   public UVDisk(double radius, int slices, int rings) {
      this(radius,0,slices,rings);
   }
   
   /**
    * Create a disk or annulus lying in the xy-plane centered at the origin.
    * (A disk is just an annulus with the inner radius equal to zero).
    * The first two parameters give the radii of an annulus.  If one of the
    * radii is zero, the result is an annulus rather than a disk.
    * @param outerRadius  The radius of the disk or outer radius of the annulus.
    *    (Note: if the outerRadius is smaller than the innerRadius, then the
    *    values are swapped, so that in effect the first two parameters are
    *    the two radii in eithe order.)
    * @param innerRadius  If 0, the result is a disk.  If greater than zero,
    *    the result is an annulus in which a disk with radius equal to innerRadius
    *    is removed from a disk with radius equal to outerRadius.
    * @param slices The number of sices of the polygon that approximates the
    *    disk.  (Like slices of a pi.)  Must be 3 or more
    * @param rings The number of radial divisions of the disk.  Must be 1 or more.
    */
   public UVDisk(double outerRadius, double innerRadius, int slices, int rings) {
      if (outerRadius == innerRadius)
         throw new IllegalArgumentException("Outer and inner radii can't be the same.");
      if (outerRadius < innerRadius) {
         double temp = outerRadius;
         outerRadius = innerRadius;
         innerRadius = temp;
      }
      if (innerRadius < 0)
         throw new IllegalArgumentException("Radius can't be negative.");
      if (slices < 3)
         throw new IllegalArgumentException("Number of slices must be 3 or more.");
      if (rings < 1)
         throw new IllegalArgumentException("Number of rings must be 1 or more.");
      this.outerRadius = outerRadius;
      this.innerRadius = innerRadius;
      this.slices = slices;
      this.rings = rings;
   }

   public boolean getMakeTexCoords() {
      return makeTexCoords;
   }

   /**
    * Set whether or not texture coordinates should be generated
    * when this cylinder is rendered.  The default is to generate them.
    * @param makeTexCoords
    */
   public void setMakeTexCoords(boolean makeTexCoords) {
      this.makeTexCoords = makeTexCoords;
   }

   public void render(GL gl) {
      gl.glNormal3d(0,0,1);
      double diff = outerRadius - innerRadius;
      for (int j = 0; j < rings; j++) {
         double d1 = innerRadius + (diff/rings) * j;
         double d2 = innerRadius + (diff/rings) * (j+1);
         gl.glBegin(GL.GL_QUAD_STRIP);
         for (int i = 0; i <= slices; i++) {
            double angle = (2*Math.PI/slices) * i;
            double sin = Math.sin(angle);
            double cos = Math.cos(angle);
            if (makeTexCoords)
               gl.glTexCoord2d(0.5*(1+cos*d1/outerRadius),0.5*(1+sin*d1/outerRadius));
            gl.glVertex3d(cos*d1,sin*d1,0);
            if (makeTexCoords)
               gl.glTexCoord2d(0.5*(1+cos*d2/outerRadius),0.5*(1+sin*d2/outerRadius));
            gl.glVertex3d(cos*d2,sin*d2,0);
         }
         gl.glEnd();
      }
   }

}
