package glutil;

import javax.media.opengl.GL;

/**
 * Represents a polyhedral approximation for a sphere.  The polyhedron is created
 * by lines of longitude and latitude on the sphere.  The sphere is centered at
 * (0,0,0), and its axis lies along the z-axis.  When the sphere is rendered,
 * texture coordinates are generated, by default, although this feature can be
 * turned off.  The sphere is rendered as a surface.  (Note the "UV" in the name
 * refers to the fact that the sphere is generated as a parametric surface, and
 * parametric surfaces are defined, by convention, by functions of the parameters
 * u and v.)
 */
public class UVSphere implements Renderable {
   
   private boolean makeTexCoords = true;
   private final int slices;
   private final int stacks;
   private final double radius;
   
   /**
    * Create a sphere with radius 1, 20 slices, and 10 stacks.
    */
   public UVSphere() {
      this(1,20,10);
   }
   
   /**
    * Create a sphere with a given radius, 20 slices, and 10 stacks.
    */
   public UVSphere(double radius) {
      this(radius, 20, 10);
   }
   
   /**
    * Create a sphere with radius 1 and with a given number of slices and
    * a given number of stacks.
    */
   public UVSphere(int slices, int stacks) {
      this(1, slices, stacks);
   }
   
   /**
    * Create a sphere with a given radius, number of slices, and number
    * of stacks.  The number of slices is the number of lines of longitude
    * (like the slices of an orange).  The number of stacks is the number
    * of divisions perpendicular the axis; the lines of latitude are the
    * dividing lines between stacks, so there are stacks-1 lines of latitude.
    */
   public UVSphere(double radius, int slices, int stacks) {
      if (radius <= 0)
         throw new IllegalArgumentException("Radius must be positive.");
      if (slices < 3)
         throw new IllegalArgumentException("Number of slices must be at least 3.");
      if (slices < 2)
         throw new IllegalArgumentException("Number of stacks must be at least 2.");
      this.radius = radius;
      this.slices = slices;
      this.stacks = stacks;
   }
   
   public boolean getMakeTexCoords() {
      return makeTexCoords;
   }

   /**
    * Set whether or not texture coordinates should be generated
    * when this sphere is rendered.
    * @param makeTexCoords
    */
   public void setMakeTexCoords(boolean makeTexCoords) {
      this.makeTexCoords = makeTexCoords;
   }

   public int getSlices() {
      return slices;
   }

   public int getStacks() {
      return stacks;
   }
   
   public double getRadius() {
      return radius;
   }

   /**
    * Render the sphere in an OpenGL context.
    */
   public void render(GL gl) {
      for (int j = 0; j < stacks; j++) {
         double latitude1 = (Math.PI/stacks) * j - Math.PI/2;
         double latitude2 = (Math.PI/stacks) * (j+1) - Math.PI/2;
         double sinLat1 = Math.sin(latitude1);
         double cosLat1 = Math.cos(latitude1);
         double sinLat2 = Math.sin(latitude2);
         double cosLat2 = Math.cos(latitude2);
         gl.glBegin(GL.GL_QUAD_STRIP);
         for (int i = 0; i <= slices; i++) {
            double longitude = (2*Math.PI/slices) * i;
            double sinLong = Math.sin(longitude);
            double cosLong = Math.cos(longitude);
            double x1 = cosLong * cosLat1;
            double y1 = sinLong * cosLat1;
            double z1 = sinLat1;
            double x2 = cosLong * cosLat2;
            double y2 = sinLong * cosLat2;
            double z2 = sinLat2;
            gl.glNormal3d(x2,y2,z2);
            if (makeTexCoords)
               gl.glTexCoord2d(1.0/slices * i, 1.0/stacks * (j+1));
            gl.glVertex3d(radius*x2,radius*y2,radius*z2);
            gl.glNormal3d(x1,y1,z1);
            if (makeTexCoords)
               gl.glTexCoord2d(1.0/slices * i, 1.0/stacks * j);
            gl.glVertex3d(radius*x1,radius*y1,radius*z1);
         }
         gl.glEnd();
      }
   }

}
