package glutil;

import javax.media.opengl.GL;

/**
 * Represents a torus (donut shape) centered at the origin, bisected by
 * the xy-plane, with its axis along the z-axis.  The torus is approximated
 * by a polyhedron. Texture coordinates are generated by default that map the 
 * texture once around the torus, in each direction.
 */
public class UVTorus implements Renderable {
   
   private final double outerRadius;
   private final double innerRadius;
   private final int slices;
   private final int rings;
   private boolean makeTexCoords = true;
   
   /**
    * Create a torus with outer radius 1 and inner radius 1/3, 
    * with 48 slices and 32 rings.
    */
   public UVTorus() {
      this(1, 1.0/3, 48, 72);
   }
   
   /**
    * Create a torus with a given outer radius, inner radius equal to
    * 1/3 of the outer radius, 48 slices, and 32 rings.
    */
   public UVTorus(double radius) {
      this(radius, radius/3, 48, 32);
   }
   
   /**
    * Create a disk of outer radius 1 and inner radius 1/3, 
    * with given numbers of slices and rings.
    */
   public UVTorus(int slices, int rings) {
      this(1, 1.0/3, 48, 32);
   }
      
   /**
    * Create a torus lying bisected by the xy-plane, centered at the origin,
    * and with its central axis lying along the z-axis.
    * The first two parameters give the outer and inner radii of the torus.
    * @param outerRadius  The outer radius of the torus.
    *    (Note: if the outerRadius is smaller than the innerRadius, then the
    *    values are swapped, so that in effect the first two parameters are
    *    the two radii in either order.)
    * @param innerRadius  The inner radius of the torus.
    * @param slices The number of slices, like the slices of an orange, that are
    *    used to create the approximate the torus.  The slices are cross-sections
    *    of the tube of the torus. Must be 3 or more.
    * @param rings The number of divisions around the circumference of the tube
    *    of the torus.  Must be 3 or more.
    */
   public UVTorus(double outerRadius, double innerRadius, int slices, int rings) {
      if (outerRadius == innerRadius)
         throw new IllegalArgumentException("Outer and inner radii can't be the same.");
      if (outerRadius < innerRadius) {
         double temp = outerRadius;
         outerRadius = innerRadius;
         innerRadius = temp;
      }
      if (innerRadius < 0)
         throw new IllegalArgumentException("Radius can't be negative.");
      if (slices < 3)
         throw new IllegalArgumentException("Number of slices must be 3 or more.");
      if (rings < 3)
         throw new IllegalArgumentException("Number of rings must be 3 or more.");
      this.outerRadius = outerRadius;
      this.innerRadius = innerRadius;
      this.slices = slices;
      this.rings = rings;
   }

   public boolean getMakeTexCoords() {
      return makeTexCoords;
   }

   /**
    * Set whether or not texture coordinates should be generated
    * when this cylinder is rendered.  The default is to generate them.
    * @param makeTexCoords
    */
   public void setMakeTexCoords(boolean makeTexCoords) {
      this.makeTexCoords = makeTexCoords;
   }

   public void render(GL gl) {
      double centerRadius = (innerRadius + outerRadius) / 2;
      double tubeRadius = outerRadius - centerRadius;
      for (int i = 0; i < slices; i++) {
         double s1 = 1.0/slices * i;
         double s2 = 1.0/slices * (i+1);
         double centerCos1 = Math.cos(2*Math.PI*s1);
         double centerSin1 = Math.sin(2*Math.PI*s1);
         double centerCos2 = Math.cos(2*Math.PI*s2);
         double centerSin2 = Math.sin(2*Math.PI*s2);
         gl.glBegin(GL.GL_QUAD_STRIP);
         for (int j = 0; j <= rings; j++) {
            double t = 1.0/rings * j;
            double cos = Math.cos(2*Math.PI*t - Math.PI);
            double sin = Math.sin(2*Math.PI*t - Math.PI);
            double x1 = centerCos1*(centerRadius + tubeRadius*cos);
            double y1 = centerSin1*(centerRadius + tubeRadius*cos);
            double z1 = sin*tubeRadius;
            gl.glNormal3d(centerCos1*cos,centerSin1*cos,sin);
            if (makeTexCoords)
               gl.glTexCoord2d(s1,t);
            gl.glVertex3d(x1,y1,z1);
            double x2 = centerCos2*(centerRadius + tubeRadius*cos);
            double y2 = centerSin2*(centerRadius + tubeRadius*cos);
            double z2 = sin*tubeRadius;
            gl.glNormal3d(centerCos2*cos,centerSin2*cos,sin);
            if (makeTexCoords)
               gl.glTexCoord2d(s2,t);
            gl.glVertex3d(x2,y2,z2);
         }
         gl.glEnd();
      }
   }

}
