package scenegraph3d;

import java.awt.Color;

import javax.media.opengl.GL;

/**
 * A LightNode is a screen graph node that represents
 * a light.  The light uses the ambient, diffuse, and
 * specular colors that it inherits from SceneNode3D.
 * However, in a light node, these colors cannot be null.
 * (The emissive material color from SceneNode3D is ignored.)
 * Lights also have position, and they can have attenuation
 * and spotlight properties.  Note that a LightNode does
 * not have a fixed OpenGL light number.  Light numbers
 * are managed by the turnOnLights() method (in SceneNode3D).
 * A LightNode can be turned off, in which case no OpenGL
 * light number will be used for the light.
 * Because of limitations built into OpenGL, you can't have
 * more than 8 lights in a scene that are turned on at the
 * same time.  Since the scene graph is a graph, not a 
 * tree, one LightNode in the graph can become several lights
 * in the scene.  The 8-light limit applies to the number of
 * lights in the scene, not just to the number of LightNodes.
 * 
 * Note that all the setter methods in this class return a 
 * pointer to the same object, meant to be used for chaining 
 * of methods calls.
 */
public class LightNode extends SceneNode3D {
   
   private float[] position = { 0, 0, 1, 0};
   
   private float[] attenuationConstants = { 1, 0, 0 };
   
   private float[] spotDirection = { 0, 0, -1 };
   private float spotCutoff = 180;
   private float spotExponent = 0;
   
   private boolean lit = true;
   
   /**
    * Create a LightNode with default properties.  Ambient
    * intensity is (0,0,0).  Diffuse and specular intensities
    * are both (1,1,1).  The light is a directional light,
    * from the direction (0,0,1).
    */
   public LightNode() {
      setAmbientColor(0,0,0);
      setDiffuseColor(1,1,1);
      setSpecularColor(1,1,1);
   }
   
   /**
    * This method enables the light and sets the position, colors, 
    * and other properties of the light -- if the light is on.
    * Lights that are off have no effect and do not consume
    * an OpenGL light number.  Also, if more than 8 lights have
    * already been used, then this light will have no effect.
    */
   protected int turnOnLightsBasic(GL gl, int nextLightNumber) {
      if (! lit || nextLightNumber > 7)
         return nextLightNumber;
      int light = GL.GL_LIGHT0 + nextLightNumber;
      gl.glEnable(light);
      gl.glLightfv(light, GL.GL_POSITION, position, 0);
      gl.glLightfv(light, GL.GL_AMBIENT, getAmbientColor(), 0);
      gl.glLightfv(light, GL.GL_DIFFUSE, getDiffuseColor(), 0);
      gl.glLightfv(light, GL.GL_SPECULAR, getSpecularColor(), 0);
      if (position[3] != 0) {
         gl.glLightf(light, GL.GL_SPOT_CUTOFF, spotCutoff);
         if (spotCutoff != 180) {
            gl.glLightfv(light, GL.GL_SPOT_DIRECTION, spotDirection, 0);
            gl.glLightf(light, GL.GL_SPOT_EXPONENT, spotExponent);
         }
         gl.glLightf(light, GL.GL_CONSTANT_ATTENUATION, attenuationConstants[0]);
         gl.glLightf(light, GL.GL_LINEAR_ATTENUATION, attenuationConstants[1]);
         gl.glLightf(light, GL.GL_QUADRATIC_ATTENUATION, attenuationConstants[2]);
      }
      return nextLightNumber + 1;
   }
   
   protected void drawBasic(GL gl) {
      // nothing to draw!
   }
   
   /**
    * Say whether this light is on or not.  Lights that are not on have
    * no effect on the scene.  The default is for the light to be on.
    */
   public LightNode setLit(boolean lit) {
      this.lit = lit;
      return this;
   }
   
   /**
    * Tells whether this light is turned on.  Lights are on by default.
    */
   public boolean isLit() {
      return lit;
   }
   
   /**
    * Sets the ambient color of this light to specified RGB values.
    * This method does the same thing as super.setAmbientColor(r,g,b)
    * but is overridden here to return a LightNode rather than a
    * SceneNode3D.
    */
   public LightNode setAmbientColor(float r, float g, float b) {
      super.setAmbientColor(r,g,b);
      return this;
   }
   
   /**
    * Sets the diffuse color of this light to specified RGB values.
    * This method does the same thing as super.setDiffuseColor(r,g,b)
    * but is overridden here to return a LightNode rather than a
    * SceneNode3D.
    */
   public LightNode setDiffuseColor(float r, float g, float b) {
      super.setDiffuseColor(r,g,b);
      return this;
   }
   
   /**
    * Sets the specular color of this light to specified RGB values.
    * This method does the same thing as super.setSpecularColor(r,g,b)
    * but is overridden here to return a LightNode rather than a
    * SceneNode3D.
    */
   public LightNode setSpecularColor(float r, float g, float b) {
      super.setSpecularColor(r,g,b);
      return this;
   }
   
   /**
    * Sets both the ambient and diffuse color of this light to a specified
    * color.  If the specified color is null, the colors are set to
    * their default values: black for ambient and white for diffuse.
    */
   public LightNode setAmbientAndDiffuseColor(Color c) {
      if (c == null) {
         super.setAmbientColor(0, 0, 0, 1);
         super.setDiffuseColor(0, 0, 0, 1);
      }
      else 
         super.setAmbientAndDiffuseColor(c);
      return this;
   }

   /**
    * Sets the ambient color of this light to a specified
    * color.  If the specified color is null, the ambient color
    * is set to its default value, black.
    */
   public LightNode setAmbientColor(Color c) {
      if (c == null)
         super.setAmbientColor(0, 0, 0);
      else 
         super.setAmbientColor(c);
      return this;
   }

   /**
    * Sets the diffuse color of this light to a specified
    * color.  If the specified color is null, the diffuse color
    * is set to its default value, white.
    */
   public LightNode setDiffuseColor(Color c) {
      if (c == null)
         super.setDiffuseColor(1, 1, 1);
      else 
         super.setDiffuseColor(c);
      return this;
   }

   /**
    * Sets the specular color of this light to a specified
    * color.  If the specified color is null, the specular color
    * is set to its default value, white.
    */
   public LightNode setSpecularColor(Color c) {
      if (c == null)
         super.setSpecularColor(0, 0, 0);
      else 
         super.setSpecularColor(c);
      return this;
   }
   
   /**
    * Makes this a positional light at (x,y,z).  (The position
    * is subject to whatever transformations are applied to
    * this LightNode.)
    */
   public LightNode setPosition(float x, float y, float z) {
      position = new float[] { x, y, z, 1 };
      return this;
   }
   
   /**
    * Makes this a positional light with light coming from the
    * direction of (x,y,z).  x, y, and z should not all be zero.
    * (The direction is subject to whatever transformations are
    * applied to this LightNode.)
    */
   public LightNode setDirection(float x, float y, float z) {
      position = new float[] { x, y, z };
      return this;
   }

   /**
    * Sets the position or direction of this light using
    * homogeneous coordinates, as in standard OpenGL, and
    * subject to whatever transformations are applied to
    * this LightNode.
    */
   public LightNode setPosition(float x, float y, float z, float w) {
      position = new float[] { x, y, z, w };
      return this;
   }
   
   /**
    * Sets the attenuation factors for this light.  Note that they are
    * ignored in the case of a directional light.  By default, the values
    * are (1,0,0), and there is no attenuation.  The constants should
    * not all be 0.
    */
   public LightNode setAttenuationConstants(float constantAttenuation, 
                         float linearAttenuation, float quadraticAttenuation) {
      this.attenuationConstants = new float[] { constantAttenuation, 
                                      linearAttenuation, quadraticAttenuation};
      return this;
   }
   
   /**
    * Sets the value of the property SPOT_CUTOFF for this light.
    * This is ignored if the light is directional.  For a positional
    * light, setting a value in the range 0 to 90 inclusive makes this
    * light into a spotlight.  For any other value, the light is a 
    * regular positional light.  (Values other than 0 to 90 are 
    * converted to 180, which is the OpenGL value for a non-spotlight.)
    * The default value is 180.
    */
   public LightNode setSpotCutoff(float spotCutoff) {
      if (spotCutoff < 0 || spotCutoff > 90)
         this.spotCutoff = 180;
      else
         this.spotCutoff = spotCutoff;
      return this;
   }

   /**
    * Sets the spotlight direction for this light.  This is ignored unless
    * the light is a spotlight.  For a spotlight, the center of the light
    * cone will shine from the position of the light in the direction given
    * by the vector (x,y,z).  x, y, and z should not all be zero.  The
    * direction is subject to transformations that are applied to this LightNode.
    */
   public LightNode setSpotDirection(float x, float y, float z) {
      this.spotDirection = new float[] { x, y, z };
      return this;
   }

   /**
    * Sets the spot exponent property of this light.  This is ignored unless
    * the light is a spotlight.  The default value is 0.  Values are clamped
    * to the range 0 to 128, which are the only legal values in OpenGL.
    * @param spotExponent
    * @return
    */
   public LightNode setSpotExponent(float spotExponent) {
      if (spotExponent < 0)
         this.spotExponent = 0;
      else if (spotExponent > 128)
         this.spotExponent = 128;
      else
         this.spotExponent = spotExponent;
      return this;
   }

   public float[] getPosition() {
      return position;
   }

   public float[] getAttenuationConstants() {
      return attenuationConstants;
   }

   public float[] getSpotDirection() {
      return spotDirection;
   }

   public float getSpotCutoff() {
      return spotCutoff;
   }

   public float getSpotExponent() {
      return spotExponent;
   }

   
}
