package scenegraph3d;

import java.awt.Color;
import java.util.ArrayList;

import javax.media.opengl.GL;

/**
 * Represents a node in 3D scene graphs that support lights and
 * materials, cameras, and basic "renderable" objects.
 */
abstract public class SceneNode3D {

   private double scaleX = 1, scaleY  = 1, scaleZ = 1;
   private double rotationAngle = 0;
   private double rotationAxisX = 0, rotationAxisY = 0, rotationAxisZ = 1;
   private double translateX = 0, translateY = 0, translateZ = 0;
   
   private float[] color;
   private float[] specularColor, diffuseColor, ambientColor, emissionColor;
   
   private int shininess = -1;
   
   abstract protected void drawBasic(GL gl);  // Draw the object represented by this node, with no transform.

   /**
    * Enable all lights that are contained in the scene graph, and set their
    * properties.  Note:  If the number of lights exceeds 8, extra lights are
    * ignored.  This method can be called from the top level with parameter
    * zero if all lights are in the scene graph.  If there are other lights,
    * they should be turned on separately, and this method should be called
    * at the top level with a number greater than 0.
    * @param gl The OpenGL drawing context.
    * @param nextLightNumber  The number (0-7) of the next available light BEFORE
    *    this method is called.  It is assumed that all lights starting with this
    *    number are available for use in the scene graph.  The numbers correspond
    *    to the OpenGL light numbers GL_LIGHT0, GL_LIGHT1, etc.
    * @return The number (0-8) of the next available light, AFTER this method
    *    is called.  The return value will not exceed 8, and a return value of
    *    8 means that the supply of lights has been exhausted.
    */
   final public int turnOnLights(GL gl, int nextLightNumber) {
          // Applies transforms, then calls turnOnLightsBasic() to activate
          // the lights themselves, subject to the appropriate transformation.
      gl.glPushMatrix();
      gl.glTranslated(translateX, translateY, translateZ);
      gl.glRotated(rotationAngle, rotationAxisX, rotationAxisY, rotationAxisZ);
      gl.glScaled(scaleX, scaleY, scaleZ);
      nextLightNumber = turnOnLightsBasic(gl,nextLightNumber);
      gl.glPopMatrix();
      return nextLightNumber;
   }
   
   protected int turnOnLightsBasic(GL gl, int nextLightNumber) { // turn on lights, w/o transforms.
      return nextLightNumber;
   }
   
   /**
    * Renders the object represented by this node, applying the transforms and
    * colors set for this node.  Transformations are always applied in the
    * order: scale, rotate, translate.
    * @param gl the OpenGL context into which the object is rendered.
    */
   final public void draw(GL gl) {
      boolean hasColor = (color != null) || (specularColor != null) 
                 || ambientColor != null || diffuseColor != null || emissionColor != null
                 || shininess >= 0;
      if (hasColor) {
         gl.glPushAttrib(GL.GL_LIGHTING_BIT | GL.GL_CURRENT_BIT);
         if (color != null)
            gl.glGetFloatv(GL.GL_CURRENT_COLOR, color, 0);
         if (ambientColor != null)
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT, ambientColor, 0);
         if (diffuseColor != null)
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_DIFFUSE, diffuseColor, 0);
         if (specularColor != null)
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, specularColor, 0);
         if (emissionColor != null)
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_EMISSION, emissionColor, 0);
         if (shininess >= 0)
            gl.glMateriali(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, shininess);
      }
      gl.glPushMatrix();
      gl.glTranslated(translateX, translateY, translateZ);
      gl.glRotated(rotationAngle, rotationAxisX, rotationAxisY, rotationAxisZ);
      gl.glScaled(scaleX, scaleY, scaleZ);
      drawBasic(gl);
      gl.glPopMatrix();
      if (hasColor)
         gl.glPopAttrib();
   }
   
   /**
    * Set the scaling transformation for this node, with the same scaling factor 
    * in the x, y, and z directions.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setScaling( double uniformScaleFactor ) {
      scaleX = scaleY = scaleZ = uniformScaleFactor;
      return this;
   }
   
   /**
    * Set the scaling transformation for this node, with the specified scaling factors
    * in the x, y, and z directions.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setScaling( double scaleX, double scaleY, double scaleZ ) {
      this.scaleX = scaleX;
      this.scaleY = scaleY;
      this.scaleZ = scaleZ;
      return this;
   }
   
   /**
    * Set the translation transformation for this node, with given translation
    * amounts in the x, y, and z directions.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setTranslation( double dx, double dy, double dz ) {
      translateX = dx;
      translateY = dy;
      translateZ = dz;
      return this;
   }
   
   /**
    * Set the rotation transformation for this node, with the angle of rotation and
    * the axis about which the rotation will be done.  The axis passes through the
    * origin and the point (axisX,axisY,axisZ).
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setRotation( double angle, double axisX, double axisY, double axisZ) {
      rotationAngle = angle;
      rotationAxisX = axisX;
      rotationAxisY = axisY;
      rotationAxisZ = axisZ;
      return this;
   }
   
   /**
    * Returns an array containing the x, y, and z components of the translation.
    */
   public double[] getTranslation() {
      return new double[] { translateX, translateY, translateZ };
   }
   
   /**
    * Returns an array containing the x, y, and z scaling factors.
    */
   public double[] getScaling() {
      return new double[] { scaleX, scaleY, scaleZ };
   }
   
   /**
    * Returns an array containing angle of rotation and the x, y, and z
    * values that determine the axis of rotation.
    */
   public double[] getRotation() {
      return new double[] { rotationAngle, rotationAxisX, rotationAxisY, rotationAxisZ }; 
   }

   /**
    * Set the current color for this node.  If the color is null, which is the default,
    * then this node inherits its color from its parent node (or from the OpenGL
    * environment, if it has no parent node).  This color applies if lighting is off
    * or if GL_COLOR_MATERIAL is enabled.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setColor(Color color) {
      this.color = color == null? null : color.getRGBColorComponents(null);
      return this;
   }
   
   /**
    * Set the current color for this node to given RGB values, with alpha equal
    * to 1.  This color applies if lighting is off or if GL_COLOR_MATERIAL is enabled.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setColor(float r, float g, float b) {
      this.color = new float[] { r, g, b, 1 };
      return this;
   }
   
   /**
    * Set the current color for this node to given RGBA values
    * This color applies if lighting is off or if GL_COLOR_MATERIAL is enabled.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setColor(float r, float g, float b, float alpha) {
      this.color = new float[] { r, g, b, alpha };
      return this;
   }
   
   public float[] getColor() {
      return color;
   }
   
   /**
    * Sets both ambient and diffuse material color of this node.  If the color is null, which
    * is the default, then the ambient and diffuse colors are inherited from the parent node or from
    * the OpenGL environment if there is no parent.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setAmbientAndDiffuseColor(Color c) {
      this.ambientColor = c == null? null : c.getRGBColorComponents(null);
      this.diffuseColor = c == null? null : c.getRGBColorComponents(null);
      return this;
   }

   /**
    * Set both ambient and diffuse material color of this node to given a RGB, with alpha = 1.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setAmbientAndDiffuseColor(float r, float g, float b) {
      this.ambientColor = new float[] { r, g, b, 1 };
      this.diffuseColor = new float[] { r, g, b, 1 };
      return this;
   }

   /**
    * Set both ambient and diffuse material color of this node to given a RGBA.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setAmbientAndDiffuseColor(float r, float g, float b, float alpha) {
      this.ambientColor = new float[] { r, g, b, alpha };
      this.diffuseColor = new float[] { r, g, b, alpha };
      return this;
   }
   
   /**
    * Set the current ambient material color of this node.  If the color is null, which
    * is the default, then the ambient color is inherited from the parent node or from
    * the OpenGL environment if there is no parent.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setAmbientColor(Color c) {
      this.ambientColor = c == null? null : c.getRGBColorComponents(null);
      return this;
   }

   /**
    * Set the current ambient material color of this node to given a RGB, with alpha = 1.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setAmbientColor(float r, float g, float b) {
      this.ambientColor = new float[] { r, g, b, 1 };
      return this;
   }

   /**
    * Set the current ambient material color of this node to given a RGBA.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setAmbientColor(float r, float g, float b, float alpha) {
      this.ambientColor = new float[] { r, g, b, alpha };
      return this;
   }
   
   public float[] getAmbientColor() {
      return ambientColor;
   }

   /**
    * Set the current diffuse material color of this node.  If the color is null, which
    * is the default, then the diffuse color is inherited from the parent node or from
    * the OpenGL environment if there is no parent.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setDiffuseColor(Color c) {
      this.diffuseColor = c == null? null : c.getRGBColorComponents(null);
      return this;
   }

   /**
    * Set the current diffuse material color of this node to given a RGB, with alpha = 1.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setDiffuseColor(float r, float g, float b) {
      this.diffuseColor = new float[] { r, g, b, 1 };
      return this;
   }

   /**
    * Set the current diffuse material color of this node to given a RGBA.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setDiffuseColor(float r, float g, float b, float alpha) {
      this.diffuseColor = new float[] { r, g, b, alpha };
      return this;
   }
   
   public float[] getDiffuseColor() {
      return diffuseColor;
   }

   /**
    * Set the current specular material color of this node.  If the color is null, which
    * is the default, then the specular color is inherited from the parent node or from
    * the OpenGL environment if there is no parent.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setSpecularColor(Color c) {
      this.specularColor = c == null? null : c.getRGBColorComponents(null);
      return this;
   }

   /**
    * Set the current specular material color of this node to given a RGB, with alpha = 1.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setSpecularColor(float r, float g, float b) {
      this.specularColor = new float[] { r, g, b, 1 };
      return this;
   }

   /**
    * Set the current specular material color of this node to given a RGBA.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setSpecularColor(float r, float g, float b, float alpha) {
      this.specularColor = new float[] { r, g, b, alpha };
      return this;
   }
   
   public float[] getSpecularColor() {
      return specularColor;
   }

   /**
    * Set the current emission material color of this node.  If the color is null, which
    * is the default, then the emission color is inherited from the parent node or from
    * the OpenGL environment if there is no parent.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setEmissionColor(Color c) {
      this.emissionColor = c == null? null : c.getRGBColorComponents(null);
      return this;
   }

   /**
    * Set the current emission material color of this node to given a RGB, with alpha = 1.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setEmissionColor(float r, float g, float b) {
      this.emissionColor = new float[] { r, g, b, 1 };
      return this;
   }

   /**
    * Set the current emission material color of this node to given a RGBA.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setEmissionColor(float r, float g, float b, float alpha) {
      this.emissionColor = new float[] { r, g, b, alpha };
      return this;
   }
   
   public float[] getEmissionColor() {
      return emissionColor;
   }
   
   /**
    * Set the shininess material property for this node.  A value of
    * -1 (or any negative value) means that this node inherits the
    * value of this property from its parent node or from the OpengL
    * Otherwise, the value should be in the range 0 to 128.
    * If the value is greater than 128, then 128 is used.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setShininess(int shininess) {
      if (shininess < 0)
         this.shininess = -1;
      else
         this.shininess = Math.max(128, shininess);
      return this;
   }
   
   /**
    * Returns the shininess material property of this node, in the range
    * 0 to 128, or returns -1 if this node's shininess is inherited.
    */
   public int getShininess() {
      return shininess;
   }
      

   // ----- Support for adding a camera to a scene graph, implemented as package private members ---
   
   ArrayList<SceneNode3D> parents;  // The ComplexObject3D nodes, if any, that contain this node.
                                     // This property is maintained entirely by the ComplexObject3D class.

   void  applyInverseTransform(GL gl) { // applies the inverse of the transform in this node.
      double sx = scaleX == 0 ? 1 : 1/scaleX;
      double sy = scaleY == 0 ? 1 : 1/scaleY;
      double sz = scaleZ == 0 ? 1 : 1/scaleZ;
      gl.glScaled(sx,sy,sz);
      gl.glRotated(-rotationAngle, rotationAxisX, rotationAxisY, rotationAxisZ);
      gl.glTranslated(-translateX, -translateY, -translateZ);
   }
   

}
