package selection;

import glutil.*;
import simplescenegraph3d.*;

import java.awt.*;
import java.awt.event.*;
import java.nio.IntBuffer;

import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.GLU;

import com.sun.opengl.util.BufferUtil;

/**
 * Creates a "world" consisting of a 41-by-41 plane with various shapes
 * scattered randomly on it.  The user has the experience of walking
 * around, by using the up/down arrow keys to move forward and backward
 * and the left/right arrow keys to turn.
 *    The user can select one of the shapes by clicking on it with
 * the mouse.  The size of the selected shape oscillates to show that
 * it is selected.
 */
public class SelectionDemo extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Selection Demo");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new SelectionDemo());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private SceneNode3D theWorld;  // Contains the plane and all the shapes.

   private UVSphere sphere = new UVSphere(48, 32);     // Some shapes to put in the world.
   private Cube cube = new Cube();
   private UVCone cone = new UVCone(0.5,2);
   private UVCylinder cylinder = new UVCylinder(0.5,2);
   
   private SimpleAvatar viewer;  // Represents the viewpoint of the viewer.
   
   private ComplexObject3D selectables; // Contains all the objects that can be selected.
   private SceneNode3D selectedItem;    // If non-null, the currently selected item.
   private Timer selectionTimer;        // Drives the animation of the selected item's size.
   private int selectionPhase;          // Used to indicate the phase of the size oscillation of the selected item.
   private Point clickedPoint;          // If non-null, the point that the user clicked, for use in selection.
   private IntBuffer selectionBuffer;   // A buffer to receive the hit records from the OpenGL selection data.
   
   private GLU glu = new GLU();  // Used for the gluPickMatrix method, to help with selection.

   /**
    * Create the world and the viewer's "avatar".  Set up keyboard, focus, and
    * mouse handling.
    */
   public SelectionDemo() {
      drawable = new GLJPanel();
      drawable.setPreferredSize(new Dimension(700,475));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      final JLabel message = new JLabel("CLICK TO GET FOCUS", JLabel.CENTER);
       message.setOpaque(true);
       message.setBackground(Color.LIGHT_GRAY);
       message.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
       add(message,BorderLayout.SOUTH);
      viewer = new SimpleAvatar();
      buildWorld();
      setBorder(BorderFactory.createLineBorder(Color.GRAY,4));
      addMouseListener(new MouseAdapter() {
         public void mousePressed(MouseEvent evt) {
            requestFocus();
         }
      });
      addFocusListener(new FocusListener() {
         public void focusLost(FocusEvent e) {
            setBorder(BorderFactory.createLineBorder(Color.GRAY,4));
            message.setText("CLICK TO GET FOCUS");
         }
         public void focusGained(FocusEvent e) {
            setBorder(BorderFactory.createLineBorder(Color.CYAN,4));
            message.setText("CLICK TO SELECT; Arrow Keys = Move; R = reset world; HOME = reset view");
         }
      });
      addKeyListener(new KeyAdapter() {
         public void keyPressed(KeyEvent e) {
            int code = e.getKeyCode();
            if (code == KeyEvent.VK_UP)
               viewer.move(0.1);
            else if (code == KeyEvent.VK_DOWN)
               viewer.move(-0.1);
            else if (code == KeyEvent.VK_LEFT)
               viewer.turn(2);
            else if (code == KeyEvent.VK_RIGHT)
               viewer.turn(-2);
            else if (code == KeyEvent.VK_HOME)
               viewer.home();
            else if (code == KeyEvent.VK_R) {
               if (selectionTimer.isRunning())
                  selectionTimer.stop();
               selectedItem = null;
               buildWorld();
               viewer.home();
            }
            drawable.repaint();
         }
      });
      addMouseListener( new MouseAdapter() {
            // When the user clicks the mouse while the panel has the input focus,
            // this method sets the value of clickedPoint to the point where the user
            // clicked and calls the drawable's repaint() method.  When that method
            // sees a non-null clickedPoint, it does a selection pass to find the object
            // that the user clicked on.
         public void mousePressed(MouseEvent evt) {
            if (hasFocus()) {
               clickedPoint = new Point(evt.getX(), evt.getY());
               drawable.repaint();
            }
         }
      } );
      selectionTimer = new Timer(20,new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            selectionPhase++;
            drawable.repaint();
         }
      });
   }
   
   /**
    * Create a SceneNode3D to represent the contents of the world.  The
    * plane is a scaled cube that has been translated in the negative y
    * direction so that its top is at y = -1.  The other objects have
    * their centers on the xz-plane, which is the same plane in which the
    * viewer moves, so the centers of the objects are at eye height.
    * The viewer starts out at (0,0), so we avoid placing objects within
    * 2 units of that position.  Otherwise, each position on a 41-by-41
    * grid has a 2% chance of being occupied. All the objects are scaled to
    * just fit within a square of side 1.  Cubes, cones, and cylinders are
    * scaled so that they sit on the plane.  Spheres have radius 0.5 and
    * float above the plane.
    */
   private void buildWorld() {
      selectables = new ComplexObject3DWithSelectionNames();  // Will contain selectable items.
      SceneNode3D ground = new RenderableNode(cube).setScaling(21, 1, 21)
            .setColor(Color.WHITE)
            .setTranslation(0,-2,0);
      for (int i = -20; i <= 20; i++)
         for (int j = -20; j <= 20; j++) {
            if (Math.random() < 0.02 && (Math.abs(i) > 2 || Math.abs(j) > 2)) {
               SceneNode3D r;
               switch ((int)(4*Math.random())) {
               case 0:
                  r = new ComplexObject3D(new RenderableNode(cube).setScaling(0.5,1,0.5));
                  break;
               case 1:
                  r = new ComplexObject3D(new RenderableNode(sphere).setScaling(0.5));
                  break;
               case 2:
                  r = new RenderableNode(cone).setRotation(90,-1,0,0).setTranslation(0,-1,0);
                  break;
               default:
                  r = new RenderableNode(cylinder).setRotation(90,-1,0,0).setTranslation(0,-1,0);
               }
               Color c = Color.getHSBColor((float)Math.random(), (float)(0.5+0.5*Math.random()), 0.8f);
               selectables.add(new ComplexObject3D(r).setColor(c).setTranslation(i,0,j));
            }
         }
      ComplexObject3D world = new ComplexObject3D();
      world.add(selectables);
      world.add(ground);
      theWorld = world;
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0.3f,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_COLOR_MATERIAL);
      gl.glEnable(GL.GL_NORMALIZE);
      selectionBuffer = BufferUtil.newIntBuffer(200);
      gl.glSelectBuffer(200,selectionBuffer); // Results of GL_SELECT mode hits will be placed in selectionBuffer.
   }

   /**
    * If clickedPoint is non-null, the display method will use OpenGL SELECT render mode
    * figure out which item was clicked.  The clickable items are all stored in a
    * ComplexObject3DWithSelectionNames, which manages the name stack.  The name
    * of each item is the same as its position in the list of items in this complex
    * object.  Whether or not it does a selection pass, this method also
    * draws a the world, from the point of view of the viewer's avatar.
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      
      /* If clickedPoint is non-null, then the user has clicked on the drawable, and
       * we need to figure out which object, if any, was clicked.  We do with by using
       * the GL_SELECT render mode.
       */
      
      if (clickedPoint != null) {
         System.out.println("Selecting point at " + clickedPoint.getX() + "," + clickedPoint.getY());
         gl.glRenderMode(GL.GL_SELECT);  // Objects will be processed in SELECT mode, not rendered.
         int[] viewport = new int[4];         // The viewport, required for gluPickMatrix().
         float[] projMatrix = new float[16];  // The projection matrix used on the PREVIOUS render.
         gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
         gl.glGetFloatv(GL.GL_PROJECTION_MATRIX, projMatrix, 0);
         
         /* Set up a projection matrix for selection.  The view volume includes only the
          * pixel that the user clicked.  When the objects are processed in SELECT mode, we
          * will get back a list of "names" of objects that intersect that view volume.  These
          * are all the candidates for the object that the user wants to select, and we just
          * have to pick the closest one.  The names are just integers.  SELECT mode requires
          * the name to be on a "name stack", which is maintained in this program by
          * "selectables", an object of type ComplexObject3DWithSelectionNames.
          *     The projection for SELECT mode is made by applying glu.gluPickMatrix() to
          * the clickedPoint, composed with the projection matrix from the previous render
          * (which, after all, produced the picture that the user clicked on).
          */
         
         gl.glMatrixMode(GL.GL_PROJECTION);
         gl.glLoadIdentity();
         glu.gluPickMatrix(clickedPoint.x, drawable.getHeight()-clickedPoint.y, 1, 1, viewport, 0);
         gl.glMultMatrixf(projMatrix, 0); // following code from "OpenGL Distilled"
         gl.glMatrixMode(GL.GL_MODELVIEW);
         
         gl.glInitNames();   // Initialize the name stack.
         theWorld.draw(gl);  // Draw the world (putting item names on name stack as appropriate).
         
         /* Changing back to GL_RENDER mode returns the number of items that were in the view volume.
          * Data about each of the "hits" has been placed into the selectionBuffer, which was
          * set up in the init() method.
          */
         
         int hitCount = gl.glRenderMode(GL.GL_RENDER);
         System.out.println("   " + hitCount + " items hit");
         
         if (selectedItem != null) {
               // If there was a previous selected item, reset it to normal size.
            selectedItem.setScaling(1);
         }
         selectedItem = null;  // No selected item.  If an item was clicked, the value will be changed.
         
         /* We have to parse the selectionBuffer to get information about the items that were hit.
          * It contains one hit record for each hit.  A hit record contains:  The number of items
          * on the name stack; the minimum and maximum depth values for the object that was hit;
          * the contents of the name stack at the time of the hit, from bottom to top.
          *    Note that the depth range values are unsigned ints, scaled from 0 to 2^32-1.  Since
          * Java does not have an unsigned int type, we convert them to long values keeping only
          * the bottom 32 bits; this converts negative ints to the corresponding unsigned int
          * value expressed as a long.
          *    The goal here is to set selectedItem to the hit object that was closest to the
          * viewer, that is, that has that smallest minimum depth value.  This is the object
          * that the user could see at the point that was clicked.
          */
         
         int pos = 0;  // Index into the selection buffer.
         long closestZ = Long.MAX_VALUE;
         for (int i = 0; i < hitCount; i++) {
            int stackDepth = selectionBuffer.get(pos);  // In fact can only be 0 or 1 in this program.
            long minZ = selectionBuffer.get(pos+1) & 0xFFFFFFFFL;
            pos += 3;
            if (stackDepth > 0 && minZ < closestZ) {
               closestZ = minZ;
               int itemName = selectionBuffer.get(pos + stackDepth-1);
               selectedItem = selectables.getItem(itemName); // The item name is just the item index in selectables.
               System.out.println("Selected item " + itemName);
            }
            pos += stackDepth;
         }
         
         if (selectedItem == null) { // No item was selected; make sure the timer is not running.
            if (selectionTimer.isRunning())
               selectionTimer.stop();
         }
         else {  // An item was selected and has to be animated; make sure that the timer is running.
            selectionPhase = 0;
            if (!selectionTimer.isRunning())
               selectionTimer.start();
         }
         System.out.println();
         clickedPoint = null; // We have finished processing this click!
      }
      
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      
      if (selectedItem != null) { // For animating the size of the selected item.
         int phase = selectionPhase % 50;
         if (phase > 25)
            phase = 50-phase;
         selectedItem.setScaling( 0.75 + 0.58*(phase/25.0));
      }
      
      viewer.apply(gl);  // Set up projection and viewing transforms.
      theWorld.draw(gl);
   }
      
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
