package simplescenegraph3d;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

/**
 * Represents a viewer (camera or avatar) that can be placed
 * into a scene graph and that will see things from the point
 * of view of an object at the same point in the scene.  To
 * apply this point of view, the AvatarNode's apply() method
 * should be called at the beginning of the display() method,
 * before drawing anything.  The apply method sets up both
 * the projection and the viewing transformation.  Although
 * an AvatarNode is a node in a scene graph, it does not have
 * any visual appearance in the image.
 */
public class AvatarNode extends SceneNode3D {

   private double fieldOfView = 45;  // data for gluPerspective
   private double near = 0.1;
   private double far = 100;
   
   private GLU glu = new GLU();

   /**
    * Apply the projection and viewing transformation needed to implement
    * the avatar's view to a given OpenGL drawing context.  This is meant
    * to be called at the beginning of display, before drawing the world.
    */
   public void apply(GL gl) {
      int[] viewPort = new int[4];
      gl.glGetIntegerv(GL.GL_VIEWPORT, viewPort, 0);
      double aspect = (double)viewPort[2] / viewPort[3];
      gl.glMatrixMode(GL.GL_PROJECTION);
      gl.glLoadIdentity();
      glu.gluPerspective(fieldOfView, aspect, near, far);
      gl.glMatrixMode(GL.GL_MODELVIEW);
      gl.glLoadIdentity();
      SceneNode3D node = this;
      while (true) {
            // Apply the inverse of the transformation that is applied
            // to this node in the scene graph -- this is the viewing
            // transformation that needs to be applied.  Note that
            // when a node has more than one parent in the graph,
            // only the first parent is constructed.   When an AvatarNode
            // is used as a camera, only the first path the scene graph
            // to the camera applies.  (In fact, it doesn't make a lot
            // of sense to have multiple copies of a camera.)
         node.applyInverseTransform(gl);
         if (node.parents == null) {  // We're at the top of the scene graph.
            break;
         }
         node = node.parents.get(0);
      }
   }

   /**
    * Sets the parameters that are used to create the projection
    * transformation.  The transformation is created using the 
    * gluPerspective() method from the GLU library, and the parameters
    * to this method are the same as the parameters to that method.
    * @param fieldOfView number of angles that the view subtends in
    *     the vertical direction.  Default is 45.  Must be strictly between
    *     0 and 180.
    * @param near The distance of the near clipping plane from the user.
    *     Default 0.1.  Must be greater than zero.
    * @param far  The distance of the far clipping plane from the user.
    *     Default is 100.  Must be greater than near.
    */
   public void setPerspectiveParamerters(double fieldOfView, double near, double far) {
      if (fieldOfView <= 0 || fieldOfView >= 180)
         throw new IllegalArgumentException("Field of view must be 0 to 180, exclusive");
      if (near <= 0)
         throw new IllegalArgumentException("Near distance must be strictly positive.");
      if (far <= near)
         throw new IllegalArgumentException("Far distance must be greater than near");
      this.fieldOfView = fieldOfView;
      this.near = near;
      this.far = far;
   }
   
   public double getFieldOfView() {
      return fieldOfView;
   }

   public double getNear() {
      return near;
   }

   public double getFar() {
      return far;
   }

   protected void drawBasic(GL gl) {
      // Nothing is drawn.
   }
   
}
