package simplescenegraph3d;

import java.awt.Color;
import java.util.ArrayList;

import javax.media.opengl.GL;

/**
 * Represents a node in a simple of version of 3D scene graphs.
 */
abstract public class SceneNode3D {

   private double scaleX = 1, scaleY  = 1, scaleZ = 1;
   private double rotationAngle = 0;
   private double rotationAxisX = 0, rotationAxisY = 0, rotationAxisZ = 1;
   private double translateX = 0, translateY = 0, translateZ = 0;
   
   private Color color;
   
   abstract protected void drawBasic(GL gl);  // Draw the object represented by this node, with no transform.
   
   /**
    * Renders the object represented by this node, applying the transforms and
    * colors set for this node.  Transformations are always applied in the
    * order: scale, rotate, translate.
    * @param gl the OpenGL context into which the object is rendered.
    */
   public void draw(GL gl) {
      float[] currentColor = new float[4];
      if (color != null)
         gl.glGetFloatv(GL.GL_CURRENT_COLOR, currentColor, 0);
      gl.glPushMatrix();
      gl.glTranslated(translateX, translateY, translateZ);
      gl.glRotated(rotationAngle, rotationAxisX, rotationAxisY, rotationAxisZ);
      gl.glScaled(scaleX, scaleY, scaleZ);
      if (color != null) {
         float[] hsb = color.getColorComponents(null);
         gl.glColor3f(hsb[0], hsb[1], hsb[2]);
      }
      drawBasic(gl);
      gl.glPopMatrix();
      if (color != null)
         gl.glColor3f(currentColor[0], currentColor[1], currentColor[2]);
   }
   
   /**
    * Set the scaling transformation for this node, with the same scaling factor 
    * in the x, y, and z directions.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setScaling( double uniformScaleFactor ) {
      scaleX = scaleY = scaleZ = uniformScaleFactor;
      return this;
   }
   
   /**
    * Set the scaling transformation for this node, with the specified scaling factors
    * in the x, y, and z directions.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setScaling( double scaleX, double scaleY, double scaleZ ) {
      this.scaleX = scaleX;
      this.scaleY = scaleY;
      this.scaleZ = scaleZ;
      return this;
   }
   
   /**
    * Set the translation transformation for this node, with given translation
    * amounts in the x, y, and z directions.
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setTranslation( double dx, double dy, double dz ) {
      translateX = dx;
      translateY = dy;
      translateZ = dz;
      return this;
   }
   
   /**
    * Set the rotation transformation for this node, with the angle of rotation and
    * the axis about which the rotation will be done.  The axis passes through the
    * origin and the point (axisX,axisY,axisZ).
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setRotation( double angle, double axisX, double axisY, double axisZ) {
      rotationAngle = angle;
      rotationAxisX = axisX;
      rotationAxisY = axisY;
      rotationAxisZ = axisZ;
      return this;
   }
   
   /**
    * Returns an array containing the x, y, and z components of the translation.
    */
   public double[] getTranslation() {
      return new double[] { translateX, translateY, translateZ };
   }
   
   /**
    * Returns an array containing the x, y, and z scaling factors.
    */
   public double[] getScaling() {
      return new double[] { scaleX, scaleY, scaleZ };
   }
   
   /**
    * Returns an array containing angle of rotation and the x, y, and z
    * values that determine the axis of rotation.
    */
   public double[] getRotation() {
      return new double[] { rotationAngle, rotationAxisX, rotationAxisY, rotationAxisZ }; 
   }

   /**
    * Set the color for this node.  If the color is null, which is the default,
    * then this node inherits its color from its parent node (or from the OpenGL
    * environment, if it has no parent node).
    * @return a reference to this SceneNode3D, which can be used for chaining calls
    *    to setter methods.
    */
   public SceneNode3D setColor(Color color) {
      this.color = color;
      return this;
   }
   
   public Color getColor() {
      return color;
   }

   // ----- Support for adding a camera to a scene graph, implemented as package private members ---
   
   ArrayList<SceneNode3D> parents;  // The ComplexObject3D nodes, if any, that contain this node.
                                     // This property is maintained entirely by the ComplexObject3D class.

   void  applyInverseTransform(GL gl) { // applies the inverse of the transform in this node.
      double sx = scaleX == 0 ? 1 : 1/scaleX;
      double sy = scaleY == 0 ? 1 : 1/scaleY;
      double sz = scaleZ == 0 ? 1 : 1/scaleZ;
      gl.glScaled(sx,sy,sz);
      gl.glRotated(-rotationAngle, rotationAxisX, rotationAxisY, rotationAxisZ);
      gl.glTranslated(-translateX, -translateY, -translateZ);
   }
   

}
