package stereoGraph3d;

import java.awt.*;
import java.awt.event.*;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.swing.*;
import javax.swing.event.*;
import javax.media.opengl.*;

import com.sun.opengl.util.BufferUtil;


/**
 * A Jogl/OpenGL application that draws the graph of a function
 * z = f(x,y), using a 201-by-201 grid of points to approximate
 * the graph.  The function can refer to two parameters a and b.
 * The user inputs the function definition and controls the values 
 * of the parameters a and b.  The program optionally draws axes
 * and grid lines on the surface.
 *    This version of the program uses glDrawElements to draw
 * the surface and lines.  It uses vertex buffer objects to hold
 * the data, if the OpenGL version is 1.5 or higher.
 */
public class StereoGrapher extends JPanel implements GLEventListener {

   public static void main(String[] args) {
      JFrame window = new JFrame("Simple Grapher");
      window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      window.setContentPane(new StereoGrapher());
      window.pack();
      window.setVisible(true);
   }
   
   private GLJPanel drawable;  // The panel where the drawing is done.
   
   private CameraWithStereo camera;  // Contains the current view transform and projection.
                           // The user can drag with the mouse to rotate the view.
   
   private Expr2 func;  // The function that is being drawn.
   
   private JSlider aSlider, bSlider;       // Controls, added to bottom of window.
   private JLabel aValue, bValue;
   private JTextField functionInput;
   private JButton newFunctionButton;
   private JCheckBox axesCheckbox;
   private JCheckBox linesCheckbox;
   private JCheckBox stereoCheckbox;

   private boolean version_1_5;  // Check is OpenGL 1.5 is available; set in init().
   
   private boolean dataIsValid;  // Set to true whenever data needs to be recomputed.
                                 // This is checked in the display() method before drawing.
   
   /* Buffers to hold the points and normals for the surface. */
   
   private FloatBuffer vBuf = BufferUtil.newFloatBuffer(201*201*3);
   private FloatBuffer nBuf = BufferUtil.newFloatBuffer(201*201*3);
   
   /* Buffers to hold the indices for drawing the surface and lines with glDrawElements*/

   private IntBuffer surfaceIndexBuffer = BufferUtil.newIntBuffer(200*201*2);
   private IntBuffer xLineIndexBuffer = BufferUtil.newIntBuffer(21*201);
   private IntBuffer yLineIndexBuffer = BufferUtil.newIntBuffer(21*201);

   /* VBO ID numbers for holding the data when OpenGL version is 1.5 or higher */
   
   private int vertexVBO, normalVBO;            // VBO IDs for surface data.
   private int xLineVBO, yLineVBO, surfaceVBO;  // VBO IDs for index data.
   
   
   /**
    * Create the GLJPanel where the surface is drawn, with a set of controls
    * underneath for setting the function to be drawn and the values of
    * the parameters a and b, and for setting whether axes and grid lines
    * are to be drawn.
    */
   public StereoGrapher() {
      GLCapabilities cap = new GLCapabilities();
      cap.setSampleBuffers(true);
      cap.setNumSamples(4);
      drawable = new GLJPanel(cap);
      drawable.setPreferredSize(new Dimension(600,600));
      drawable.addGLEventListener(this);
      setLayout(new BorderLayout());
      add(drawable, BorderLayout.CENTER);
      aValue = new JLabel(" a = +1.000");
      bValue = new JLabel(" b = +1.000");
      aValue.setFont(new Font("Monospace", Font.PLAIN, 12));
      bValue.setFont(new Font("Monospace", Font.PLAIN, 12));
      aSlider = new JSlider(-2000, 2000, 1000);
      bSlider = new JSlider(-2000, 2000, 1000);
      axesCheckbox = new JCheckBox("Axes ", true);
      linesCheckbox = new JCheckBox("Lines");
      stereoCheckbox = new JCheckBox("Stereo");
      JPanel bottom = new JPanel();
      bottom.setBackground(Color.GRAY);
      bottom.setBorder(BorderFactory.createLineBorder(Color.GRAY, 2));
      bottom.setLayout(new GridLayout(0,1,2,2));
      add(bottom, BorderLayout.SOUTH);
      JPanel panel;
      functionInput = new JTextField("2*b*sin( a*(x^2+y^2) ) / (1 + (x^2+y^2))");
      newFunctionButton = new JButton("Graph it!");
      ActionListener al = new ActionListener() {
         public void actionPerformed(ActionEvent evt) {
            if (evt.getSource() instanceof JCheckBox) {
                  // Draw line or draw axes setting has changed.
                  // Just repaint, no need to recompute.
               drawable.repaint();
               return;
            }
            try {
                  // Since the ActionEvent is not from a checkbox, it's the graph
                 // definition that is changing.
               func = new Expr2(functionInput.getText());
               double a = aSlider.getValue()/1000.0;
               double b = bSlider.getValue()/1000.0;
               func.setVariable('a', a);
               func.setVariable('b', b);
               dataIsValid = false;  // Force recomputation of data with new graph definition.
               drawable.repaint();
            }
            catch (IllegalArgumentException e) {
               JOptionPane.showMessageDialog(drawable, 
                     "There is an error in the function definition:\n" + e.getMessage());
            }
         }
      };
      functionInput.addActionListener(al);
      newFunctionButton.addActionListener(al);
      axesCheckbox.addActionListener(al);
      linesCheckbox.addActionListener(al);
      stereoCheckbox.addActionListener(al);
      panel = new JPanel();
      panel.setLayout(new BorderLayout());
      panel.add(new JLabel(" f(x,y) = "), BorderLayout.WEST);
      panel.add(functionInput, BorderLayout.CENTER);
      panel.add(newFunctionButton, BorderLayout.EAST);
      JPanel row1 = new JPanel();
      row1.setLayout(new BorderLayout(5,5));
      row1.add(panel, BorderLayout.CENTER);
      row1.add(stereoCheckbox, BorderLayout.EAST);
      bottom.add(row1);
      ChangeListener cl = new ChangeListener() {
         public void stateChanged(ChangeEvent e) {
              // Value of a or b has changed.
            double a = aSlider.getValue()/1000.0;
            double b = bSlider.getValue()/1000.0;
            aValue.setText(String.format(" a = %+1.3f", a));
            bValue.setText(String.format(" b = %+1.3f", b));
            if (func != null) {
               func.setVariable('a', a);
               func.setVariable('b', b);
               dataIsValid = false;  // Force recomputation with new values of a and b.
               drawable.repaint();
            }
         }
      };
      aSlider.addChangeListener(cl);
      bSlider.addChangeListener(cl);
      panel = new JPanel();
      panel.setLayout(new BorderLayout());
      panel.add(aSlider, BorderLayout.CENTER);
      panel.add(aValue, BorderLayout.EAST);
      panel.add(axesCheckbox, BorderLayout.WEST);
      bottom.add(panel);
      panel = new JPanel();
      panel.setLayout(new BorderLayout());
      panel.add(bSlider, BorderLayout.CENTER);
      panel.add(bValue, BorderLayout.EAST);
      panel.add(linesCheckbox, BorderLayout.WEST);
      bottom.add(panel);
      camera = new CameraWithStereo();
      camera.setView(40, 12, 12, 0, 0, 0, 0, 0, 1);
      camera.setScale(7);
      new TrackBallForStereoCamera(drawable,camera);  
   }

   /**
    * This method will be called when the GLJPanel is first
    * created.  It can be used to initialize the GL context.
    */
   public void init(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClearColor(0,0,0,1);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_LIGHT0);
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glLightModeli(GL.GL_LIGHT_MODEL_TWO_SIDE, GL.GL_TRUE);
      version_1_5 = gl.isExtensionAvailable("GL_VERSION_1_5");
      System.out.println("Version 1.5 available is: " + version_1_5);
      if (gl.isExtensionAvailable("GL_VERSION_1_3"))
         gl.glEnable(GL.GL_MULTISAMPLE);
      makeElementBuffers();  // Generate lists of indices for glDrawElements.  This data never changes.
      if (version_1_5) {
            // Generate VBOs for the data, and fill the ones that are for index data with
            // data from Java nio buffers.  The VBOs for index data won't change again and
            // so use GL.GL_STATIC_DRAW.
         int[] ids = new int[5];
         gl.glGenBuffers(5, ids, 0);
         vertexVBO = ids[0];
         normalVBO = ids[1];
         xLineVBO = ids[2];
         yLineVBO = ids[3];
         surfaceVBO = ids[4];
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vertexVBO);
         gl.glVertexPointer(3,GL.GL_FLOAT,0,0);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, normalVBO);
         gl.glNormalPointer(GL.GL_FLOAT,0,0);
         gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, surfaceVBO);
         gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, 4*2*200*201, surfaceIndexBuffer, GL.GL_STATIC_DRAW);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, xLineVBO);
         gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, 4*21*201, xLineIndexBuffer, GL.GL_STATIC_DRAW);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, yLineVBO);
         gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, 4*21*201, yLineIndexBuffer, GL.GL_STATIC_DRAW);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
      }
      else {
         gl.glVertexPointer(3,GL.GL_FLOAT,0,vBuf);
         gl.glNormalPointer(GL.GL_FLOAT,0,nBuf);
      }
   }

   /**
    * Draws a set of xyz-axes and a surface (depending on settings).
    */
   public void display(GLAutoDrawable drawable) {
      GL gl = drawable.getGL();
      gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      if ( ! stereoCheckbox.isSelected()) {
         camera.apply(gl);
         draw(gl,false);
      }
      else {
         camera.apply(gl,-3);
         gl.glColorMask(true,false,false,false);
         draw(gl,true);
         gl.glClear(GL.GL_DEPTH_BUFFER_BIT);
         gl.glColorMask(false,true,false,false);
         camera.apply(gl,3);
         draw(gl,true);
         gl.glColorMask(true,true,true,true);
      }
   }
   
   /**
    * Does the drawing of one view; called twice for stereo views.
    * @param forStereo  This is only used to determine the colors
    * that are used.  For stereo viewing, only white and gray
    * are used.
    */
   private void draw(GL gl, boolean forStereo) {
      if (axesCheckbox.isSelected()) {
         gl.glDisable(GL.GL_LIGHTING);
         gl.glColor3f(1,1,1);
         gl.glLineWidth(2);
         gl.glBegin(GL.GL_LINES);
         if (!forStereo)
            gl.glColor3f(1, 0, 0);
         gl.glVertex3f(0,0,0);
         gl.glVertex3f(6,0,0);
         if (!forStereo)
            gl.glColor3f(0, 0.7f, 0);
         gl.glVertex3f(0,0,0);
         gl.glVertex3f(0,6,0);
         if (!forStereo)
            gl.glColor3f(0.25f, 0.25f, 1);
         gl.glVertex3f(0,0,0);
         gl.glVertex3f(0,0,6);
         gl.glEnd();
         gl.glLineWidth(1);
         gl.glEnable(GL.GL_LIGHTING);
      }
      if (func != null) {
         if (forStereo) {
            gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, new float[] {1,1,1,1}, 0);
            gl.glMaterialfv(GL.GL_BACK, GL.GL_AMBIENT_AND_DIFFUSE, new float[] {0.75f,0.75f,0.75f,1}, 0);
         }
         else {
            gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, new float[] {0.7f,0.7f,1}, 0 );
            gl.glMaterialfv(GL.GL_BACK, GL.GL_AMBIENT_AND_DIFFUSE, new float[] {0.8f,0.8f,0.5f}, 0 );
         }
         if (!dataIsValid) {
            computeSurfaceData();
            if (version_1_5) {  
                 // Set up VBOs for surface points and normals. Since these change
                 // pretty regularly, use GL.GL_DYNAMIC_DRAW.
               gl.glBindBuffer(GL.GL_ARRAY_BUFFER, vertexVBO);
               gl.glBufferData(GL.GL_ARRAY_BUFFER, 4*3*201*201, vBuf, GL.GL_DYNAMIC_DRAW);
               gl.glBindBuffer(GL.GL_ARRAY_BUFFER, normalVBO);
               gl.glBufferData(GL.GL_ARRAY_BUFFER, 4*3*201*201, nBuf, GL.GL_DYNAMIC_DRAW);
               gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
            }
         }
         gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
         gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
         if ( ! linesCheckbox.isSelected() ) {
            drawSurface(gl);  // Just draw the surface.
         }
         else {
            gl.glPolygonOffset(1,1);
            gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
            drawSurface(gl);
            gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);
            gl.glDisable(GL.GL_LIGHTING);
            gl.glColor3f(0,0,0);
            gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
            drawLinesOnSurface(gl);
            gl.glEnable(GL.GL_LIGHTING);
         }
         gl.glDisableClientState(GL.GL_VERTEX_ARRAY);
         gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
      }
   }
      
   
   /**
    * This method, which is called just once from init(), fills the index
    * buffers with the indices that are needed  by glDrawElements for 
    * drawing the surface and the lines on the surface.
    */
   private void makeElementBuffers() {
      for (int i = 0; i < 201; i+=10) {  // indices for drawing lines in x-direction
         for (int j = 0; j < 201; j++)
            xLineIndexBuffer.put(201*i + j);
      }
      for (int i = 0; i < 201; i+=10) {  // indices for drawing lines in y-direction
         for (int j = 0; j < 201; j++)
            yLineIndexBuffer.put(201*j + i);
      }
      for (int i = 0; i < 200; i++) {   // indices for drawing surface with GL_TRIANGLE_STRIPs
         for (int j = 0; j < 201; j++) {
            surfaceIndexBuffer.put(201*(i+1) + j);
            surfaceIndexBuffer.put(201*i + j);
         }
      }
      xLineIndexBuffer.rewind();
      yLineIndexBuffer.rewind();
      surfaceIndexBuffer.rewind();
   }
   
   
   /**
    * This method is called when the surface data needs to be recomputed.  The data
    * that is stored consists of two 201-by-201 grids of vectors, one for surface points
    * and one for surface normals.  The indices that are computed by makeElementBuffers()
    * and used by glDrawArrays() are indices into the lists of points/vectors in those grids.
    */
   private void computeSurfaceData() {
      double xmin = -5;
      double xmax = 5;
      double ymin = -5;
      double ymax = 5;
      double xRes = 200;
      double yRes = 200;
      float[] surfaceData = new float[201*3];
      float[] normalData = new float[201*3];
      double dx = (xmax - xmin)/xRes;
      double dy = (ymax - ymin)/yRes;
      for (int i = 0; i <= xRes; i++) {
         int v = 0;
         int n = 0;
         double y1 = ymin + dy*i;
         for (int j = 0; j <= yRes; j++) {
            double x = xmin + dx*j;
            func.setVariable('x', x);
            func.setVariable('y', y1);
            double z1 = func.value();
            float[] normal1 = computeUnitNormal(x, y1);
            surfaceData[v++] = (float)x;
            surfaceData[v++] = (float)y1;
            surfaceData[v++] = (float)z1;
            normalData[n++] = normal1[0];
            normalData[n++] = normal1[1];
            normalData[n++] = normal1[2];
         }
         vBuf.put(surfaceData,0,201*3);
         nBuf.put(normalData,0,201*3);
      }
      vBuf.rewind();
      nBuf.rewind();
      dataIsValid = true;
   }
   
   
   /**
    * Draw the surface as a series of triangle strips.
    */
   private void drawSurface(GL gl) {
      if (version_1_5) {
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, surfaceVBO);
         for (int i = 0; i < 200; i++)
            gl.glDrawElements(GL.GL_TRIANGLE_STRIP, 402, GL.GL_UNSIGNED_INT, 402*i*4);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
      }
      else {
         for (int i = 0; i < 200; i++) {
            surfaceIndexBuffer.position(402*i);
            gl.glDrawElements(GL.GL_TRIANGLE_STRIP, 402, GL.GL_UNSIGNED_INT, surfaceIndexBuffer);
         }
      }
   }
   
   /**
    * Draw lines on the surface, but only every 10-th line.
    */
   private void drawLinesOnSurface(GL gl) {
      if (version_1_5) {
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, xLineVBO);
         for (int i = 0; i <= 20; i++)
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, 201*i*4);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, yLineVBO);
         for (int i = 0; i <= 20; i++)
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, 201*i*4);
         gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
      }
      else {
         for (int i = 0; i <= 20; i++) {
            xLineIndexBuffer.position(201*i);
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, xLineIndexBuffer);
            yLineIndexBuffer.position(201*i);
            gl.glDrawElements(GL.GL_LINE_STRIP, 201, GL.GL_UNSIGNED_INT, yLineIndexBuffer);
         }
      }
   }
   
   /**
    * Compute a unit normal to the graph of z = func(x,y).
    * This is only an approximation, using nearby points instead
    * of exact derivatives.
    */
   private float[] computeUnitNormal(double x, double y) {
      double epsilon = 0.00001;
      func.setVariable('x',x);
      func.setVariable('y',y);
      double z = func.value();
      func.setVariable('x',x+epsilon);
      double z1 = func.value();
      func.setVariable('x',x);
      func.setVariable('y',y+epsilon);
      double z2 = func.value();
      // normal is (epsilon,0,z1-z) X (0,epsilon,z2-z)
      double a = -epsilon*(z1-z);
      double b = -epsilon*(z2-z);
      double c = epsilon*epsilon;
      double length = Math.sqrt(a*a + b*b + c*c);
      if (Double.isNaN(length) || Double.isInfinite(length))
         return new float[] { 0,0,1 };
      else
         return new float[] { (float)(a/length), (float)(b/length), (float)(c/length) };
   }
   
   public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
   }

   public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
         boolean deviceChanged) {
   }
}
